module Error
  ( Error(..), success, value, addErrors )
 where

import Monad

data Error a = FatalError String | Error String | Success a 
               deriving (Eq, Show, Read)

instance Functor Error where
   fmap f (Success a)    = Success (f a)
   fmap _ (Error e)      = Error e
   fmap _ (FatalError e) = FatalError e

instance Monad Error where 
   (Success x)    >>= k   = k x
   (Error e)      >>= _   = Error e
   (FatalError e) >>= _   = FatalError e
   return                 = Success
   fail s                 = FatalError s

instance MonadPlus Error where
   mzero                     = Error ""
   (FatalError e) `mplus` _  = FatalError e
   (Error _)      `mplus` ys = ys
   xs             `mplus` _  = xs

success :: (Error a) -> (String -> b) -> (a -> b) -> b
success (Error e) err _ = err e
success (Success a) _ f = f a

value e = success e error id

addErrors (Error msg) (Error msg1) = Error (msg ++ " or " ++ msg1)
addErrors (Error msg) _ = Error msg
addErrors emsg@(FatalError _) _ = emsg
addErrors (Success _) emsg = emsg
