module Main 
   ( module SourceLang 
   , module SourceLangParse 
   , ppsourceexpr
   , main
   ) where

import Monad
import System
--import Time

import Error

import SourceLang
import SourceLangParse
import SourceLangPrint
import SourceTypeCheck
import ResidualLang
import ResidualLangPrint
import ResidualReduction
import Spec
import SpecState
import SpecMonad
import PredsSimplify
import PredsSolving
--Added by @
import Scripts 
-- end @

solvingPredicates   =                        True
eliminatingEvidence = solvingPredicates   && False
raisingArity        = eliminatingEvidence && False


main = runEnvironment nullEnv
{-

main = do putStrLn "Specialising!"
          putStr "Enter filename: "; fn <- getLine
          --let fn = "tmp"
          mres <- pts fn
          case mres of
           Error msg -> error msg
           Success _ -> return ()
-}

altmain2 = do putStrLn "Started!"
              putStr "Reading preds..."
              strDelta <- readFile "inTest.delta"
              let delta = read strDelta

              let fn = "tmp"

              let fnorig = "examples/ex" ++ fn ++ ".pts"
              s  <- readFile fnorig
              let fnpts = "examples/ex" ++ fn ++ ".pts.tmp"
              let s1 = takeWhile (/='\12') s
              writeFile fnpts s1
              appendFile fnpts "\12\n\n{-\n"
              t0 <- return 0 --getClockTime
              markTime fnpts t0
              Success ((Data ddefs tst,tau),source_state) 
                       <- runIOE (processSource fnpts s1)
              let SM f = 
                    do setSourceTCstate source_state
                       exceptVblesUsedIn ddefs
                       ddefs'             <- specDatas ddefs
                       res <- simplify [] [] delta
                       traceSM "What's bad: " res
                       return res
              case f initialState of
                Error e        -> writeFile "outTest" e
                Success (cd,_) -> writeFile "outTest" (show cd)
               
altmain = do putStrLn "Started!"
             putStr "Reading state..."
             strState <- readFile "inTest.state"
             let state = read strState
             state `seq` putStr "Reading delta..."
             strDelta <- readFile "inTest.delta"
             let delta = read strDelta
             delta `seq` putStr "Reading expression and type..."
             stretau <- readFile "inTest.etau"
             let (e,tau) = (read stretau::(ResTerm,ResType))
             putStr "Reading variables..."
             strfvs <- readFile "inTest.fvs"
             let (ftvs,fsvs) = read strfvs
             ftvs `seq` fsvs `seq` putStrLn "Solving..."
             let SM f = do (cxt, delta') <- solveRT True ftvs fsvs delta
                           traceSMcooked "Final delta:" delta'
                           traceSMcooked "Context:" (noredfill cxt (RVar "X"))
                           return (cxt, delta')
             case f state of
              Error e        -> writeFile "outTest.error" e
              Success (cd,_) -> writeFile "outTest.error" (show cd)

specall = do mapM_ (\fn -> do putStr ("Example " ++ fn ++ "...")
                              pts fn
                              putStrLn (" Done!")
                   )
               (examples++dataexamples)

dataexamples = 
  [
    "Data01-1", "Data01-2"
  , "Data02-1", "Data02-2", "Data02-3", "Data02-4"
  , "Data03-1", "Data03-2", "Data03-3", "Data03-4", "Data03-5"
  , "Data04-1", "Data04-2", "Data04-3"
  , "Data05-1", "Data05-2", "Data05-3"
  , "Data06-1", "Data06-2", "Data06-3", "Data06-4"
  , "Data07-1", "Data07-2", "Data07-3", "Data07-4"
  , "Data07-5", "Data07-6", "Data07-7"
  , "Data08-1"
  ]

examples = 
  [
    "00-1", "00-2"
  , "01-1", "01-2"
  , "02-1", "02-2", "02-3", "02-4", "02-5"
  , "03-1", "03-2"
  , "04-1", "04-2", "04-3"
  , "05"
  , "06-1", "06-2", "06-3"
  , "07-1", "07-2", "07-3", "07-4"
  , "08"
  , "09-1", "09-2", "09-3", "09-4"
  , "10-1", "10-2", "10-3"
  , "11"
  , "12"
  , "13"
  , "14-1", "14-2", "14-3"
  , "15"
  , "16"
  , "20-1", "20-2", "20-3"
  , "21"
  --, "22-1" -- Not to solve = loops with new heuristic!
  --, "23"   -- Not to solve = loops with new heuristic!
  ] ++
  [
    "24"
  , "25-1", "25-2", "25-3", "25-4"
  , "25-5", "25-6", "25-7", "25-8"
  --, "30"
  --, "31"
  , "32-1", "32-2"
  , "100"
  , "00-0", "02-0", "14-0"
  --, "22-0" -- LOOPS!
  ]

moreexamples =
  [
    "Power"
  , "Fib"
  , "EE",   "EE1",  "EE2", "EE3", "EE4"
  , "PTS"
  ]

pts fn = 
  -- DOS! : do let fnorig = "examples\\ex" ++ fn ++ ".pts"
  do let fnorig = "examples/ex" ++ fn ++ ".pts"
     s  <- readFile fnorig
  -- DOS! :   let fnpts = "examples\\ex" ++ fn ++ ".pts.tmp"
     let fnpts = "examples/ex" ++ fn ++ ".pts.tmp"
     let s1 = takeWhile (/='\12') s
     writeFile fnpts s1
     appendFile fnpts "\12\n\n{-\n"
     t0 <- return 0 --getClockTime
     markTime fnpts t0 
     res <- runIOE (processPTS t0 fnpts s1)
     markTime fnpts t0
     appendFile fnpts "\n-}\n"
  -- DOS! :   system ("move /Y " ++ fnpts ++ " " ++ fnorig)
     system ("mv " ++ fnpts ++ " " ++ fnorig)
     return res
          
processSource fnpts s =
  do let me = parsesourceexpr s
     case me of
       Error msg -> 
        do ioeAppendFile fnpts "Cannot parse the program!\n\n"
           ioeAppendFile fnpts msg
           fail msg
       Success e -> 
        do let mtetau = typechecksource e
           case mtetau of
             Error msg -> 
              do ioeAppendFile fnpts
                            "Cannot infer a type for the source program!\n\n"
                 ioeAppendFile fnpts msg
                 fail msg
             Success (st@((te, tau), source_state)) -> 
              do ioeAppendFile fnpts "The source program.\n\n"
                 --ioeAppendFile fnpts (show te)
                 --ioeAppendFile fnpts "\n\n\n"
                 ioeAppendFile fnpts (ppsourceexpr te)
                 ioeAppendFile fnpts "\n ::\n\n"
                 ioeAppendFile fnpts (ppsourcetype tau)
                 return st

processPTS t0 fnpts s =
  do ((te,tau),source_state) <- processSource fnpts s
     ioeLift (markTime fnpts t0)
     specInterface "PTS..."
                   source_state
                   (te,tau) fnpts
                   False -- (not solvingPredicates)
                   False -- (not eliminatingEvidence)
                   False -- (not raisingArity)
                   "It does not specialise!\n\n"
                   "Its principal type specialisation.\n\n"
     when solvingPredicates 
          (do ioeLift (markTime fnpts t0)
              specInterface "Solving..."
                            source_state
                            (te,tau) fnpts
                            (solvingPredicates)
                            False -- (not eliminatingEvidence)
                            False -- (not raisingArity)
                            "It cannot be solved!\n\n"
                            "Its solution.\n\n"
          )
     when eliminatingEvidence
          (do ioeLift (markTime fnpts t0)
              specInterface "EE..."
                            source_state
                            (te,tau) fnpts
                            (solvingPredicates)
                            (eliminatingEvidence)
                            False -- (not raisingArity)
                            "Its evidence cannot be eliminated!\n\n"
                            "Its solution with evidence elimination.\n\n"
          )
     when raisingArity
          (do ioeLift (markTime fnpts t0)
              specInterface "Raising..."
                            source_state
                            (te,tau) fnpts
                            (solvingPredicates)
                            (eliminatingEvidence)
                            (raisingArity)
                            "It cannot be arity raised!\n\n"
                            "Its solution with arity raising.\n\n"
          )
     return ()
      
specInterface phase source_state te fnpts sP eE aR eMsg relStr =
  do let mesigma = specialisesource source_state sP eE aR te
     ioeLift (putStr phase)
     case mesigma of
       Error msg -> 
        do ioeAppendFile fnpts eMsg
           ioeAppendFile fnpts msg
           fail msg
       Success (e', sigma) -> 
        do ioeAppendFile fnpts relStr 
           --ioeAppendFile fnpts (show e')
           --ioeAppendFile fnpts "\n\n\n"
           --ioeAppendFile fnpts (show sigma)
           --ioeAppendFile fnpts "\n\n\n"
           ioeAppendFile fnpts (ppresidualterm e')
           ioeAppendFile fnpts "\n ::\n\n"
           ioeAppendFile fnpts (ppresidualscheme sigma)
           return ()

markTime fn t0 = 
{-
  do t <- getClockTime
     appendFile fn ("\nMark: " ++ show (diffClockTime t t0) ++ "\n")
-}
  do appendFile fn "\n---------------------------------------------\n"
     appendFile fn ("Mark: ")
  -- DOS! :   system ("date /t >>" ++ fn)
  -- DOS! :   system ("time /t >>" ++ fn)
     system ("date >>" ++ fn)
     appendFile fn "---------------------------------------------\n\n"

---------------------------------------------------------
data IOE a = IOE (IO (Error a))

instance Monad IOE where
  return x    = IOE (return (Success x))
  IOE m >>= f = IOE (do mres <- m
                        case mres of 
                          Error msg -> return (Error msg)
                          Success v -> runIOE (f v)
                    )
  fail e = IOE (return (Error e))

ioeLift io = IOE (do res <- io; return (Success res))
ioeAppendFile fn str = ioeLift (appendFile fn str)

runIOE (IOE m) = m

