-- MetaVble is a type to keep information about metavariables of certain
--   kind. The integer is the seed for the next unused variable, and the
--   table is used to keep instantiations of variables during type inference.

module MetaVbles(module MyPrelude, module Table, module MetaVbles) where

import List
import MyPrelude
import Table

data MetaVble a = MV (Int,[Int]) (Table (Perhaps a)) [Int] [[Int]]
                deriving (Read, Show)

initialmetavbles :: MetaVble a
initialmetavbles = MV (0,[]) (initT (No "")) [] []

setexceptsmetavble expt (MV (n,_) ts ms sms) = ((), MV (n,expt) ts ms sms)

freshmetavble :: MetaVble a -> (Int, MetaVble a)
freshmetavble (MV (n,expt) ts ms sms) = 
  let (n1,expt1) = next (n+1) expt 
      next n [] = (n,[])
      next n xs@(x:xs') = 
         case compare n x  of
           LT -> (n,xs)
           EQ -> next (n+1) xs'
           GT -> next n xs'
   in n1 `seq` expt1 `seq` (n1, MV (n1,expt1) ts ms sms) 

getmetavble :: Int -> MetaVble a -> (Perhaps a, MetaVble a)
getmetavble x mv@(MV _ ts _ _) = let pa = assocT ts x in pa `seq` (pa, mv)

setmetavble :: Int -> a -> MetaVble a -> ((), MetaVble a)
setmetavble x v (MV n ts ms sms) = ((), MV n (updateT x (Yes v) ts) ms sms)

setmetavbles :: [(Int, Perhaps a)] -> MetaVble a -> ((), MetaVble a)
setmetavbles xvs (MV n ts ms sms) = ((), MV n (setmvbles xvs ts) ms sms)
  where setmvbles [] mv = mv
        setmvbles ((x,v):xvs) ts = setmvbles xvs (updateT x v ts)

type GetTable a = MetaVble a -> (Table (Perhaps a), MetaVble a)
gettable :: GetTable a
gettable mv@(MV _ ts _ _) = (ts, mv)

getflattable mv@(MV _ ts _ _) = (flatTableExcept (No "") ts, mv)

seenmetavble :: Int -> MetaVble a -> (Bool, MetaVble a)
seenmetavble v mv@(MV _ _ ms  _) = (v `elem` ms, mv)

remembermetavble :: Int -> MetaVble a -> ((), MetaVble a)
remembermetavble v (MV n ts ms sms) = ((), MV n ts (v:ms) sms)

forgetmetavble :: Int -> MetaVble a -> ((), MetaVble a)
forgetmetavble v (MV n ts ms sms) = ((), MV n ts (ms\\[v]) sms)

forgetallmetavbles :: MetaVble a -> ((), MetaVble a)
forgetallmetavbles (MV n ts _  sms) = ((), (MV n ts [] sms))

pushmemmetavble :: MetaVble a -> ((), MetaVble a)
pushmemmetavble (MV n ts ms sms) = ((), MV n ts ms (ms:sms))

popmemmetavble :: MetaVble a -> ((), MetaVble a)
popmemmetavble (MV n ts _ (ms:sms)) = ((), MV n ts ms sms)

