module MyPrelude 
  (Perhaps(..), isYes, stripYes, fork, cross
 , ST
 , strictfoldl, toHead, appendsFile, (.*.), indexOf
 , subseteq, seteq, includes, whenM
 , accuml, pick, change, unite, mkId
 , fst3, snd3, thd3
 , Sum3(..)
 , isInL, stripInL, onlyInL
 , isInC, stripInC, onlyInC
 , isInR, stripInR, onlyInR
 , IfCase(..), ifcase
 , isLeft, stripLeft, onlyLeft
 , isRight, stripRight, onlyRight
 , dig2num, str2num
 , canthpnSo, outboundIn
 )
where

import List(union)
import Char

data Perhaps a = Yes a | No String deriving (Eq,Read, Show)

isYes (Yes _) = True
isYes _ = False

stripYes (Yes a) = a

type ST s a = s -> (a,s)

-- By John Hughes
strictfoldl op z [] = z
strictfoldl op z (x:xs) = 
    let z' = z `op` x 
     in z' `seq` (strictfoldl op z' xs)

toHead f [] = []
toHead f (x:xs) = f x : xs
toHeadSnd f (x, y) = (x, toHead f y)

appendsFile file [str] = appendFile file str
appendsFile file (str:strs) = appendFile file (str ++ "\n") >> appendsFile file strs

f .*. g = \(a,b) -> (f a, g b)
cross f g (x,y) = (f x, g y)
fork f g x = (f x, g x)

-- See elemIndex in List!
indexOf n xs = indexOf' 0 n xs
indexOf' i n [] = i
indexOf' i n (x:xs) = if n == x 
                       then i+1 
                       else indexOf' (i+1) n xs

subseteq, seteq :: Eq a => [a] -> [a] -> Bool
subseteq [] ys = True
subseteq (x:xs) [] = False
subseteq (x:xs) ys = x `elem` ys && subseteq xs ys

seteq xs ys = xs `subseteq` ys && ys `subseteq` xs

includes xs ys = subseteq ys xs

whenM x b = if b then Just x else Nothing

fst3 :: (a,b,c) -> a
fst3 (a,_,_) = a

snd3 :: (a,b,c) -> b
snd3 (_,b,_) = b

thd3 :: (a,b,c) -> c
thd3 (_,_,c) = c

change :: [(Int,b)] -> [b] -> [b]
change [] xs = xs
change ((i,x):ixs) xs = change' i x (change ixs xs)
   where change' 0 x (_:ys) = x:ys
         change' n x (y:ys) = y: change' (n-1) x ys

accuml :: (a -> b -> c -> a) -> a -> [b] -> [c] -> a
accuml f e []     []     = e
accuml f e (x:xs) (y:ys) = accuml f (f e x y) xs ys
accuml _ _ _  _  = error "Wrong number of parameters..."

pick :: [Int] -> [a] -> [a]
pick [] _ = []
pick (i:is) xs = xs!!i : pick is xs

unite :: Eq a => [[a]] -> [a]
unite [] =[] 
unite (xs:xss) = xs `union` unite xss

mkId x 0 = x
mkId x 1 = x++"'"
mkId x n = x++show n

data Sum3 a b c = InL a | InC b | InR c

isInL (InL _) = True
isInL _ = False

isInC (InC _) = True
isInC _ = False

isInR (InR _) = True
isInR _ = False

stripInL (InL a) = a
stripInC (InC a) = a
stripInR (InR a) = a

onlyInL = map stripInL . filter isInL
onlyInC = map stripInC . filter isInC
onlyInR = map stripInR . filter isInR

data IfCase a = Bool :--> a

ifcase :: [IfCase a] -> a
ifcase ((b :--> a):bs) = if b then a else ifcase bs

isLeft (Left _) = True
isLeft _ = False

isRight (Right _) = True
isRight _ = False

stripLeft (Left a) = a
stripRight (Right a) = a

onlyLeft = map stripLeft . filter isLeft
onlyRight = map stripRight . filter isRight

dig2num :: Char -> Int
dig2num d = ord d - ord '0'

str2num :: String -> Int
str2num ds = foldl (\n d -> n*10+dig2num d) 0 ds

cm2inch :: Double -> Double
cm2inch cm = cm / 2.54

inch2cm :: Double -> Double
inch2cm inch = inch * 2.54

canthpnSo msg = "Cannot happen! " ++ msg
outboundIn msg = "Index out of bound in " ++ msg

