module Nat (Nat) where

data Nat = N Integer
          -- El constructor N es abstracto!

-- Estos numeros naturales se usan como cualquier otro (incluidas
-- las constantes numericas), gracias al sistema de clases de Haskell
-- para numeros.
-- La unica diferencia es que el control de que un numero es natural o no 
--  esta hecho en tiempo de ejecucion, pero provista en la construccion de
--  los numeros. 
-- Eso hace que, por ejemplo, (2-3) :: Nat, y solo da error si trato de 
--  evaluarlo.

-- Constructor interno
nat :: Integer -> Nat
nat n = if n>=0 
         then N n 
         else error (show n ++ " no es un nmero natural!")

instance Eq Nat where
  (N n) == (N m) = n==m

instance Ord Nat where
  (N n) <= (N m) = n <= m

instance Enum Nat where
  fromEnum (N n) = fromEnum n
  toEnum n = nat (toEnum n)

instance Num Nat where
  (N n) + (N m) = nat (n+m)
  (N n) - (N m) = nat (n-m)
  (N n) * (N m) = nat (n*m)
  negate (N n)  = error "No se puede invertir un natural!"
  abs n = n
  signum (N n) = nat (signum n)
  fromInteger n = nat n

instance Show Nat where
  showsPrec p (N n) = showsPrec p n

instance Read Nat where
  readsPrec p s = [ (nat n, s') | (n,s') <- readsPrec p s ]
  
instance Real Nat where
  toRational (N n) = toRational n

instance Integral Nat where
  quot (N n) (N m) = nat (quot n m)
  rem (N n) (N m) = nat (rem n m)
  div (N n) (N m) = nat (div n m)
  mod (N n) (N m) = nat (mod n m)
  quotRem (N n) (N m) = let (q,r) = quotRem n m in (nat q, nat r)
  divMod (N n) (N m) = let (q,r) = divMod n m in (nat q, nat r)
  toInteger (N n) = n

