module PredsSimplify
--   (glb, glbs, simplify, simplifySolving, isMoreGeneral)
 where

import Monad

import List(intersect)
import Maybe

import MyPrelude
import MoreMonad

import SourceLang
import qualified SourceTypeCheckMonad
import qualified SourceUnification
import ResidualLang
import ResidualLangPrint
import SpecState
import SpecMonad
import ResidualReduction
import ResidualUnification
import ResidualGeneralizable
import ResidualLangFVs
import SpecAuxRels

-- ---------------------------
-- Simplification of contexts
--  * it keeps the number of predicates to the minimum needed
--  * it is applicable at any stage, provided the term is changed
--     with the resulting ContextTerm
-- ---------------------------
simplify :: [ResTypeVble] -> [ResSchemeVble]
            -> Preds -> SpecM (ContextTerm, Preds)
simplify ftvs fsvs delta = 
                do (hs,vs,delta') <- simplify' 1 ftvs fsvs False delta
                   return (replaceEv hs vs, delta')

simplifySolving :: [ResTypeVble] -> [ResSchemeVble]
                   -> Preds -> SpecM ([EvVble],[EvidenceTerm], Preds)
simplifySolving ftvs fsvs delta = 
        do (hs,vs,delta') <- simplify' 1 ftvs fsvs True delta
           ----------------------------------
           ------- ALEJANDRO PELIGRO!!!!!!!!!
           --------- Cambiar a TRUE x q????
           ----------------------------------
           return (hs, vs, delta')

-- ---------------------------
-- First and second arguments are the type and scheme variables 
--   free in the context
-- ---------------------------
simplify' :: Int -> [ResTypeVble] -> [ResSchemeVble] -> Bool
                   -> Preds -> SpecM ([EvVble],[EvidenceTerm], Preds)
simplify' i ftvs fsvs forcing delta = 
 do --traceSM "simplify'' : " delta
    res@(hs,vs,delta') <- simplify'' ftvs fsvs forcing [] delta
    if delta' == delta
     then return res
     else do (hs',vs',delta'') <- simplify' (i+1) ftvs fsvs forcing delta'
             return (hs, map (\v -> accuml evsubst v hs' vs') vs, delta'')

-- Third argument is a flag, indicating which function to use for MoreGeneral
-- Fourth argument contains those preds already considered and not simplified
simplify'' _    _    _       _      [] = return ([], [], [])
simplify'' ftvs fsvs forcing delta2 (d@(v@(EVar h), pred):delta) = 
 do pred' <- unfoldAll pred
    traceSM "Entling h=" (v,pred)
    mv    <- entl True forcing (delta2++delta) pred'
    case mv of
     Just (v',ps) -> 
      -- PARCHE
      if null ps
       then 
          do traceSM "Adding to context (for entl): " (h,ps)
             (hs,vs,delta') <- simplify'' ftvs fsvs forcing delta2 (delta++ps)
             return (h:hs, v':vs, delta')
       else 
          do mres <- entl'' True forcing (delta2++delta) pred'
             case mres of
              Nothing -> 
               do traceSM "Adding to context (for entl): " (h,ps)
                  (hs,vs,delta') <- simplify'' ftvs fsvs forcing delta2 
                                                                 (delta++ps)
                  return (h:hs, v':vs, delta')
              Just (v'',preds') -> 
               do (hs,vs,delta') <- simplify'' ftvs fsvs forcing delta2 
                                                                 (delta++preds')
                  return (h:hs, v'':vs, delta')
      -- PARCHE
     Nothing      -> 
      do mpreds <- simplifyPred ftvs fsvs v pred'
         case mpreds of
           Nothing -> 
             do (hs,vs,delta') <- simplify'' ftvs fsvs forcing 
                                             (d:delta2) delta
                return (h:hs, v:vs, (v,pred'):delta')
           Just (v',preds) ->
             do traceSM "Adding to context (for simplPred): " (h,preds)
                (hs,vs,delta') <- simplify'' ftvs fsvs forcing 
                                             delta2 (delta++preds)
                return (h:hs, v':vs, delta')

-- ---------------------------
-- Simple entailment for simplification (Delta,Delta)
--  (Just v, if h:delta ||- v : delta'; Nothing, otherwise)
-- ---------------------------
entls :: Bool -> Preds -> Preds -> SpecM (Maybe ([EvidenceTerm],Preds))
entls _ _ [] = return (Just ([],[]))
entls efecUnif delta ((h,d):delta') = 
   do mv  <- entl efecUnif False delta d
      case mv of
       Nothing      -> return Nothing
       Just (v,ps1) ->
         do mvs <- entls efecUnif delta delta'
            case mvs of
              Nothing       -> return Nothing
              Just (vs,ps2) -> return (Just (v:vs, ps1++ps2))

------
-- Establishes which predicates of delta' are not entailed by delta
------
notentlby :: Preds -> Preds -> SpecM ([EvidenceTerm],Preds)
notentlby [] _ = return ([],[])
notentlby ((_,d):delta') delta =
   do newTemporarySubst
      mv               <- entl False False delta d
      case mv of
        Nothing        -> do discardUnification
                             (vs, deltaKeep') <- delta' `notentlby` delta
                             h <- freshevidence
                             return (h:vs, (h,d):deltaKeep')
        Just (v,preds) -> if null preds
                           then do confirmUnification
                                   (vs, deltaKeep') <- delta' `notentlby` delta
                                   return (v:vs, deltaKeep')
                           else do traceSM "Tricky entl: " (delta,d)
                                   traceSM "results in: " (v,preds)
                                   fail (canthpnSo "Have you messed with entl?")

-- ---------------------------
-- Simple entailment for simplification (delta,d)
--  (Just v, if h:delta ||- v : d; Nothing, otherwise)
-- ---------------------------
entl, entl' :: Bool  -> Bool 
            -> Preds -> Pred -> SpecM (Maybe (EvidenceTerm,Preds)) 
entl efecUnif forcing delta d = 
 do d' <- unfoldAll d
    delta' <- unfoldAll delta -- mmmmmmmmmmmmm
    entl' efecUnif forcing delta' d'

entl' efecUnif forcing delta pred@(IsInt tau) = 
   case tau of
    RTInt n      -> return (Just (EInt  n,[]))
    RTVar _      -> entl'' efecUnif forcing delta pred
    RTFixS _ _ _ -> let msg = "Static fix with numbers may not terminate!" in
                    failentl efecUnif forcing msg
    _            -> let msg = "Non-integer as number!" in
                    failentl efecUnif forcing msg

entl' efecUnif forcing delta pred@(IsBool tau) =
   case tau of
    RTBool b     -> return (Just (EBool b,[]))
    RTFixS _ _ _ -> let msg = "Static fix with booleans may not terminate!" in
                    failentl efecUnif forcing msg
    RTVar  _     -> entl'' efecUnif forcing delta pred
    _            -> let msg = "Non-boolean as truth-value!" in
                    failentl efecUnif forcing msg

entl' efecUnif forcing delta pred@(IsIf tau taub tau1 tau2) = 
  case taub of
    RTBool b -> if b then do yes <- efectiveUnify tau tau1
                             if not yes 
                              then do msg <- unifErrMsg tau tau1
                                      failentl efecUnif forcing msg
                              else return (Just (evoid, []))
                     else do yes <- efectiveUnify tau tau2 
                             if not yes 
                              then do msg <- unifErrMsg tau tau1
                                      failentl efecUnif forcing msg
                              else return (Just (evoid, []))
    RTVar _      -> entl'' efecUnif forcing delta pred
    RTFixS _ _ _ -> let msg = "Static fix with booleans may not terminate!" in
                    failentl efecUnif forcing msg
    _            -> let msg = "Non-boolean as conditional!" in
                    failentl efecUnif forcing msg

entl' efecUnif forcing delta pred@(IsCase tau taue cts) = 
  case taue of
    RTConstruc c targs -> 
      let chooseBranch _ c' _ [] = 
            fail ("Case statment does not cover "++c++".")
          chooseBranch tau' c' targs' ((ctau,taur):cts') =
            case ctau of
             RTConstruc c'' targs'' -> 
               if c'==c''
                then 
                 do yeses <- mapM (\(t,t') -> efectiveUnify t t')
                                  (zip targs' targs'')
                    if not (and yeses)
                     then do let msg = "Cannot unify arguments in a pattern!"
                             failentl efecUnif forcing msg
                     else do yes <- efectiveUnify tau' taur
                             if not yes 
                              then do msg <- unifErrMsg tau taur
                                      failentl efecUnif forcing msg
                              else return (Just (evoid, []))
                else chooseBranch tau' c' targs' cts'
             _ -> fail "Cannot happen! Non-constructor in case patterns!"
       in chooseBranch tau c targs cts
    RTVar _      -> entl'' efecUnif forcing delta pred
    RTFixS _ _ _ -> let msg = "Static fix with datatypes may not terminate!" in
                    failentl efecUnif forcing msg
    _            -> let msg = "Non-constructor in constr_guard!" in
                    failentl efecUnif forcing msg

entl' efecUnif forcing delta pred@(IsArithOp nop tau tau1 tau2) = 
  case tau1 of
   RTInt n1 -> 
         case tau2 of
          RTInt n2 -> do let n = arithop nop n1 n2
                         yes <- efectiveUnify tau (RTInt n)
                         if not yes 
                          then do msg <- unifErrMsg tau (RTInt n)
                                  if forcing 
                                   then fail msg
                                   else entl'' efecUnif forcing delta pred
                          else return (Just (EInt n, []))
          RTVar _      -> entl'' efecUnif forcing delta pred
          RTFixS _ _ _ -> let msg = ("Static fix with numbers "
                                    ++"may not terminate!") in
                          failentl efecUnif forcing msg
          _            -> let msg = "Non-intenger as number!" in
                          failentl efecUnif forcing msg
   RTVar _      -> entl'' efecUnif forcing delta pred
   RTFixS _ _ _ -> let msg = "Static fix with numbers may not terminate!" in
                   failentl efecUnif forcing msg
   _            -> let msg = "Non-intenger as number!" in
                   failentl efecUnif forcing msg

entl' efecUnif forcing delta pred@(IsBoolOp  bop tau tau1 tau2) =               
  case tau1 of
   RTBool b1 -> 
          case tau2 of
           RTBool b2 -> do let b = boolop bop b1 b2
                           yes <- efectiveUnify tau (RTBool b)
                           if not yes 
                            then do msg <- unifErrMsg tau (RTBool b)
                                    if forcing 
                                     then fail msg
                                     else entl'' efecUnif forcing delta pred
                            else return (Just (EBool b, []))
           RTVar _      -> entl'' efecUnif forcing delta pred
           RTFixS _ _ _ -> let msg = ("Static fix with booleans "
                                      ++"may not terminate!") in
                           failentl efecUnif forcing msg
           _            -> let msg = "Non-boolean as truth-value!" in
                           failentl efecUnif forcing msg
   RTVar _      -> entl'' efecUnif forcing delta pred
   RTFixS _ _ _ -> let msg = "Static fix with booleans may not terminate!" in
                   failentl efecUnif forcing msg
   _            -> let msg = "Non-boolean as truth-value!" in
                   failentl efecUnif forcing msg

entl' efecUnif forcing delta pred@(IsRelOp rop tau tau1 tau2) =
  case tau1 of
   RTBool b1 -> 
         case tau2 of
          RTBool b2 -> do let b = relop rop b1 b2
                          yes <- efectiveUnify tau (RTBool b)
                          if not yes 
                           then 
                            do msg <- unifErrMsg tau (RTBool b)
                               failentl efecUnif forcing msg
                           else return (Just (EBool b, []))
          RTVar _      -> entl'' efecUnif forcing delta pred
          RTFixS _ _ _ -> let msg = ("Static fix with booleans "
                                     ++"may not terminate!") in
                          failentl efecUnif forcing msg
          _            -> let msg = "Non-boolean as truth-value!" in
                          failentl efecUnif forcing msg
   RTInt n1 -> 
        case tau2 of
         RTInt n2 -> do let b = relop rop n1 n2
                        yes <- efectiveUnify tau (RTBool b)
                        if not yes 
                         then do msg <- unifErrMsg tau (RTBool b)
                                 failentl efecUnif forcing msg
                         else return (Just (EBool b, []))
         RTVar _      -> entl'' efecUnif forcing delta pred
         RTFixS _ _ _ -> let msg = ("Static fix with numbers "
                                    ++"may not terminate!") in
                         failentl efecUnif forcing msg
         _            -> let msg = "Non-intenger as number!" in
                         failentl efecUnif forcing msg
   RTVar _      -> entl'' efecUnif forcing delta pred
   RTFixS _ _ _ -> let msg = ("Static fix with numbers or booleans "
                                    ++"may not terminate!") in
                   failentl efecUnif forcing msg
   _            -> let msg = "Relation between non-base types!" in
                   failentl efecUnif forcing msg

entl' efecUnif forcing delta pred@(IsConstrOf datatype tau) = 
  case tau of
   RTConstruc c args -> 
     do source_args <- sourceTypeCheck 
         (do c_sourcetype <- SourceTypeCheckMonad.constructorType c
             let (args, taud) = splitConstrType c_sourcetype
             SourceUnification.unify taud datatype
             args' <- SourceTypeCheckMonad.unfoldVbles True args
             return args'
         )
        (deltas,yes) <- checkSRfor args source_args
        if not yes
         then do msg <- return "Constructor argument failed check in IsConstrOf"
                 failentl efecUnif forcing msg
         else if null (concat deltas)
               then return (Just (EConstr c,[]))
               else do mvps <- entls efecUnif delta (concat deltas)
                       case mvps of
                        Nothing -> entl'' efecUnif forcing delta pred
                        Just (_,ps) -> if null ps 
                                        then return (Just (EConstr c,[]))
                                        else entl'' efecUnif forcing delta pred
   RTVar  _     -> entl'' efecUnif forcing delta pred
   _            -> let msg = "Non-constructor as static data!" in
                   failentl efecUnif forcing msg

-----------------------------------
---- Begin Alejandro --------------
-----------------------------------
entl' efecUnif forcing delta pred@(HasC td c tau) = 
    do td' <- indir td 
       case td' of 
          (RTData dname _ _) -> do (d:numberdata) <- return dname
                                   (d:_:number) <- return dname
                                   n <- return ((read number)::Int)
                                   traceSM "Provando HASC >>>>" pred
                                   traceSM "Con >>>>>>>" n
                                   rt <- getconstructorRType (c++numberdata)
                                   -- h <- freshevidence
                                   -- Se tiene que hacer la unificacion
                                   -- si devuelvo el predicado no anda
                                   -- pues esto se tiene que probar 
                                   -- desde el vacio 
                                   -- fidel?
                                   entl' efecUnif forcing 
                                         delta (Unify rt tau)
                                   return (Just (EInt n, []))
{-
                                   return (Just (EInt n, 
                                           [(h, Unify rt tau)]))
-} 

          RTVar _      -> entl'' efecUnif forcing delta pred
          _            -> return Nothing

entl' efecUnif forcing delta gpred@(GuardedDinData c tau p) = 
     do td' <- indir tau 
        case td' of 
            (RTData dname _ _) -> do names <- getconstructorRnames dname
                                     let (d:number) = dname 
                                     let w = (c++number) `elem` names
                                     if not w then return (Just (evoid, []))
                                              else do h <- freshevidence
                                                      entl' efecUnif 
                                                            forcing delta p
{-
                                                      return (Just (h, [(h, p)]))
-}
            RTVar _      -> entl'' efecUnif forcing delta gpred
            _ -> return Nothing
 

entl' efecUnif forcing delta pred@(ConstrsOf tau) = 
     do td' <- indir tau 
        case td' of 
            (RTData dname _ _) -> do names <- getconstructorRnames dname
 
                                     traceSM "Probando ConstrsOf con" (EConstrs names) 
                                     return (Just (EConstrs names, []))
            RTVar _      -> entl'' efecUnif forcing delta pred
            _ -> return Nothing

-----------------------------------
------ End Alejandro --------------
-----------------------------------
entl' efecUnif forcing delta gpred@(Guarded taub pred) = 
  case taub of
   RTBool b     -> if b 
                    then do mv <- entl' efecUnif forcing delta pred
                            case mv of
                             Just _  -> return mv
                             Nothing -> entl'' efecUnif forcing delta gpred
                    else return (Just (evoid, []))
   RTVar  _     -> entl'' efecUnif forcing delta gpred
   RTFixS _ _ _ -> let msg = "Static fix with booleans may not terminate!" in
                   failentl efecUnif forcing msg
   _            -> let msg = "Non-boolean as guard!" in
                   failentl efecUnif forcing msg

entl' efecUnif forcing delta gpred@(GuardedNeg taub pred) = 
  case taub of
   RTBool b     -> if not b 
                    then do mv <- entl' efecUnif forcing delta pred
                            case mv of
                             Just _  -> return mv
                             Nothing -> entl'' efecUnif forcing delta gpred
                    else return (Just (evoid, []))
   RTVar  _     -> entl'' efecUnif forcing delta gpred
   RTFixS _ _ _ -> let msg = "Static fix with booleans may not terminate!" in
                   failentl efecUnif forcing msg
   _            -> let msg = "Non-boolean as guard!" in
                   failentl efecUnif forcing msg

entl' efecUnif forcing delta gpred@(GuardedConstr c tauc pred) = 
  case tauc of
   RTConstruc c' _ -> if c==c' 
                       then do mv <- entl' efecUnif forcing delta pred
                               case mv of
                                Just _  -> return mv
                                Nothing -> entl'' efecUnif forcing delta gpred
                       else return (Just (evoid, []))
   RTVar  _     -> entl'' efecUnif forcing delta gpred
   RTFixS _ _ _ -> let msg = "Static fix with datatypes may not terminate!" in
                   failentl efecUnif forcing msg
   _            -> let msg = "Non-constructor in constr_guard!" in
                   failentl efecUnif forcing msg

entl' efecUnif forcing delta pred@(Unify tau1 tau2) = 
   if not efecUnif then return Nothing else
   do yes <- efectiveUnify tau1 tau2
      if not yes 
       then do msg <- unifErrMsg tau1 tau2
               failentl efecUnif forcing msg
       else return (Just (evoid, []))

entl' efecUnif forcing delta pred@(Fail msg) = 
   entl'' efecUnif forcing delta pred

entl' efecUnif forcing delta 
      pred@(IsFunS t@(RTTypedS taus  taue  sigma@(RSForAll _ _ _)) 
                   t'@(RTTypedS taus' taue' sigma'@(RSForAll _ _ _))) = 
   let ismg = if forcing 
               then forceMoreGeneral 
               else isMG efecUnif
    in do mcxtp <- ismg delta sigma sigma'
          case mcxtp of
           Just (cxt, preds) -> do
              case taue of
               RTStaticExp e -> 
                 do let e' = redfill cxt e
                    yes <- efectiveUnify taue' (RTStaticExp e')
                    if not yes
                     then do msg <- unifErrMsg taue (RTStaticExp e')
                             failentl efecUnif forcing msg
                     else do yes <- efectiveUnify taus taus'
                             if not yes
                              then do msg <- unifErrMsg taus taus'
                                      failentl efecUnif forcing msg
                              else return (Just (EStaticExp e', preds))
               RTInvertCxt c e ->
                 do c' <- unfoldAll c
                    cxt' <- unfoldAll cxt
                    return (Just (EStaticExp e, preds))
                    --if c'==cxt' then return (Just (EStaticExp e, preds))
                    --            else fail ("WARNING: " ++ show c' ++ 
                    --                       " not equal to " ++ show cxt')
                    -- What happens when c/=cxt?
               RTVar _  -> 
                 case taue' of
                  RTStaticExp e' ->
                             do yes <- efectiveUnify taus taus'
                                if not yes
                                 then do msg <- unifErrMsg taus taus'
                                         failentl efecUnif forcing msg
                                 else return (Just (EStaticExp e', preds))
                  _ -> entl'' efecUnif forcing delta pred
               _        -> let msg = ("Non-function as closure! (" 
                                       ++ show taue) in
                           failentl efecUnif forcing msg
           Nothing -> let msg = "Cannot satisfy the predicate\n"
                                ++ ppresidualpredicate pred      in
                      failentl efecUnif forcing msg

entl' efecUnif forcing delta 
      pred@(IsFunS tauU@(RTFixS   taus  taue  sigma@(RSForAll _ _ _)) 
                   tauL@(RTTypedS taus' taue' sigma'@(RSForAll _ _ _))) = 
   if forcing
    then do h  <- freshevidence
            h' <- freshevidence
            t  <- freshresidualtype
            ts <- freshresidualtype
            te <- freshresidualtype
            sf <- freshresidualtermBasedOn "rf"
            let th  = evidence2residual h
            let th' = evidence2residual h'
            let v = EStaticExp 
                     (RLam sf 
                       (REAppS th' (REAppS (REAppS th (RVar sf)) (RVar sf)))
                     )
            return (Just (v
                         ,[ (h', IsFunS t tauL)
                          , (h , IsFunS (RTTypedS taus taue sigma)
                                        (RTTypedS ts   te   
                                                  (scheme (RTFun tauU t))))
                          ]
                         ))
    else entl'' efecUnif forcing delta pred

entl' efecUnif forcing delta 
      pred@(IsFunS tauU@(RTTypedS taus  taue  sigma@(RSForAll _ _ _)) 
                   tauL@(RTFixS   taus' taue' sigma'@(RSForAll _ _ _))) = 
{-
  -- This loops! The addition of pred to the context is not representable
  --  in this way...
   if forcing
    then do h  <- freshevidence
            h' <- freshevidence
            t  <- freshresidualtype
            ts <- freshresidualtype
            te <- freshresidualtype
            return (Just (h,
                          [ (h, IsFunS tauU t)
                          , (h', IsFunS (RTTypedS taus' taue' sigma')
                                        (RTTypedS ts    te    (scheme (RTFun tauL t))))
                          ]))
    else entl'' efecUnif forcing delta pred
-}
  -- What about this? 
  -- Warning: it needs extensions to singleentl (case FunS TypedS FixS)
   if forcing
    then 
     do h  <- freshevidence
        t  <- freshresidualtype
        ts <- freshresidualtype
        te <- freshresidualtype
        mv <- entl' efecUnif forcing ((h,pred):delta) 
                    (IsFunS (RTTypedS taus' taue' sigma')
                            (RTTypedS ts    te    (scheme (RTFun tauL t)))
                    )
        case mv of
          Nothing -> do let msg = ("Check of upper bound to a FixS failed (1)"
                                  )
                        failentl efecUnif forcing msg
          Just (_,preds) ->
            do mv' <- entl efecUnif forcing (preds++delta) (IsFunS tauU t)
               case mv' of
                 Nothing -> 
                  do ((tauU',t'),pred') <- unfoldAll ((tauU,t),pred)
                     let msg = ("Check of upper bound to a FixS failed (2)"
                                ++"\nUpper bound checked: "
                                ++ ppresidualpredicate (IsFunS tauU' t')
                                ++ "\nentailed by "
                                ++ ppresidualpredicates (preds++delta)
                                ++ "\n\nWhen attempting to solve " 
                                ++ ppresidualpredicate pred'
                               )
                     failentl efecUnif forcing msg
                 Just (v,preds') -> return (Just (v, preds++preds'))
    else entl'' efecUnif forcing delta pred    

entl' efecUnif forcing delta 
      pred@(IsFixS (RTTypedS taus  taue  sigma@(RSForAll _ _ _)) 
                   (RTFixS   taus' taue' sigma'@(RSForAll _ _ _))) = 
   let ismg = if forcing 
               then forceMoreGeneral 
               else isMG efecUnif
    in do mcxtp <- ismg delta sigma sigma'
          case mcxtp of
           Just (cxt, preds) -> 
              case taue of
               RTStaticExp e -> 
                 do let e' = redfill cxt e
                    yes <- efectiveUnify taue' (RTStaticExp e')
                    if not yes
                     then do msg <- unifErrMsg taue (RTStaticExp e')
                             failentl efecUnif forcing msg
                     else do yes <- efectiveUnify taus taus'
                             if not yes
                              then do msg <- unifErrMsg taus taus'
                                      failentl efecUnif forcing msg
                              else return (Just (EStaticExp e', preds))
               RTVar _  -> entl'' efecUnif forcing delta pred
               _        -> let msg = "Non-function as closure!" in
                           failentl efecUnif forcing msg
           Nothing -> let msg = "Cannot satisfy the predicate\n"
                                ++ ppresidualpredicate pred      in
                      failentl efecUnif forcing msg

entl' efecUnif forcing delta 
      pred@(IsMG sigma@(RSForAll  _ _ _)
                 sigma'@(RSForAll _ _ _)) =
   let ismg = if forcing 
               then forceMoreGeneral 
               else isMG efecUnif
    in do mcxtp <- ismg delta sigma sigma'
          case mcxtp of
           Just (cxt, preds) -> 
                      do traceSM ("More general only if: " ++ 
                                   show (takeEvidence preds)) (sigma, sigma')
                         return (Just (ECxt cxt, preds))
           Nothing -> let msg = "Cannot satisfy the predicate\n"
                                ++ ppresidualpredicate pred      in
                      failentl efecUnif forcing msg

entl' efecUnif forcing delta pred = entl'' efecUnif forcing delta pred

------
entl'' efecUnif forcing delta dr = entl''' efecUnif forcing [] delta dr

entl''' efecUnif _       _ []                      _  = return Nothing
entl''' efecUnif forcing delta2 (vd@(v@(EVar h), d):delta) dr = 
     do mv <- singleentl_unfold efecUnif forcing (delta2++delta) h d dr
        case mv of
         Just (nv,preds) -> 
           if null preds
            then return mv
            else do mv' <- entls efecUnif (delta++(vd:delta2)) preds
                    case mv' of
                      Nothing -> entl''' efecUnif forcing (vd:delta2) 
                                                          delta dr
                      Just (_,ps') -> 
                        if null ps' 
                         then return (Just (nv,[]))
                         else entl''' efecUnif forcing (vd:delta2) 
                                               delta dr
         Nothing -> entl''' efecUnif forcing 
                                     (vd:delta2) delta dr
------
failentl efecUnif forcing msg = 
  if efecUnif
   then if forcing 
         then fail msg
         else do h <- freshevidence
                 return (Just (h, [(h,Fail msg)]))
   else return Nothing

-- ---------------------------
-- Simple entailment for simplification (delta,delta)
-- Just v, if h:d ||- v:d', Nothing, otherwise
-- ---------------------------
--singleentl :: Preds -> Int -> Pred -> Pred 
--              -> SpecM (Maybe (EvidenceTerm,Preds))

--------------------------
-- Begin Alejandro -------
--------------------------
-- Now, unification can happens at this level
singleentl_unfold 
      efecUnif forcing delta h d d' = do delta' <- unfoldAll delta
                                         du <- unfoldAll d
                                         du' <- unfoldAll d' 
                                         singleentl efecUnif forcing
                                                    delta' h du du'
--------------------------
---- End Alejandro -------
--------------------------

singleentl _ _ _ h d d' | d==d' = return (Just (EVar h, []))

singleentl _ _ _ h (IsArithOp _ tau _ _) (IsInt  tau') = 
                       if tau==tau'
                        then return (Just (EVar h, []))
                        else return Nothing

singleentl _ _ _ h (IsBoolOp  _ tau _ _) (IsBool tau') = 
                       if tau==tau'
                        then return (Just (EVar h, []))
                        else return Nothing

singleentl _ _ _ h (IsRelOp   _ tau _ _) (IsBool tau') = 
                       if tau==tau'
                        then return (Just (EVar h, []))
                        else return Nothing

{- Bad patch! Better solved when introducing IsIf... 
singleentl _ forcing delta h (IsIf tau taub tau1 tau2) d =
  case d of
   IsInt tau'  -> if tau==tau'
                   then do p1<- case tau1 of
                                 RTInt _ -> return []
                                 _ -> do h1 <- freshevidence
                                         return [(h1,Guarded taub 
                                                      (IsInt tau1))]
                           p2<- case tau2 of
                                 RTInt _ -> return []
                                 _ -> do h2 <- freshevidence
                                         return [(h2,GuardedNeg taub 
                                                      (IsInt tau2))]
                           return (Just (EVar h,p1++p2))
                   else return Nothing
   IsBool tau' -> if taub==tau'
                   then return (Just (EVar h, []))
                   else 
                     if tau==tau'
                      then do p1<- case tau1 of
                                    RTBool _ -> return []
                                    _ -> do h1 <- freshevidence
                                            return [(h1,Guarded taub 
                                                         (IsBool tau1))]
                              p2<- case tau2 of
                                    RTBool _ -> return []
                                    _ -> do h2 <- freshevidence
                                            return [(h2,GuardedNeg taub 
                                                         (IsBool tau2))]
                              return (Just (EVar h,p1++p2))
                      else return Nothing
   IsConstrOf dt tau' -> 
      if tau==tau'
       then do p1<- case tau1 of
                     RTConstruc c [] ->
                      do source_args <- sourceTypeCheck
                          (do c_sourcetype <- SourceTypeCheckMonad.constructorType c
                              let (args, taud) = splitConstrType c_sourcetype
                              SourceUnification.unify taud dt
                              return args
                          )
                         if null source_args
                          then return []
                          else 
                           do h1 <- freshevidence
                              return [(h1,Guarded taub 
                                           (IsConstrOf dt tau1))]
                     _ -> do h1 <- freshevidence
                             return [(h1,Guarded taub 
                                          (IsConstrOf dt tau1))]
               p2<- case tau2 of
                     RTConstruc c [] ->
                      do source_args <- sourceTypeCheck
                          (do c_sourcetype <- SourceTypeCheckMonad.constructorType c
                              let (args, taud) = splitConstrType c_sourcetype
                              SourceUnification.unify taud dt
                              return args
                          )
                         if null source_args 
                          then return []
                          else 
                           do h2 <- freshevidence
                              return [(h2,GuardedNeg taub 
                                           (IsConstrOf dt tau2))]
                     _ -> do h2 <- freshevidence
                             return [(h2,GuardedNeg taub 
                                          (IsConstrOf dt tau2))]
               return (Just (EVar h,p1++p2))
       else return Nothing
   _           -> return Nothing
-}

-----------------------
--- Begin Alejandro ---
-----------------------
singleentl efecUnif forcing delta h 
           pred@(HasC td c tau) 
           pred2@(HasC td' c' tau') = 
                if td==td' && c==c'  
                   then do yes <- efectiveUnify tau tau'
                           if not yes 
                                  then do msg <- unifErrMsg tau tau'
                                          failentl efecUnif forcing msg
                                  else return (Just (EVar h, []))
                   else return Nothing

singleentl efecUnif forcing delta h 
           pred@(HasC td c tau) 
           g@(GuardedDinData c' td' p) = 
                if td==td' && c==c' 
                   then do hp <- freshevidence
                           entl efecUnif forcing ((EVar h, pred):delta) p 
                   else return Nothing

{-
singleentl efecUnif forcing delta h 
           (GuardedDinData c td p)
           (GuardedDinData c' td' p') = 
                if td==td' && c==c' 
                   then singleentl efecUnif forcing delta h p p' 
                   else return Nothing
-}

-----------------------
---- End Alejandro ----
-----------------------
singleentl efecUnif forcing delta h 
           (Guarded tau' pred') 
           (Guarded tau  pred) = 
                       if tau==tau'
                        then singleentl efecUnif forcing delta h pred' pred
                        else return Nothing

singleentl efecUnif forcing delta h 
           (GuardedNeg tau' pred') 
           (GuardedNeg tau  pred) = 
                       if tau==tau'
                        then singleentl efecUnif forcing delta h pred' pred
                        else return Nothing

singleentl efecUnif forcing delta h 
           (GuardedConstr c  tau' pred') 
           (GuardedConstr c' tau  pred) = 
                       if c==c'&&tau==tau'
                        then singleentl efecUnif forcing delta h pred' pred
                        else return Nothing

singleentl efecUnif forcing delta h (Guarded (RTBool True) pred') d =
   singleentl efecUnif forcing delta h pred' d

singleentl efecUnif forcing delta h (GuardedNeg (RTBool False) pred') d =
   singleentl efecUnif forcing delta h pred' d

singleentl efecUnif forcing delta h 
           (GuardedConstr c (RTConstruc c' _) pred') d | c==c' =
   singleentl efecUnif forcing delta h pred' d

--singleentl efecUnif _ delta h
--      (IsFunS (RTTypedS taus  taue  sigma@(RSForAll _ _ _))  (RTVar t)) 
--      (IsFunS (RTTypedS taus' taue' sigma'@(RSForAll _ _ _)) (RTVar t')) = 
--   if t==t'
singleentl efecUnif _ delta h
           (IsFunS (RTTypedS taus  taue  sigma@(RSForAll _ _ _))  tau) 
           (IsFunS (RTTypedS taus' taue' sigma'@(RSForAll _ _ _)) tau') = 
   let ismg = isMG efecUnif in
   if tau==tau'
    then do mc1 <- ismg delta sigma' sigma
            taue'' <- indir taue'
            case mc1 of
              Just (c1,ps)
                -> case (taue, taue') of
                    (RTStaticExp e, RTStaticExp e') -> 
                     do let e'' = fill c1 e'
                        yes <- efectiveUnify taue (RTStaticExp e'')
                        if not yes
                         then return Nothing
                         else do yes <- efectiveUnify taus taus'
                                 if not yes 
                                  then return Nothing
                                  else return (Just (EVar h, ps))
                    (_, RTVar _) -> return (Just (EVar h, ps))
                    _ -> return Nothing
              _ -> return Nothing
    else return Nothing

singleentl efecUnif _ delta h
           (IsFunS (RTFixS taus  taue  sigma@(RSForAll _ _ _))  tau) 
           (IsFunS (RTFixS taus' taue' sigma'@(RSForAll _ _ _)) tau') = 
   let ismg = isMG efecUnif in
   if tau==tau'
    then do mc1 <- ismg delta sigma' sigma
            taue'' <- indir taue'
            case mc1 of
              Just (c1,ps)
                -> case (taue, taue') of
                    (RTStaticExp e, RTStaticExp e') -> 
                     do let e'' = fill c1 e'
                        yes <- efectiveUnify taue (RTStaticExp e'')
                        if not yes
                         then return Nothing
                         else do yes <- efectiveUnify taus taus'
                                 if not yes 
                                  then return Nothing
                                  else return (Just (EVar h, ps))
                    (_, RTVar _) -> return (Just (EVar h, ps))
                    _ -> return Nothing
              _ -> return Nothing
    else return Nothing

singleentl efecUnif _ delta 
           h (IsMG sigma@(RSForAll _ _ _)  (RSVar s)) 
             (IsMG sigma'@(RSForAll _ _ _) (RSVar s')) = 
                       let ismg = isMG efecUnif in
                       if s==s'
                        then do mc1 <- ismg delta sigma' sigma
                                mc2 <- ismg delta sigma  sigma'
                                case (mc1, mc2) of
                                  (Just (c1,ps),Just _)
                                    -> if c1==CHole
                                        then return (Just (EVar h, ps))
                                        else return (Just (EComp (EVar h) 
                                                                 (ECxt c1), ps))
                                  _ -> return Nothing
                        else return Nothing

singleentl efecUnif _ delta 
           h (IsMG (RSVar s)  sigma@(RSForAll _ _ _)) 
             (IsMG (RSVar s') sigma'@(RSForAll _ _ _)) = 
                       let ismg = isMG efecUnif in
                       if s==s'
                        then do mc1 <- ismg delta sigma' sigma
                                mc2 <- ismg delta sigma sigma'
                                case (mc1, mc2) of
                                  (Just _,Just (c2,ps)) 
                                    -> if c2==CHole
                                        then return (Just (EVar h, ps))
                                        else return (Just (EComp (ECxt c2) 
                                                                 (EVar h), ps))
                                  _ -> return Nothing
                        else return Nothing

singleentl _ _ _ _ _ _ = return Nothing

---------------------------------------------------------
-- isMoreGeneral relationship: 
--   evidence, if it's true; nothing, otherwise
--   If it is more general, confirm unifications
---------------------------------------------------------
isMoreGeneral :: Preds -> ResidualTypeScheme
                       -> ResidualTypeScheme 
                       -> SpecM (Maybe (ContextTerm,Preds))
isMoreGeneral delta s1 s2 = isMG True delta s1 s2

isMG :: Bool  -> Preds -> ResidualTypeScheme
                       -> ResidualTypeScheme 
                       -> SpecM (Maybe (ContextTerm,Preds))
isMG _ _ sigma1 sigma2 | sigma1==sigma2 = return (Just (CHole, []))
isMG efecUnif delta sigma1@(RSForAll tm  sm  _)
                    sigma2@(RSForAll tm' sm' _) =
   do sigma2_1                        <- addOffset tm sm sigma2
      (delta'' :=> tau'', _,    _)    <- instantiate sigma2_1
      (delta'  :=> tau',  tvs1, svs1) <- instantiate sigma1
{- Alejandro
      traceSM "Instancia 1 ----->"  delta'
      traceSM ":=>" tau'
      traceSM "Instancia 2 ----->"  delta''
      traceSM ":=>" tau''
-}
      success                         <- match tvs1 svs1 tau' tau''
      if not success then do discardUnification; return Nothing else do 
      delta'_1  <- unfoldAll delta'
{- Alejandro
      traceSM "---------------------------" ""
      traceSM "---------------------------" ""
      traceSM "Contexto" delta 
      traceSM "---------------------------" ""
      traceSM "---------------------------" ""
      traceSM "\n" delta''
      traceSM "||-" delta'_1 
      traceSM "-----------------------------" ""
-}
      mvs       <- entls efecUnif (delta''++delta) delta'_1
      -- Alejandro
	  {-
      traceSM "Se intento ver que:" (delta''++delta)
      traceSM "entailea" delta'_1
      traceSM "para lo que falta" mvs
	  -}
      case mvs of
        Nothing      -> do discardUnification; return Nothing
        Just (vs,[]) -> 
                    do let h' = takeEvidence delta''
                       confirmUnification
                       return (Just ( CELam h' (CEApp CHole vs)
                                    , []
                                    ))
        _ -> error "Cannot happen! Have you messed with entls?"

---------------------------------------------------------
-- forceMoreGeneral is like isMoreGeneral, but enriches the global context
---------------------------------------------------------
forceMoreGeneral :: Preds -> ResidualTypeScheme 
                          -> ResidualTypeScheme
                          -> SpecM (Maybe (ContextTerm,Preds))
forceMoreGeneral _ sigma1 sigma2 | sigma1==sigma2 = return (Just (CHole, []))
forceMoreGeneral delta sigma1@(RSForAll tm  sm  _)
                       sigma2@(RSForAll tm' sm' _) =
   do sigma2_1                        <- addOffset tm sm sigma2
      (delta'' :=> tau'', tvs2, svs2) <- instantiate sigma2_1
      (delta'  :=> tau',  tvs1, svs1) <- instantiate sigma1
      success                         <- matchUnif tvs1 svs1 tau' tau''
      if not success then do discardUnification; return Nothing else do
      delta'_1     <- unfoldAll delta'
      let deltaG   = delta''++delta
      --DEBUG
      when ((takeEvidence delta' == [128,129,130,131])
            && (takeEvidence delta'' == [128,129,131,147]))
           (do traceSM "Forcing MG: " ((deltaG, tau''), (delta', tau'))
           )
      --DEBUG
      (vs, deltaKeep') <- delta'_1 `notentlby` deltaG
      confirmUnification
      --sigma1_1 <- generalize (tvs1++tvs2) (svs1++svs2) (delta'_1:=>tau')
      --sigma1_2 <- regularize sigma1_1
      let h' = takeEvidence delta''
      return (Just ( CELam h' (CEApp CHole vs)
                   , deltaKeep'
                   ))

------
-- Renames the evidence variables to fresh ones
------
simplifyPred ftvs fsvs v (Unify tau tau') = 
  case (tau, tau') of
   (RTVar t,  RTVar t')  -> if t `elem` ftvs && t' `elem` ftvs
                             then return Nothing
                             else do yes <- efectiveUnify tau tau'
                                     if yes
                                      then return (Just (evoid, []))
                                      else unifErr tau tau'
   (RTVar t,  _       )  -> if t `elem` ftvs
                             then return Nothing
                             else do yes <- efectiveUnify tau tau'
                                     if yes
                                      then return (Just (evoid, []))
                                      else unifErr tau tau'
   (_      ,  RTVar t')  -> if t' `elem` ftvs
                             then return Nothing
                             else do yes <- efectiveUnify tau tau'
                                     if yes
                                      then return (Just (evoid, []))
                                      else unifErr tau tau'
   (RTInt n,  RTInt n')  -> if n==n' 
                             then return (Just (evoid, []))
                             else do msg <- unifErrMsg tau tau'
                                     return (Just (v,[(v,Fail msg)]))
   (RTBool b, RTBool b') -> if b==b' 
                             then return (Just (evoid, []))
                             else do msg <- unifErrMsg tau tau'
                                     return (Just (v,[(v,Fail msg)]))
   (RTIntD,   RTIntD)    -> return (Just (evoid, []))
   (RTBoolD,  RTBoolD)   -> return (Just (evoid, []))
   (RTFun tau1 tau2,
    RTFun tau1' tau2')   -> do let pred1 = Unify tau1 tau1'
                               let pred2 = Unify tau2 tau2'
                               mpreds1 <- simplifyPred ftvs fsvs v pred1
                               mpreds2 <- simplifyPred ftvs fsvs v pred2
                               case mpreds1 of 
                                Nothing                 -> 
                                 case mpreds2 of
                                  Nothing               -> 
                                    return (Just (v, [(v,pred1),(v,pred2)]))
                                  Just (_,(_,Fail _):_) -> return mpreds2
                                  Just (v2,preds2)      -> 
                                    return (Just (v2, (v,pred1):preds2))
                                Just (_,(_,Fail _):_)   -> return mpreds1
                                Just (v1,preds1)        -> 
                                 case mpreds2 of
                                  Nothing               -> 
                                    return (Just (v1, (v,pred2):preds1))
                                  Just (_,(_,Fail _):_) -> return mpreds2
                                  Just (v2,preds2)      -> 
                                    return (Just (combEv v1 v2, preds1++preds2))
   (RTTuple taus, 
    RTTuple taus')       -> if length taus == length taus'
                             then simplifyTuples ftvs fsvs v taus taus'
                             else do msg <- unifErrMsg tau tau'
                                     return (Just (v,[(v,Fail msg)]))
   (RTTupleS taus,
    RTTupleS taus')      -> if length taus == length taus'
                             then simplifyTuples ftvs fsvs v taus taus'
                             else do msg <- unifErrMsg tau tau'
                                     return (Just (v,[(v,Fail msg)]))
   _                     -> do ftvt  <- freeRTV tau
                               fsvt  <- freeRSV tau
                               ftvt' <- freeRTV tau'
                               fsvt' <- freeRSV tau'
                               yes   <- unify tau tau'
                               if not yes 
                                then do msg <- unifErrMsg tau tau'
                                        return (Just (v,[(v,Fail msg)]))
                                else do
                               if (   ftvt  `intersect` ftvs == [] 
                                   && fsvt  `intersect` fsvs == [])
                                  ||
                                  (   ftvt' `intersect` ftvs == [] 
                                   && fsvt' `intersect` fsvs == [])
                                then do confirmUnification
                                        return (Just (evoid,[]))
                                else do discardUnification
                                        return Nothing

simplifyPred ftvs fsvs v (IsConstrOf datatype tau) = 
  case tau of
   RTConstruc c args -> 
     do source_args <- 
         sourceTypeCheck 
             (do c_sourcetype <- SourceTypeCheckMonad.constructorType c
                 let (args, taud) = splitConstrType c_sourcetype
                 SourceUnification.unify taud datatype
                 args' <- SourceTypeCheckMonad.unfoldVbles True args
                 return args'
             )
        (deltas,yes) <- checkSRfor args source_args
        if not yes
         then do msg <- return "Constructor argument failed check in IsConstrOf"
                 return (Just (v,[(v,Fail msg)]))
         else return (Just (v ,map (\(h,d)->(v,d)) (concat deltas)))
   _            -> return Nothing

simplifyPred ftvs fsvs v (Guarded tau pred) =
  case tau of
   RTBool b -> if b
                then do mpreds <- simplifyPred ftvs fsvs v pred
                        case mpreds of
                          Nothing -> return (Just (v, [(v,pred)]))
                          Just _  -> return mpreds
                else return (Just (evoid, []))
   RTVar t  -> return Nothing 
               {-do (consistentWithT
                   , superfluousForT) <- attemptWith t True  ftvs fsvs v pred
                  if consistentWithT
                   then if superfluousForT
                         then return (Just (evoid, []))
                         else do mpreds <- simplifyPred ftvs fsvs v pred
                                 case mpreds of
                                  Nothing         -> return Nothing
                                  Just (v',preds) -> 
                                   return (Just (v', map (\(v,pred') -> (v,
                                                          Guarded tau pred')
                                                         ) preds))
                   else do set "RTVar" residualtype t (RTBool False)
                           return (Just (evoid, []))
                -}
   RTFixS _ _ _ -> return (Just (v,[(v,Fail "Static fix with booleans may not terminate.")]))
   _            -> return (Just (v,[(v,Fail "Non-boolean guard!")]))

simplifyPred ftvs fsvs v (GuardedNeg tau pred) =
  case tau of
   RTBool b -> if b
                then return (Just (evoid, []))
                else do mpreds <- simplifyPred ftvs fsvs v pred
                        case mpreds of
                          Nothing -> return (Just (v, [(v,pred)]))
                          Just _  -> return mpreds
   RTVar t  -> return Nothing
               {-do (consistentWithF
                   , superfluousForF) <- attemptWith t False ftvs fsvs v pred
                  if consistentWithF
                   then if superfluousForF
                         then return (Just (evoid, []))
                         else do mpreds <- simplifyPred ftvs fsvs v pred
                                 case mpreds of
                                   Nothing         -> return Nothing
                                   Just (v',preds) -> 
                                    return (Just (v', map (\(v,pred') -> (v,
                                                           GuardedNeg tau pred')
                                                          ) preds))
                   else do set "RTVar" residualtype t (RTBool True)
                           return (Just (evoid, []))
               -}
   RTFixS _ _ _ -> return (Just (v,[(v,Fail "Static fix with booleans may not terminate.")]))
   _            -> return (Just (v,[(v,Fail "Non-boolean guard!")]))

simplifyPred ftvs fsvs v (GuardedConstr c tau pred) =
  case tau of
   RTConstruc c' _ -> 
               if c==c'
                then do mpreds <- simplifyPred ftvs fsvs v pred
                        case mpreds of
                          Nothing -> return (Just (v, [(v,pred)]))
                          Just _  -> return mpreds
                else return (Just (evoid, []))
   RTVar t      -> return Nothing
   RTFixS _ _ _ -> let msg = "Static fix with datatypes may not terminate!" 
                    in return (Just (v,[(v,Fail msg)]))
   _            -> let msg = "Non-constructor in constr_guard!" 
                    in return (Just (v,[(v,Fail msg)]))

simplifyPred ftvs fsvs v (IsIf tau taub tau1 tau2) =
  case taub of
   RTBool b -> {-if b 
                then do yes <- efectiveUnify tau tau1
                        if yes 
                         then return (Just (evoid, []))
                         else unifErr tau tau1 
                else do yes <- efectiveUnify tau tau2
                        if yes 
                         then return (Just (evoid, []))
                         else unifErr tau tau2 
               -}error "Cannot happen! It should have be removed by entl..."
   RTVar _  -> if tau1 == tau2
                then do yes <- efectiveUnify tau tau1
                        if yes 
                         then return (Just (evoid, []))
                         else unifErr tau tau1 
                else case tau1 of
                      RTBool b1   -> 
                       case tau2 of
                        RTBool b2 -> if b1 && not b2
                                      then do yes <- efectiveUnify tau taub
                                              if yes
                                               then return (Just (evoid, []))
                                               else unifErr tau taub
                                      else return Nothing
                        _         -> return Nothing
                      _           -> return Nothing
   RTFixS _ _ _ -> return (Just (v,[(v,Fail "Static fix with booleans may not terminate.")]))
   _            -> return (Just (v,[(v,Fail "Non-boolean condition!")]))

--simplifyPred ftvs fsvs v (IsBool (RTBool b)) = return (Just (evoid, []))

simplifyPred _ _ _ _ = return Nothing

------
simplifyTuples _    _    v []         []           = return (Just (evoid,[]))
simplifyTuples ftvs fsvs v (tau:taus) (tau':taus') =
  do mpreds <- simplifyPred ftvs fsvs v (Unify tau tau')
     case mpreds of
       Nothing               -> return Nothing
       Just (_,(_,Fail _):_) -> return mpreds
       Just (v',preds)       -> 
         do mpredss <- simplifyTuples ftvs fsvs v taus taus'
            case mpredss of
             Nothing               -> return Nothing
             Just (_,(_,Fail _):_) -> return mpredss
             Just (v'',predss)     -> return (Just (combEv v' v''
                                                   ,preds++predss))
simplifyTuples _ _ _ _ _ = 
  error "Cannot happen! Have you messed with simplifyPred?"

------
combEv (ETupleS []) v = v
combEv v _            = v
     -- Choice! What happens if the second one is not a evoid?

------
attemptWith t b ftvs fsvs v pred = 
  do newTemporarySubst
     set "RTVar in attempt" tmptype t (RTBool b)
     pred'  <- unfoldAll pred
     mpreds <- simplifyPred ftvs fsvs v pred'
     discardUnification
     traceSM "Attempt result: " mpreds
     case mpreds of
       Just (_,(_,Fail _):_) -> return (False, False)
       Just (_,[])           -> return (True,  True)
       _                     -> return (True,  False)
                                    -- (consistent, superfluous)

---------------------------------------------------------
-- Greatest lower bound of two schemes, if it exists.
---------------------------------------------------------
glb :: [ResTypeVble] -> [ResSchemeVble]
        -> ResidualTypeScheme -> ResidualTypeScheme
        -> SpecM (Maybe ResidualTypeScheme)
glb ftvs fsvs sigma1 sigma2 | sigma1==sigma2 = return (Just sigma1)
glb ftvs fsvs sigma1@(RSForAll _ _ _) sigma2@(RSForAll _ _ _) =
   do (delta'' :=> tau'', tvs2, svs2) <- instantiate sigma2
      (delta'  :=> tau',  tvs1, svs1) <- instantiate sigma1
      success      <- efectiveUnify tau' tau''   -- Why effective??
      if not success then return Nothing else do 
      delta_1      <- unfoldAll (delta'++delta'')
      tau'_1       <- unfoldAll tau'
      (_,delta_2)  <- simplify ftvs fsvs delta_1
      (sigma,_,_)  <- genThese emptyAssgn (tvs1++tvs2) (svs1++svs2) 
                               (delta_2:=>tau'_1)
      sigma_1      <- regularize sigma
      return (Just sigma_1)

------
-- Calculates de glb of a list of schemes
------
glbs _    _    [sigma]        = return (Just sigma)
glbs ftvs fsvs (sigma:sigmas) = do mplb <- glbs ftvs fsvs sigmas
                                   case mplb of
                                    Nothing  -> return Nothing
                                    Just plb -> glb ftvs fsvs sigma plb

