module ResidualDataManagement where

import Error

import BasicData
import ResidualLang

------------------------------
-- Residual data management --
------------------------------
type RDataManager = ([RDataDef],[(DataName,ResType)])

initialRDataMngr = ([],[])

-------------------------
-- buildConstructorType
buildConstructorType reps_ts reps_ss dataName args =
  let datatype = RTData dataName (map snd reps_ts) (map snd reps_ss)
      args' = replaceIn (reps_ts, reps_ss) args
   in buildConstrType args' datatype

-------------------------
--mngrDataNamed
mngrDataNamed (ddefs,fws) d = 
  let mr = mngrDataNamed' ddefs d
   in case mr of
        Success ddef -> Success (Left ddef)
        Error msg    -> case lookup d fws of
                         Nothing -> Error msg
                         Just td -> Success (Right td)

mngrDataNamed' :: [RDataDef] -> DataName -> Error RDataDef
mngrDataNamed' [] d = Error ("Cannot happen! Have you messed with datatypes?"
                            ++ " (data " ++ d ++ " not found)")
mngrDataNamed' (ddef:ddefs) d =
     if d == rdatadefName ddef 
      then Success ddef
      else mngrDataNamed' ddefs d

-------------------------
--mngrDataOf
mngrDataOf (ddefs,_) = mngrDataOf' ddefs

mngrDataOf' :: [RDataDef] -> ConstructorName 
              -> Error (RDataDef, [ResType])
mngrDataOf' [] c = Error ("Cannot happen! Have you messed with datatypes?"
                            ++ " (constructor " ++ c ++ " not found)")
mngrDataOf' (ddef@(RDataDef _ _ _ _ cs):ddefs) cname =
  case (constrArgsOf cname cs) of
    Nothing   -> mngrDataOf' ddefs cname 
    Just args -> Success (ddef, args)

-------------------------
--mngrDataOf
mngrAddData (ddefs,fws) d = (d:ddefs,fws)

mngrAddForwardData (ddefs,fws) fd = (ddefs,fd:fws)

--------------------------
--------------------------
--------------------------
class ReplaceRT t where
  replaceIn :: ([(ResType,ResType)],[(ResTypeScheme,ResTypeScheme)]) -> t -> t
  replaceIn _ t = t

--------------------------
instance ReplaceRT ResType where
  replaceIn (reps_ts,_) t@(RTVar _) = 
     case lookup t reps_ts of
       Nothing  -> error ("Cannot happen! Free variable "
                         ++"in constructor argument after check...")
       Just nrt -> nrt
  replaceIn reps (RTTuple ts)       = RTTuple      (reps `replaceIn` ts)
  replaceIn reps (RTTupleS ts)      = RTTupleS     (reps `replaceIn` ts)
  replaceIn reps (RTStaticExp e')   = RTStaticExp  (reps `replaceIn` e')
  replaceIn reps (RTFixS t1 t2 s)   = RTFixS       (reps `replaceIn` t1)
                                                   (reps `replaceIn` t2)
                                                   (reps `replaceIn` s)
  replaceIn reps (RTTypedS t1 t2 s) = RTTypedS     (reps `replaceIn` t1)
                                                   (reps `replaceIn` t2)
                                                   (reps `replaceIn` s)
  replaceIn reps (RTFun t1 t2)      = RTFun        (reps `replaceIn` t1)
                                                   (reps `replaceIn` t2)
  replaceIn reps (RTPoly t)         = RTPoly       (reps `replaceIn` t)
  replaceIn reps (RTData n ts ss)   = RTData n     (reps `replaceIn` ts)
                                                   (reps `replaceIn` ss)
  replaceIn reps (RTConstruc c t)   = RTConstruc c (reps `replaceIn` t)
  replaceIn _    t                  = t

--------------------------
instance ReplaceRT ResTypeScheme where
  replaceIn (_,reps_ss) s@(RSVar _) = 
     case lookup s reps_ss of
       Nothing  -> error ("Cannot happen! Free variable "
                         ++"in constructor argument after check...")
       Just nrs -> nrs
  replaceIn reps (RSTupleS ts)      = RSTupleS      (reps `replaceIn` ts)
  replaceIn reps (RSForAll t s qt)  = RSForAll t s  (reps `replaceIn` qt)
  replaceIn _    s                  = s

--------------------------
instance ReplaceRT QualType where
  replaceIn reps (ps:=>t) = (reps `replaceIn` ps) :=> (reps `replaceIn` t)

instance ReplaceRT a => ReplaceRT [a] where
  replaceIn reps = map (replaceIn reps)

instance (ReplaceRT a, ReplaceRT b) => ReplaceRT (a,b) where
  replaceIn reps (a,b) = (reps `replaceIn` a, reps `replaceIn` b)

--------------------------
instance ReplaceRT Pred where
  replaceIn reps (IsMG s1 s2)            = IsMG            (replaceIn reps s1) 
                                                           (replaceIn reps s2)
  replaceIn reps (IsMGmorph t s1 s2)     = IsMGmorph       (replaceIn reps t) 
                                                           (replaceIn reps s1) 
                                                           (replaceIn reps s2)
  replaceIn reps (IsInt t)               = IsInt           (replaceIn reps t) 
  replaceIn reps (IsBool t)              = IsBool          (replaceIn reps t) 
  replaceIn reps (IsIf t tb t1 t2)       = IsIf            (replaceIn reps t) 
                                                           (replaceIn reps tb)
                                                           (replaceIn reps t1)
                                                           (replaceIn reps t2)
  replaceIn reps (IsCase t te cts)       = IsCase          (replaceIn reps t) 
                                                           (replaceIn reps te)
                                                           (replaceIn reps cts)
  replaceIn reps (IsFunS t1 t2)          = IsFunS          (replaceIn reps t1)
                                                           (replaceIn reps t2)
  replaceIn reps (IsFixS t1 t2)          = IsFixS          (replaceIn reps t1)
                                                           (replaceIn reps t2)
  replaceIn reps (IsArithOp aop t t1 t2) = IsArithOp aop   (replaceIn reps t)
                                                           (replaceIn reps t1)
                                                           (replaceIn reps t2)
  replaceIn reps (IsBoolOp  bop t t1 t2) = IsBoolOp bop    (replaceIn reps t)
                                                           (replaceIn reps t1)
                                                           (replaceIn reps t2)
  replaceIn reps (IsRelOp   rop t t1 t2) = IsRelOp rop     (replaceIn reps t)
                                                           (replaceIn reps t1)
                                                           (replaceIn reps t2)
  replaceIn reps (IsConstrOf sdt t)      = IsConstrOf sdt  (replaceIn reps t) 
  replaceIn reps (Guarded t d)           = Guarded         (replaceIn reps t) 
                                                           (replaceIn reps d)
  replaceIn reps (GuardedNeg t d)        = GuardedNeg      (replaceIn reps t) 
                                                           (replaceIn reps d)
  replaceIn reps (GuardedConstr c t d)   = GuardedConstr c (replaceIn reps t) 
                                                           (replaceIn reps d)
  replaceIn reps (Unify t1 t2)           = Unify           (replaceIn reps t1)
                                                           (replaceIn reps t2)
  replaceIn _    d                       = d

--------------------------
instance ReplaceRT ResTerm where
  replaceIn reps (RApp e1 e2)         = RApp         (replaceIn reps e1)
                                                     (replaceIn reps e2)
  replaceIn reps (RLam x e)           = RLam x       (replaceIn reps e)
  replaceIn reps (RFix e)             = RFix         (replaceIn reps e)
  replaceIn reps (RLet xes e)         = RLet         (replaceIn reps xes)
                                                     (replaceIn reps e)
  replaceIn reps (RArithOp aop e1 e2) = RArithOp aop (replaceIn reps e1)
                                                     (replaceIn reps e2)
  replaceIn reps (RIf e e1 e2)        = RIf          (replaceIn reps e)
                                                     (replaceIn reps e1)
                                                     (replaceIn reps e2)
  replaceIn reps (RBoolOp  bop e1 e2) = RBoolOp bop  (replaceIn reps e1)
                                                     (replaceIn reps e2)
  replaceIn reps (RRelOp rop e1 e2)   = RRelOp rop   (replaceIn reps e1)
                                                     (replaceIn reps e2)
  replaceIn reps (RTupleS es)         = RTupleS      (replaceIn reps es)
  replaceIn reps (RProjS i n e)       = RProjS i n   (replaceIn reps e)
  replaceIn reps (RTuple es)          = RTuple       (replaceIn reps es)
  replaceIn reps (RProj i n e)        = RProj i n    (replaceIn reps e)
  replaceIn reps (RData dds e)        = RData dds    (replaceIn reps e)
  replaceIn reps (RConstruc c es)     = RConstruc c  (replaceIn reps es)
  replaceIn reps (RCase e cs)         = RCase        (replaceIn reps e)
                                                     (replaceIn reps cs)
  replaceIn reps (REFill v e)         = REFill       (replaceIn reps v)
                                                     (replaceIn reps e)
  replaceIn reps (RELam h e)          = RELam h      (replaceIn reps e)
  replaceIn reps (REApp e v)          = REApp        (replaceIn reps e)
                                                     (replaceIn reps v)
  replaceIn reps (RELet x e' e)       = RELet x      (replaceIn reps e')
                                                     (replaceIn reps e)
  replaceIn reps (REIf v e1 e2)       = REIf         (replaceIn reps v)
                                                     (replaceIn reps e1)
                                                     (replaceIn reps e2)
  replaceIn reps (RECase v pes)       = RECase       (replaceIn reps v)
                                                     (replaceIn reps pes)
  replaceIn reps (REAppS e1 e2)       = REAppS       (replaceIn reps e1)
                                                     (replaceIn reps e2)
  replaceIn _    e                    = e

instance ReplaceRT EvTerm where
  replaceIn reps (ECxt c)       = ECxt       (replaceIn reps c)
  replaceIn reps (EComp v1 v2)  = EComp      (replaceIn reps v1)
                                             (replaceIn reps v2)
  replaceIn reps (EStaticExp e) = EStaticExp (replaceIn reps e)
  replaceIn reps (EAppS e1 e2)  = EAppS      (replaceIn reps e1)
                                             (replaceIn reps e2)
  replaceIn reps (ETupleS v)    = ETupleS    (replaceIn reps v)
  replaceIn reps (EProjS i n v) = EProjS i n (replaceIn reps v)
  replaceIn _    v              = v

instance ReplaceRT CxtTerm where
  replaceIn reps (CELet x c e) = CELet x     (replaceIn reps c)
                                             (replaceIn reps e)
  replaceIn reps (CELam v c)   = CELam       (replaceIn reps v)
                                             (replaceIn reps c)
  replaceIn reps (CEApp c v)   = CEApp       (replaceIn reps c)
                                             (replaceIn reps v)
  replaceIn _    c             = c


instance ReplaceRT Char

instance ReplaceRT Int

