module ResidualLang where

import List(elemIndex,(\\), union)

import MyPrelude(accuml,seteq)

import BasicData
import SourceLang(SourceType,ArithOp,BoolOp,RelOp,STBase(..))

-- ------------------------------------
-- Residual Terms
-- ------------------------------------
type TypedResidualTerm = (ResidualTerm, ResidualType)

type ResTermVble  = String
type EvVble       = Int
type ResidualTerm = ResTerm
data ResTerm      = 
     RVar     ResTermVble                 
   -- Functions
   | RApp     ResTerm ResTerm  
   | RLam     ResTermVble ResTerm         
   | RFix     ResTerm         
   | RLet     [ (ResTermVble,ResTerm) ] 
              ResTerm
   -- Basic types: Int
   | RInt     Int
   | RArithOp ArithOp ResTerm ResTerm
   -- Basic types: Bool
   | RBool    Bool
   | RIf      ResTerm ResTerm ResTerm
   | RBoolOp  BoolOp ResTerm ResTerm
   | RRelOp   RelOp ResTerm ResTerm
   -- Tuples
   | RTupleS  [ ResTerm ]
   | RProjS   Int Int ResTerm 
   | RTuple   [ ResTerm ]
   | RProj    Int Int ResTerm 
   -- Data                            
   | RData [RDataDef] ResTerm          
   | RConstruc ConstructorName [ResTerm]
   | RCase ResTerm [(ResTerm,ResTerm)] 
   --------------------------------------
   -------- Begin Alejandro -------------
   --------------------------------------
   | RDConstruc ConstructorName ResTerm [ResTerm]   -- C_r [ResTerm]
   -- | RDConstruc ConstructorName EvTerm [ResTerm]   -- C_r [ResTerm]
   | RProtoCase ResTerm ResTerm [(ResTerm, ResTerm)] 
   -- | RProtoCase ResTerm EvTerm [(ResTerm, ResTerm)] 
   --------------------------------------
   ---------- End Alejandro -------------
   --------------------------------------

   -- Evidence Stuff
   | REVar    EvVble                      
   | REFill   EvTerm ResTerm 
   | RELam    [ EvVble ] ResTerm
   | REApp    ResTerm [ EvTerm ] 
   | RELet    ResTermVble ResTerm ResTerm    -- let_v x=e in e
   | REIf     EvTerm ResTerm ResTerm         -- if_v v then e else e
   | RECase   EvTerm [(ConstructorName,ResTerm)] -- case v of c_i -> e_i esac
   | REAppS   ResTerm ResTerm                -- e @_v e
   | REConstrs [ConstructorName]
   deriving (Eq, Read, Show)

evoid  = ETupleS  []
rvoid  = RTupleS  []
rtvoid = RTTupleS []

type EvidenceTerm = EvTerm
data EvTerm       = 
     EVar  EvVble                    
   | EInt  Int
   | EBool Bool
   | EConstr ConstructorName
   | ECxt  CxtTerm                
   | EComp EvTerm EvTerm             -- (v o v') = v[v'[]::t]
   | EStaticExp ResTerm                      -- Cast e into evidence
   | EAppS ResTerm ResTerm                   -- e @_v e
   | ETupleS [ EvTerm ]
   | EProjS Int Int EvTerm
   | EConstrs [ConstructorName]
   deriving (Eq, Read, Show)

type ContextTerm = CxtTerm
data CxtTerm     =
   CHole                         
 | CELet ResTermVble CxtTerm ResTerm 
 | CELam [ EvVble ] CxtTerm
 | CEApp CxtTerm [ EvTerm ]  
 deriving (Eq, Read, Show)

-- ------------------------------------
-- Residual Types
-- ------------------------------------
type ResTypeVble      = Int
type ResGenTypeVble   = Int
type ResDeBruijnLevel = Int
type ResidualType     = ResType
data ResType          = 
     RTVar ResTypeVble 
   | RTGen ResDeBruijnLevel ResGenTypeVble  -- Quantified variables 
        -- deBruijn index,  id
   | RTInt       Int
   | RTIntD  
   | RTBool      Bool
   | RTBoolD  
   | RTTupleS    [ ResType ]
   | RTTuple     [ ResType ]
   | RTInvertCxt CxtTerm ResTerm            -- c^(-1)[e'] = e'' ie. c[e'']=e'
                                            --  Patch for IsFunS solving...
   | RTStaticExp ResTerm                    -- Cast e' into a type
   | RTFixS      ResType                    -- type of the tuple of free vars??
                 ResType ResTypeScheme      -- e' :: sigma
   | RTTypedS    ResType                    -- type of the tuple of free vars
                 ResType ResTypeScheme      -- e' :: sigma
   | RTFun       ResType ResType
   | RTPoly      ResTypeScheme
   | RTData      DataName [ResType] [ResTypeScheme] 
   | RTDData 	 ResType ResType
   | RTConstruc  ConstructorName [ResType]          
   deriving (Eq, Read, Show)

isRTVar (RTVar _) = True
isRTVar _ = False

type ResSchemeVble      = Int
type ResGenSchemeVble   = Int
type ResidualTypeScheme = ResTypeScheme
data ResTypeScheme      = 
     RSVar ResSchemeVble
   | RSGen ResDeBruijnLevel ResGenSchemeVble -- Quantified variables 
        -- deBruijn index,  id
   | RSForAll ResGenTypeVble ResGenSchemeVble  QualType 
           -- maxTypes,      maxSchemes
   | RSTupleS [ ResType ]
   deriving (Eq, Read, Show)

type QualifiedType = QualType
data QualType      = Preds :=> ResType
                   deriving (Eq, Read, Show)

data Pred = IsMG       ResTypeScheme ResTypeScheme
          | IsMGmorph  ResType ResTypeScheme ResTypeScheme
          | IsInt      ResType
          | IsBool     ResType
          | IsConstrOf SourceType ResType  
            -- eg. IsConstrOf (Maybe t) (Just t')
          | IsIf       ResType ResType ResType ResType
          | IsCase     ResType ResType [(ResType,ResType)]
          | IsFunS     ResType ResType
          | IsFixS     ResType ResType
          | IsArithOp  ArithOp ResType ResType ResType 
          | IsBoolOp   BoolOp  ResType ResType ResType 
          | IsRelOp    RelOp   ResType ResType ResType 
          | Guarded    ResType Pred
          | GuardedNeg ResType Pred
          | GuardedConstr ConstructorName ResType Pred
          | Unify      ResType ResType
          | Fail       String
          --- Begin Alejandro ---
          | GuardedDinData ConstructorName ResType Pred -- (C in D?Pred)
          | HasC ResType ConstructorName ResType -- HasC t1 C t2
          | IndexOf ResType -- IndexOf t
          | UnifyData ResType ConstructorName ResType -- t[cname] ~ t'
          | ConstrsOf ResType -- ConstrsOf tau
          --- End Alejandro ---
          deriving (Eq, Read, Show)

assertIs SBInt  = IsInt
assertIs SBBool = IsBool

type Preds = [ (EvidenceTerm, Pred) ]

-- ------------------------------------
--  Auxiliaries
-- ------------------------------------
ithProj :: Int -> Int -> EvidenceTerm
ithProj n _ | n<=0 = error "Cannot produce projections for an empty tuple"
ithProj n i = EProjS i n (ECxt CHole)

rTuple [t] = t
rTuple ts  = RTTupleS ts

eTuple [e] = e
eTuple es  = RTupleS es

takeEvidence :: Preds -> [EvVble]
takeEvidence = map (unEVar . fst)

unEVar (EVar h) = h

class Scheme t where
  scheme :: t -> ResidualTypeScheme

instance Scheme ResType where
  scheme tau' = RSForAll 0 0 ([] :=> tau')

instance Scheme QualType where
  scheme qt   = RSForAll 0 0 qt

evidence2residual :: EvidenceTerm -> ResidualTerm
evidence2residual (EVar  h)      = REVar h
evidence2residual (EInt  n)      = RInt  n
evidence2residual (EBool b)      = RBool b
evidence2residual (EStaticExp e) = e
evidence2residual (EAppS e e')   = REAppS e e'
evidence2residual (ETupleS vs)   = RTupleS (map evidence2residual vs)
evidence2residual (EProjS i n v) = RProjS i n (evidence2residual v)
evidence2residual (EConstrs names) = REConstrs names
evidence2residual (ECxt _)       = error cxtMsg
evidence2residual (EComp _ _)    = error cxtMsg
evidence2residual (EConstr _)    = error ("Constructors cannot be transformed"
                                          ++" from evidence to the residual"
                                          ++" language.")
evidence2residual v = error ("Extend evidence2residual for " ++ show v)

cxtMsg = "Contexts cannot be expressed in the residual language"
          ++ " (without being filled)"

-- replaceEv h v delta t = ((/\h.[]::t)::(delta:=>t)){{v}}
--  so, h::delta, and t is the type of the hole
replaceEv :: [EvVble] -> [EvTerm] -> CxtTerm
replaceEv [] [] = CHole
replaceEv h  v  = CEApp (CELam h CHole) v

-- ------------------------------------
-- Eta contraction for residual terms!
-- ------------------------------------
class EtaStrippable t where
  stripREtas :: t -> t
  stripREtas t = t

instance EtaStrippable ResTerm where
  stripREtas (RApp e1 e2)       = RApp (stripREtas e1) (stripREtas e2)
  stripREtas e@(RLam _ _)       = let (xs, e') = stripREta [] e
                                  in foldr RLam (stripREtas e') xs
  stripREtas (RFix e)           = RFix (stripREtas e)
  stripREtas (RLet ds e)        = RLet (stripREtas ds) (stripREtas e)
  stripREtas (RArithOp o e1 e2) = RArithOp o (stripREtas e1) (stripREtas e2)
  stripREtas (RIf e e1 e2)      = RIf (stripREtas e) (stripREtas e1) 
                                                     (stripREtas e2)
  stripREtas (RBoolOp o e1 e2)  = RBoolOp o (stripREtas e1) (stripREtas e2)
  stripREtas (RRelOp o e1 e2)   = RRelOp o (stripREtas e1) (stripREtas e2)
  stripREtas (RTupleS es)       = RTupleS (stripREtas es)
  stripREtas (RProjS i n e)     = RProjS i n (stripREtas e)
  stripREtas (RTuple es)        = RTuple (stripREtas es)
  stripREtas (RProj i n e)      = RProj i n (stripREtas e)
  stripREtas (RData ddefs e)    = RData ddefs (stripREtas e)
  stripREtas (RConstruc c es)   = RConstruc c (stripREtas es)
  stripREtas (RCase e es)       = RCase (stripREtas e) (stripREtas es)
  stripREtas (REFill v e)       = REFill (stripREtas v) (stripREtas e)
  stripREtas (RELam h e)        = RELam h (stripREtas e)
  stripREtas (REApp e v)        = REApp (stripREtas e) (stripREtas v)
  stripREtas (RELet x e' e)     = RELet x (stripREtas e') (stripREtas e)
  stripREtas (REIf v e1 e2)     = REIf (stripREtas v) (stripREtas e1)
                                                      (stripREtas e2)
  stripREtas (RECase v pes)     = RECase (stripREtas v) (stripREtas pes)
  stripREtas (REAppS e e')      = REAppS (stripREtas e) (stripREtas e')
  stripREtas e                  = e
 
stripREta xs (RLam x e) = stripREta (x:xs) e
stripREta xs e@(RConstruc c args) =
  let n = length args
      m = length xs
      (xs',etas) = splitAt (m-n) xs
   in if map RVar (reverse etas) == args
       then (reverse xs', RConstruc c [])
       else (reverse xs, e)

--------------------------------
--- Begin Alejandro ------------
--------------------------------
stripREta xs e@(RDConstruc c ev args) =
  let n = length args
      m = length xs
      (xs',etas) = splitAt (m-n) xs
   in if map RVar (reverse etas) == args
       then (reverse xs', RDConstruc c ev [])
       else (reverse xs, e)
--------------------------------
---- End  Alejandro ------------
--------------------------------

stripREta xs e = (reverse xs, e)

instance EtaStrippable EvTerm where
  stripREtas (ECxt c)         = ECxt       (stripREtas c)
  stripREtas (EComp v v')     = EComp      (stripREtas v) (stripREtas v')
  stripREtas (EStaticExp e)   = EStaticExp (stripREtas e)
  stripREtas (EAppS e e')     = EAppS      (stripREtas e) (stripREtas e')
  stripREtas (ETupleS v)      = ETupleS    (stripREtas v)
  stripREtas (EProjS i n v)   = EProjS i n (stripREtas v)
  stripREtas v                = v
  
instance EtaStrippable CxtTerm where
  stripREtas (CELet x c e)    = CELet x (stripREtas c) (stripREtas e)
  stripREtas (CELam h c)      = CELam h (stripREtas c)
  stripREtas (CEApp c v)      = CEApp   (stripREtas c) (stripREtas v)
  stripREtas c                = c
  
instance EtaStrippable Char
instance EtaStrippable Int
instance EtaStrippable ResType

instance EtaStrippable a => EtaStrippable [a] where
  stripREtas [] = []
  stripREtas (e:es) = stripREtas e : stripREtas es
                         
instance (EtaStrippable a, EtaStrippable b) => EtaStrippable (a,b) where
  stripREtas (a,b) = (stripREtas a, stripREtas b)

-- ------------------------------------
-- Free evidence variables for residual terms!
-- ------------------------------------
class FreeEvidence t where
  freeHVs :: t -> [EvVble]
  freeHVs t = []

instance FreeEvidence ResTerm where
  freeHVs (RApp e1 e2)       = freeHVs e1 `union` freeHVs e2
  freeHVs (RLam _ e)         = freeHVs e
  freeHVs (RFix e)           = freeHVs e
  freeHVs (RLet ds e)        = freeHVs ds `union` freeHVs e
  freeHVs (RArithOp _ e1 e2) = freeHVs e1 `union` freeHVs e2
  freeHVs (RIf e e1 e2)      = freeHVs e  `union` freeHVs e1
                                          `union` freeHVs e2
  freeHVs (RBoolOp _ e1 e2)  = freeHVs e1 `union` freeHVs e2
  freeHVs (RRelOp _ e1 e2)   = freeHVs e1 `union` freeHVs e2
  freeHVs (RTupleS es)       = freeHVs es
  freeHVs (RProjS _ _ e)     = freeHVs e
  freeHVs (RTuple es)        = freeHVs es
  freeHVs (RProj _ _ e)      = freeHVs e
  freeHVs (RData _ e)        = freeHVs e
  freeHVs (RConstruc _ es)   = freeHVs es
  freeHVs (RCase e es)       = freeHVs e  `union` freeHVs es
  freeHVs (REFill v e)       = freeHVs v  `union` freeHVs e
  freeHVs (RELam h e)        = freeHVs e \\ h
  freeHVs (REApp e v)        = freeHVs e  `union` freeHVs v
  freeHVs (RELet x e' e)     = freeHVs e' `union` freeHVs e
  freeHVs (REIf v e1 e2)     = freeHVs v  `union` freeHVs e1 
                                          `union` freeHVs e2
  freeHVs (RECase v pes)     = freeHVs v  `union` freeHVs pes
  freeHVs (REAppS e e')      = freeHVs e  `union` freeHVs e'
  freeHVs e                  = []
 
instance FreeEvidence EvTerm where
  freeHVs (EVar h)         = [h]
  freeHVs (ECxt c)         = freeHVs c
  freeHVs (EComp v v')     = freeHVs v `union` freeHVs v'
  freeHVs (EStaticExp e)   = freeHVs e
  freeHVs (EAppS e e')     = freeHVs e `union` freeHVs e'
  freeHVs (ETupleS v)      = freeHVs v
  freeHVs (EProjS i n v)   = freeHVs v
  freeHVs v                = []
  
instance FreeEvidence CxtTerm where
  freeHVs (CELet x c e)    = freeHVs c `union` freeHVs e
  freeHVs (CELam h c)      = freeHVs c \\ h
  freeHVs (CEApp c v)      = freeHVs c `union` freeHVs v
  freeHVs c                = []
  
instance FreeEvidence Char
instance FreeEvidence Int
instance FreeEvidence ResType

instance FreeEvidence a => FreeEvidence [a] where
  freeHVs [] = []
  freeHVs (e:es) = freeHVs e `union` freeHVs es
                         
instance (FreeEvidence a, FreeEvidence b) => FreeEvidence (a,b) where
  freeHVs (a,b) = freeHVs a `union` freeHVs b

--------------------
-- Residual datatypes
----------------------
data RDataDef = RDataDef DataName [ResType] [ResTypeScheme] 
                         Preds [Constructor ResType]  
               deriving (Eq, Read, Show)                  

rdatadefName (RDataDef name _ _ _ _) = name

rdataName (RTData name _ _) = name

buildConstrType args datatype = foldr RTFun datatype args

