module ResidualLangFVs where

import List(union)
import Monad(liftM, liftM2)
import MyPrelude(unite)

import ResidualLang
import SpecMonad
import ResidualGeneralizable

-- ---------------------------------------------
-- Free type and scheme variables
-- ---------------------------------------------
freeRTVs g tau = do frvg <- freeRTV g
                    frvt <- freeRTV tau
                    fsvg <- freeRSV g
                    fsvt <- freeRSV tau
                    return (frvg `union` frvt, fsvg `union` fsvt)

freeRTV  t = do t' <- unfoldAll t; typevars t'
freeRSV  t = do t' <- unfoldAll t; schemevars t'
freeRGTV t = do t' <- unfoldAll t; resetGeneric; generictypevars t'
freeRGSV t = do t' <- unfoldAll t; resetGeneric; genericschemevars t'

typevars :: Generalizable t => t -> SpecM [ResTypeVble]
typevars t = do (_,tvs,_) <- gr (\tv -> return (RTVar tv,[tv],[]))
                                (\ot -> return (ot      ,[]  ,[]))
                                (\sv -> return (RSVar sv,[]  ,[]))
                                (\os -> return (os      ,[]  ,[]))
                                t
                return tvs

schemevars :: Generalizable t => t -> SpecM [ResSchemeVble]
schemevars t = do (_,_,svs) <- gr (\tv -> return (RTVar tv,[],[]))
                                  (\ot -> return (ot      ,[],[]))
                                  (\sv -> return (RSVar sv,[],[sv]))
                                  (\os -> return (os      ,[],[]))
                                  t
                  return svs

generictypevars t =
            do (_,tgvs,_) <- gr (\tv -> return (RTVar tv, [], []))
                                (\ot@(RTGen l g) ->
                                     do cl <- scopeLevel
                                        if cl==l then return (ot,[g],[])
                                                 else return (ot,[ ],[])
                                )
                                (\sv -> return (RSVar sv,[],[]))
                                (\os -> return (os,[],[]))
                                t
               return tgvs

genericschemevars t =
            do (_,_,sgvs) <- gr (\tv -> return (RTVar tv,[],[]))
                                (\ot -> return (ot,[],[]))
                                (\sv -> return (RSVar sv,[],[]))
                                (\os@(RSGen l g) ->
                                     do cl <- scopeLevel
                                        if cl==l then return (os,[],[g])
                                                 else return (os,[],[ ])
                                )
                                t
               return sgvs


---------------------------------------------------------
-- Classifying free variables in residual predicates
--  It assumes that t is completely unfolded
class (Unfoldable t,Generalizable t) => FRV t where
  clFreeRTV :: Bool -> t
                    -> SpecM ([ResTypeVble]  ,[ResTypeVble]  ,[ResTypeVble])
  clFreeRSV :: Bool -> t
                    -> SpecM ([ResSchemeVble],[ResSchemeVble],[ResSchemeVble])
                           -- Go             , Wait          , Other

  clFreeRTV _ t = toWait (freeRTV t)
  clFreeRSV _ t = toWait (freeRSV t)

toGo    m = do fvs <- m
               return (fvs,[],[])

toWait  m = do fvs <- m
               return ([],fvs,[])

toOther m = do fvs <- m
               return ([],[],fvs)

instance FRV EvTerm 

---------------------------------------------------------
instance FRV ResType where
  clFreeRTV top (RTVar t)              = toWait (return [t])
  clFreeRTV top (RTTupleS taus)        = clFreeRTV top taus
  clFreeRTV top (RTTuple  taus)        = clFreeRTV top taus
  clFreeRTV top (RTFixS   _ tau sigma) = clFreeRTVFor2 top tau sigma
  clFreeRTV top (RTTypedS _ tau sigma) = clFreeRTVFor2 top tau sigma
  clFreeRTV top (RTFun tau1 tau2)      = clFreeRTVFor2 top tau1 tau2
  clFreeRTV top (RTPoly sigma)         = clFreeRTV top sigma
  --clFreeRTV top (RTData -- Should not appear!!
  ---------------------------
  ---- Begin Alejandro ------
  ---------------------------
  clFreeRTV top (RTDData td ts) = clFreeRTVFor2 top td ts
  ---------------------------
  ------ End Alejandro ------
  ---------------------------

  clFreeRTV top (RTConstruc _ taus)    = clFreeRTV top taus
  clFreeRTV _   _ = return ([],[],[])
          -- Invert, StaticExp, Int, Bool, Gen, etc.

---------------------------------------------------------
instance FRV ResTypeScheme where
  clFreeRTV top (RSForAll _ _ qt) = clFreeRTV top qt
  clFreeRTV top (RSTupleS taus)   = clFreeRTV top taus
  clFreeRTV _   _                 = return ([],[],[])

---------------------------------------------------------
instance FRV QualType where
  clFreeRTV _   (preds:=>tau) =
    do pfvs <- clFreeRTV False preds
       tfvs <- clFreeRTV False tau
       return (pfvs `trunion` tfvs)

---------------------------------------------------------
instance FRV a => FRV [a] where
  clFreeRTV _   []     = return ([],[],[])
  clFreeRTV top (x:xs) = do fv1 <- clFreeRTV top x
                            fv2 <- clFreeRTV top xs
                            return (fv1 `trunion` fv2)
  clFreeRSV _   []     = return ([],[],[])
  clFreeRSV top (x:xs) = do fv1 <- clFreeRSV top x
                            fv2 <- clFreeRSV top xs
                            return (fv1 `trunion` fv2)

---------------------------------------------------------
instance (FRV a, FRV b) => FRV (a,b) where
  clFreeRTV top (a,b) = do fv1 <- clFreeRTV top a
                           fv2 <- clFreeRTV top b
                           return (fv1 `trunion` fv2)
  clFreeRSV top (a,b) = do fv1 <- clFreeRSV top a
                           fv2 <- clFreeRSV top b
                           return (fv1 `trunion` fv2)

---------------------------------------------------------
instance FRV Pred where
  clFreeRTV top (IsMG s1 s2)            = clFreeRTVFor2 top s1 s2
  clFreeRTV top (IsMGmorph tau s1 s2)   = clFreeRTVFor2 top tau (IsMG s1 s2)
  clFreeRTV top (IsInt tau)                  = clFreeVar top tau
  clFreeRTV top (IsBool tau)                 = clFreeVar top tau
  clFreeRTV top (IsConstrOf _ tau)           = clFreeRTV top tau
  clFreeRTV top (IsIf tau taub tau1 tau2)    = clFreePolyOp top tau taub 
                                                                [tau1,tau2]
  clFreeRTV top (IsCase tau taue cs)         = clFreePolyOp top tau taue cs
  clFreeRTV top (IsFunS tau1 tau2)           = clFreeFun False top tau1 tau2
  clFreeRTV top (IsFixS tau1 tau2)           = clFreeFun True  top tau1 tau2
  clFreeRTV top (IsArithOp  _ tau tau1 tau2) = clFreeOp tau tau1 tau2
  clFreeRTV top (IsBoolOp   _ tau tau1 tau2) = clFreeOp tau tau1 tau2
  clFreeRTV top (IsRelOp    _ tau tau1 tau2) = clFreeOp tau tau1 tau2
  clFreeRTV top (Guarded    taub pred)       = clFreeGuard top taub pred
  clFreeRTV top (GuardedNeg taub pred)       = clFreeGuard top taub pred
  clFreeRTV top (GuardedConstr _ taub pred)  = clFreeGuard top taub pred
  ----------------------------------
  --------- Begin Alejandro --------
  ----------------------------------
  clFreeRTV top (GuardedDinData _ taud pred) = clFreeGuardDinData 
                                                    top taud pred

  clFreeRTV top (HasC tau1 _ tau2) = clFreeHasC top tau1 tau2
  clFreeRTV top (ConstrsOf tau) = clFreeConstrsOf top tau

  clFreeRTV top (UnifyData t cname tau) = clFreeUnifyData top t tau
  clFreeRTV top (IndexOf t) = clFreeVar top t 
  ----------------------------------
  ---------- End Alejandro ---------
  ----------------------------------
  clFreeRTV _   pred                         = toWait (freeRTV pred)
            -- Unify, Fail

  -----------
  clFreeRSV top (IsMG          sigma1 sigma2) = clFreeRSV_MG top sigma1 sigma2
  clFreeRSV top (IsMGmorph tau sigma1 sigma2) = do sfvs <- clFreeRSV_MG top 
                                                                 sigma1 sigma2
                                                   tfvs <- toWait (freeRSV tau)
                                                   return (sfvs `trunion` tfvs)
  clFreeRSV _   pred                          = toWait (freeRSV pred)

----------------------------------------
trunion (as,bs,cs) (as',bs',cs') = (as`union`as',bs`union`bs',cs`union`cs')

clFreeVar top (RTVar t) = if top 
                           then toGo    (return [t])
                           else toOther (return [t])
clFreeVar top tau       = clFreeRTV top tau

   
clFreeFun isFix top tau1 tau2 =
    case (tau1,tau2) of
      (RTVar t1, RTVar t2) -> if isFix 
                               then toOther (return [t1,t2])
                               else toWait (return [t1,t2])
      (_       , RTVar t2) -> do ufvs <- clFreeRTV top tau1
                                 lfvs <- if top
                                          then toGo (return [t2])
                                          else toWait (return [t2])
                                 return (ufvs `trunion` lfvs)
      (RTVar t1, _)        -> do ufvs <- clFreeRTV top tau2
                                 lfvs <- toOther (return [t1])
                                 return (ufvs `trunion` lfvs)
      (_       , _)        -> clFreeRTVFor2 top tau1 tau2

clFreeOp tau tau1 tau2 =
  do tfvs  <- toWait  (freeRTV tau)
     t1fvs <- toOther (freeRTV tau1)
     t2fvs <- toOther (freeRTV tau2)
     return ((tfvs `trunion` t1fvs) `trunion` t2fvs)

clFreeGuard top taub pred =
  do bfvs <- toOther (freeRTV taub)
     pfvs <- clFreeRTV top pred
     return (bfvs `trunion` pfvs)

----------------------------
---- Begin Alejandro -------
----------------------------
clFreeGuardDinData top taud pred =
  do dfvs <- clFreeVar top taud 
     pfvs <- clFreeRTV top pred
     return (dfvs `trunion` pfvs)

clFreeHasC top tau1 tau2 = 
  do tfvs1  <- clFreeVar top tau1
     tfvs2  <- clFreeRTV top tau2 
     return (tfvs1 `trunion` tfvs2)

clFreeConstrsOf top tau = 
  do tfvs  <- clFreeVar top tau
     return tfvs

clFreeUnifyData top t tau = 
  do tfvs1 <- clFreeVar top t
     tfvs2 <- clFreeRTV top tau 
     return (tfvs1 `trunion` tfvs2)
----------------------------
------ End Alejandro -------
----------------------------
clFreePolyOp top tau taue taus =
  do tfvs <- toWait (freeRTV tau)
     efvs <- clFreeVar top taue
     sfvs <- clFreeRTV top taus
     return ((tfvs `trunion` efvs) `trunion` sfvs)

clFreeRTVFor2 top tau1 tau2 =
  do t1fvs <- clFreeRTV top tau1
     t2fvs <- clFreeRTV top tau2
     return (t1fvs `trunion` t2fvs)

clFreeRSV_MG top sigma1 sigma2 =
    case (sigma1, sigma2) of
     (RSVar s1, RSVar s2) -> toWait (return [s1,s2])
     (_       , RSVar s2) -> do ufvs <- toWait (freeRSV sigma1)
                                lfvs <- if top
                                         then toGo (return [s2])
                                         else toWait (return [s2])
                                return (ufvs `trunion` lfvs)
     (RSVar s1, _)        -> do ufvs <- toWait (freeRSV sigma2)
                                lfvs <- if top
                                         then toGo (return [s1])
                                         else toWait (return [s1])
                                return (ufvs `trunion` lfvs)
     (_       , _)        -> toWait (freeRSV [sigma1,sigma2])

