module ResidualLangPrint
  (ppresidualterm, ppresidualscheme, ppresidualtype
  ,ppresidualpredicate, ppresidualpredicates, showRTVar, showRSVar) 
 where

import Char
import Maybe

import MyPrelude
import PrettyPrinter

import BasicData
import SourceLang
import qualified SourceLangPrint
import ResidualLang 

---------------------------------
-- Printing residual expressions
---------------------------------
ppresidualterm :: ResidualTerm -> String
ppresidualterm   e = pretty 80 60 (ppexp (stripREtas e))

ppresidualscheme :: ResidualTypeScheme -> String
ppresidualscheme s = pretty 80 60 (pprests s)

ppresidualtype :: ResidualType -> String
ppresidualtype   t = pretty 80 60 (pprtype t)

ppresidualpredicate :: Pred -> String
ppresidualpredicate p = pretty 80 60 (pppred p [])

ppresidualpredicates :: Preds -> String
ppresidualpredicates d = pretty 80 60 (pppreds d [])

-------------------------------
-- Pretty printing expressions
-------------------------------
ppexp :: ResidualTerm -> Doc
ppexp e@(RLam _ _)      = pplam [] e
ppexp (RLet [] e)     = ppexp e 
ppexp (RLet xes 
        e@(RLet _ _)) = (text "let" <> ppdecls xes)
                        $$
                        (nest 1 (text"in"))
                        $$
                        (ppexp e)
ppexp (RLet xes e)    = sep [ text "let" <> ppdecls xes
                            , nest 1 (text "in ") <> ppexp e
                            ]
ppexp (RELam h e)     = if null h
                         then ppexp e
                         else sep [ ppelams h, nest 1 (ppexp e) ]
ppexp (RELet x e1 e2) = sep [ text "let_v" <+> ppdecl x (ppexp e1)
                            , nest 1 (text"in")
                            , ppexp e2
                            ]
ppexp (RIf e e1 e2)   = (text "if " <> ppexp e)
                        $$
                        (nest 1 (text "then ") <> ppexp e1)
                        $$
                        (nest 1 (text "else ") <> ppexp e2)
ppexp (REIf v e1 e2)  = (text "if_v " <> ppevidence v)
                        $$
                        (nest 1 (text "then ") <> ppexp e1)
                        $$
                        (nest 1 (text "else ") <> ppexp e2)
ppexp (RECase v pes)  = (text "case_v" <+> ppevidence v <+> text "of")
                        $$
                        (nest 1 (ppecasestats pes))
                        $$ text "esac"
------------------------------
-------- Alejandro -----------
------------------------------
ppexp (REConstrs names)  = text (show names)
------------------------------
------------------------------
------------- End ------------
------------------------------

ppexp (RData [] e)    = ppexp e                         
ppexp (RData ds e)    = ppdataexps ds $$ ppexp e        
ppexp e               = ppbexp e

-------------------------------
pplam xs (RLam x e) = pplam (x:xs) e
pplam xs e          = sep [ ppbounds (reverse xs)
                          , nest 1 (ppexp e)
                          ]

ppbounds [] = text ""
ppbounds [x] = text "\\" <> text (x ++ " ->")
ppbounds (x:xs) = text "\\" <> text (x ++ " ->") <+> ppbounds xs

-------------------------------
ppbexp (RBoolOp Or e1 e2)    = ppbterm e1 <+> ppboolop Or 
                                          <+> ppbexp e2
ppbexp e                     = ppbterm e

-------------------------------
ppbterm (RBoolOp And e1 e2)  = ppbfact e1 <+> ppboolop And 
                                          <+> ppbterm e2
ppbterm e                    = ppbfact e

-------------------------------
ppbfact (RRelOp rop e1 e2)   = ppnexp e1  <+> pprelop rop 
                                          <+> ppnexp e2
ppbfact e                    = ppnexp e

-------------------------------
ppnexp (RArithOp nop e1 e2)  | nop==Add || nop==Sub
                             = ppnexp e1  <+> pparithop nop 
                                          <+> ppnterm e2
ppnexp e                     = ppnterm e

-------------------------------
ppnterm (RArithOp Mul e1 e2) = ppnterm e1 <+> pparithop Mul 
                                          <+> ppnfact e2
ppnterm e                    = ppnfact e

-------------------------------
ppnfact (RArithOp Exp e1 e2) = ppnfact e1 <+> pparithop Exp 
                                          <+> ppfexp e2
ppnfact e                    = ppfexp e

-----------------------------
ppfexp  (RApp e1 e2)    = sep [ ppfexp e1
                              , nest 2 (text "@" <+> ppprim e2)
                              ]
ppfexp  e               = ppfexp' e

ppfexp' (REApp e v)     = if null v
                           then ppfexp e
                           else sep [ ppfexp' e
                                    , nest 2 (ppevs v)
                                    ]
ppfexp' (REAppS e e')   = sep [ ppfexp' e
                              , nest 2 (text "@_v" <+> ppprim e')
                              ]
ppfexp' e               = ppprim e

-----------------------------
ppprim (RFix e)         = ppopr "fix" (ppaexp e)
ppprim (RProj i n e)    = ppopr (proj i n) (ppaexp e)
ppprim (RProjS i n e)   = ppopr (proj i n ++ tokenStatic) (ppaexp e)
ppprim e                = ppaexp e

ppaexp (RVar s)         = text s
ppaexp (REVar h)        = text (showEVar h)
ppaexp (RInt n)         = text (show n)
ppaexp (RBool b)        = text (show b)
ppaexp (RTuple es)      = pptuple (map ppexp es)
ppaexp (RTupleS [])     = text "*"
ppaexp (RTupleS es)     = pptuple (map ppexp es) <> text tokenStatic
ppaexp (REFill v e)     = ppevcls v <+> (sep [ text "["
                                             , nest 1 (ppexp e)
                                             , text "]"
                                             ]
                                        )
ppaexp (RELam [] e)     = ppaexp e
ppaexp (RLet [] e)      = ppaexp e
ppaexp (RConstruc c xs) = ppargs (text c) ppaexp xs    
ppaexp e@(RCase _ _)    = ppcaseStatement e            
-------------------------------
---- Begin Alejandro ----------
-------------------------------
ppaexp (RDConstruc c et xs) = ppargs (text (c++"_") <> ppaexp et)
                              ppaexp xs
ppaexp e@(RProtoCase _ _ _ ) = ppprotocaseStatement e
-------------------------------
------ End Alejandro ----------
-------------------------------
ppaexp e                = paren (ppexp e)

ppdecl  x ppe = sep [text x <+> text "=", nest 2 ppe]
ppdecls []  = text ""
ppdecls [(x,e)] = text " " <> ppdecl x (ppexp e)
ppdecls ((x,e):ds)  = 
     (text " " <> ppdecl x (ppexp e))
     $$
     foldr1 ($$) (map (\(x,e) -> 
                         nest (-1) 
                           (text ";" <+> ppdecl x (ppexp e))
                      ) ds)
        
-----------------------------
pptuple []         = text "()"
pptuple [ppe]      = paren ppe
pptuple (ppe:ppes) = sep (   [ text "(" <+> ppe ]
                          ++ (map (\ppe -> text "," <+> ppe) ppes) 
                          ++ [ text ")" ]
                         )
proj 1 2 = "fst"
proj 2 2 = "snd"
proj i n = "proj_" ++ show i ++ "_" ++ show n

ppboolop  bop = text (showbop bop)
pprelop   rop = text (showrop rop)
pparithop nop = text (shownop nop)

--------------
--begin @F----
--------------
ppargs d pp [] = d
ppargs d pp xs = d <+> foldr1 (<+>) (map pp xs)

ppcaseStatement (RCase e cs) =
   (text "case" <+> ppaexp e <+> text "of" )
   $$
   (nest 1 (ppcasestats cs))
   $$ text "esac"
 
----------------------
-- Begin Alejandro ---
----------------------
ppprotocaseStatement (RProtoCase e h_cs cs) =
   (text "protocase" <+> ppaexp e)
   $$
   (text "with constructors" <+> ppaexp h_cs <+> text "of")
   $$
   (nest 1 (ppcasestats cs))
   $$ text "esac"



ppcasestats xs = ppcasestats' ppfexp xs
ppecasestats xs = ppcasestats' text xs

ppcasestats' :: (a->Doc) -> [(a,ResTerm)] -> Doc
ppcasestats' f []  = text ""
ppcasestats' f [c] = ppcasestat' f c <> text ";"
ppcasestats' f (c:cs) = (ppcasestat' f c <> text ";")
                        $$ ppcasestats' f cs

ppcasestat' f (st1,st2) = sep [ f st1
                              , nest 2 ((text "->") <+> (ppexp st2))
                              ]

---------------------------
ppdataexps es = foldr ($$) (text "") (map ppdataexp es)

ppdataexp (RDataDef name ts ss delta cs) = 
 sep [
      text "data"
       <+>
       ppargs (ppargs (text name)
                      (flip pprtf []) ts
              )       (flip pprsf []) ss
       <+>
       (text "=")
     , nest 2 (ppWithContext delta (ppdataconstructors cs) [])
     , text ";"
     ]

ppdataconstructors [] = text ""
ppdataconstructors [c] = ppdataconstructor c
ppdataconstructors (c:cs)
   = sep (ppdataconstructor c :
          map (\c -> nest (-2) (text "|" <+> ppdataconstructor c)) cs)

ppdataconstructor (Constructor name sts)
   = ppargs (text name) (\t -> pprtf t []) sts
------------
--end @F----
------------

----------------------------
-- Printing evidence
----------------------------
ppevidence :: EvTerm -> Doc
ppevidence (ECxt c)         = ppcontext c
ppevidence (EComp v1 v2)    = paren (sep [ ppevcls v1 
                                         , nest 2 (text "." <+> ppevcls v2)
                                         ])
ppevidence (EProjS i n v)   = ppopr (proj i n) (ppevidence v)
ppevidence (EStaticExp e)   = ppexp e
ppevidence (EConstrs names) = text (show names) 
ppevidence v                = ppevcls v

-----------------------------
ppevcls (EVar  h)    = text (showEVar h)
ppevcls (EInt  n)    = text (show n)
ppevcls (EBool b)    = text (show b)
ppevcls (EConstr c)  = text c
ppevcls (ETupleS vs) = pptuple (map ppevidence vs) <> text tokenStatic
ppevcls (EAppS e e') = sep [ ppfexp' e
                           , nest 2 (text "@_v" <+> ppprim e')
                           ]
ppevcls v            = paren (ppevidence v)
                                 
----------------------------
-- Printing contexts
-----------------------------
ppcontext :: CxtTerm -> Doc
ppcontext (CELam h c)     = if null h
                             then ppcontext c
                             else sep [ ppelams h, ppcontext c ]
ppcontext (CELet x c e)   = sep [ text "let_v" 
                                   <+> ppdecl x (ppcontext c)
                                , nest 1 (text"in")
                                , ppexp e
                                ]
ppcontext c               = ppcontextappl c

-----------------------------
ppcontextappl CHole       = text "[]"
ppcontextappl (CEApp c v) = if null v
                             then ppcontextappl c
                             else sep [ ppcontextappl c
                                      , nest 2 (ppevs v) ]
ppcontextappl c           = paren (ppcontext c)

----------------------------
-- Auxiliaries
-----------------------------
showEVar   h = mkId "h" h
showRTVar  t = "T_" ++ show t
showRSVar  s = mkId "s" s
showRTGen it = mkId "gt" it
showRSGen is = mkId "gs" is

-----------------------------
ppeslams [x]    = ppeslam x 
ppeslams (x:xs) = ppeslam x <> ppeslams xs 

ppeslam :: ResTermVble -> Doc
ppeslam x = text "\\_v" <> text (x ++ " ->")

-----------------------------
ppsapps [e]    = ppsapp e
ppsapps (e:es) = sep [ ppsapp e , ppsapps es ]

ppsapp e = text "@_v" <+> ppprim e

-----------------------------
ppelams [h]    = ppelam h 
ppelams (h:hs) = ppelam h <> ppelams hs 

ppelam :: EvVble -> Doc
ppelam h = text "/\\" <> text (showEVar h ++ ".")

-----------------------------
pptlams []     = text ""
pptlams [t]    = pptlam t 
pptlams (t:ts) = pptlam t <> pptlams ts

pptlam :: ResTypeVble -> Doc
pptlam t = text "//\\\\" <> text (showRTVar t ++ ".")

-----------------------------
ppslams []     = text ""
ppslams [s]    = ppslam s 
ppslams (s:ss) = ppslam s <> ppslams ss

ppslam :: ResSchemeVble -> Doc
ppslam s = text "//\\\\" <> text (showRSVar s ++ ".")

-----------------------------
ppevs [v]    = ppev v
ppevs (v:vs) = sep [ ppev v , ppevs vs ]

ppev v = text "{{" <+> ppevidence v <+> text "}}"

-----------------------------
pptys []     = text ""
pptys [t]    = ppty t
pptys (t:ts) = sep [ ppty t , pptys ts ]

ppty t = text "<<" <+> pprtype t <+> text ">>"

-----------------------------
ppschs []     = text ""
ppschs [s]    = ppsch s
ppschs (s:ss) = sep [ ppsch s , ppschs ss ]

ppsch s = text "<<" <+> pprests s <+> text ">>"

-----------------------------
ppidslist [] = text ""
ppidslist [x] = text x
ppidslist (x:xs) = text x <> text "," <+> ppidslist xs

----------------------------
-- Printing residual types
----------------------------
pprests :: ResidualTypeScheme -> Doc
pprests s = pprs s []

pprtype    t = pprt t []
pprtypesig t = text ""
--pprtypesig t = text " ::" <+> pprt t []

-----------------------------
pprt (RTFun t1 t2) gs = sep [ pprtf t1 gs
                            , nest 1 (text "->")
                            , nest 1 (pprt t2 gs)
                            ]
pprt t             gs = pprtpref t gs                                

pprtpref (RTData n ts ss) gs | not (null ts) || not (null ss) =      
      ppargs (ppargs (text n)                                        
                     (\t -> pprtf t gs) ts                           
             )       (\s -> pprsf s gs) ss                           
pprtpref (RTConstruc c xs@(_:_)) gs =                                
      ppargs (text c) (\t -> pprtf t gs) xs                          
pprtpref (RTPoly s)        gs = text "poly" <+> pprsf s gs
--------------------------
----- Begin Alejandro ----
--------------------------
pprtpref (RTDData td ts) gs =      
	text "data" <+> pprtf td gs <+> text "with" 
    <+> pprtf ts gs
{-
pprtpref (RTDData td ts)        gs = text "data" <+> pprtf td gs
                                     <+> text "with" <+> pptys ts
-}
--------------------------
------ End Alejandro -----
--------------------------

pprtpref t gs = pprtf t gs                                           

-----------------------------
pprtf (RTVar t)         _  = text (showRTVar t)
pprtf (RTGen l t)       gs = gent l t gs
pprtf RTIntD            gs = text "Int" 
pprtf RTBoolD           gs = text "Bool" 
pprtf (RTInt n)         gs = text (show n)
pprtf (RTBool b)        gs = text (show b)
--pprtf (RTPoly s)        gs = text "poly" <+> pprsf s gs
pprtf (RTTuple ts)      gs = pptuple (map (\t -> pprtpref t gs) ts)
pprtf (RTTupleS ts)     gs = pptuple (map (\t -> pprtpref t gs) ts) 
                               <> text tokenStatic
pprtf (RTStaticExp e)   gs = ppexp e
pprtf (RTFixS   _ t s)  gs = sep [ text "rec("
                                 , nest 4 (pprt t gs)
                                 , nest 4 (text "::")
                                 , nest 4 (text "...") --(pprs s gs)
                                 , nest 3 (text ")") 
                                 ]
pprtf (RTTypedS _ t s)  gs = sep [ text "clos("
                                 , nest 5 (pprt t gs)
                                 , nest 5 (text "::")
                                 , nest 5 (pprs s gs)
                                 , nest 4 (text ")" )
                                 ]
pprtf (RTData n [] [])  _  = text n                         
pprtf (RTConstruc c []) _  = text c                         
pprtf t                 gs = paren (pprt t gs)

-----------------------------
pprs (RSForAll tm sm qt) gs = let gs1 = ((tm,sm):gs)
                               in if tm+sm == 0
                                   then ppqt qt gs1
                                   else sep [ forAll tm sm gs1
                                            , nest 1 (ppqt qt gs1) 
                                            ]
pprs s                   gs = pprsf s gs

-----------------------------
pprsf (RSVar s)               _  = text (showRSVar s)
pprsf (RSGen l s)             gs = gens l s gs
pprsf (RSForAll 0 0 ([]:=>t)) gs = pprtf t ((0,0):gs)
pprsf (RSTupleS ts)           gs = pptuple (map (\t -> pprt t gs) ts) 
                                    <> text tokenStatic
pprsf s                       gs = paren (pprs s gs)

-----------------------------
ppqt (ps :=> t) gs = ppWithContext ps (pprt t gs) gs

ppWithContext [] doc _ = doc
ppWithContext ps doc gs = sep [ pppreds ps gs
                              , nest 1 (sep [ text "=>", nest 1 doc ])
                              ]

-----------------------------
pppreds []  _  = text ""
pppreds [p] gs = ppepred p gs
pppreds ps  gs = sep ((map (\p -> ppepred p gs <> text ",") (init ps)) 
                     ++ [ pppreds [last ps] gs ])

ppepred (EVar h,d) gs = pppred' (text (showEVar h) <> text " :: ") d gs
pppred d gs = pppred' (text "") d gs

pppred' dc (IsMG s1 s2) gs = sep [ dc <> text ("IsMG")
                                 , nest 2 (pprsf s1 gs)
                                 , nest 2 (pprsf s2 gs)
                                 ]
pppred' dc (IsMGmorph t s1 s2) gs = sep [ dc <> text ("IsPolym") 
                                        , nest 2 (paren(pprtf t gs))
                                        , nest 2 (pprsf s1 gs)
                                        , nest 2 (pprsf s2 gs)
                                        ]
pppred' dc (IsInt  t)   gs = dc <> text ("IsInt")  <+> pprtf t gs
pppred' dc (IsBool t)   gs = dc <> text ("IsBool") <+> pprtf t gs
pppred' dc (IsIf tr tb t1 t2) gs = 
  dc <> text ("(") <> 
   pprtf tr gs <+> text ":="
   <+> text "if"    <+> pprtf tb gs
   <+>  text "then" <+> pprtf t1 gs
   <+>  text "else" <+> pprtf t2 gs 
  <> text ")"
pppred' dc (IsCase tr te cts) gs = 
  dc <> text ("(") <> 
   pprtf tr gs <+> text ":="
   <+> ((text "case" <+> pprtf te gs <+> text "of")
        $$ nest 1 (pptcasestats cts gs)
        $$ text "esac"
       )
  <> text ")"
pppred' dc (IsFunS t t') gs =  sep [ dc <> text "IsFunS"
                                   , nest 1 (pprtf t gs)
                                   , nest 1 (pprtf t' gs)
                                   ]
pppred' dc (IsFixS t t') gs =  sep [ dc <> text "IsFixS"
                                   , nest 1 (pprtf t gs)
                                   , nest 1 (pprtf t' gs)
                                   ]
pppred' dc (IsArithOp nop t t1 t2) gs = 
  dc <> text ("(") <> 
   pprtf t gs <+> text ":=" <+> pprtf t1 gs <+> pparithop nop <+> pprtf t2 gs 
  <> text ")"
pppred' dc (IsBoolOp bop t t1 t2) gs = 
  dc <> text ("(") <> 
   pprtf t gs <+> text ":=" <+> pprtf t1 gs <+> ppboolop bop <+> pprtf t2 gs 
  <> text ")"
pppred' dc (IsRelOp rop t t1 t2) gs = 
  dc <> text ("(") <> 
   pprtf t gs <+> text ":=" <+> pprtf t1 gs <+> pprelop rop <+> pprtf t2 gs 
  <> text ")"
pppred' dc (IsConstrOf sdt t) gs = sep [ dc <> text ("IsConstrOf") 
                                       , nest 2 (SourceLangPrint.pptypefact sdt)
                                       , nest 2 (pprtf t gs)
                                       ]
pppred' dc (Guarded t d)    gs = dc <> pprtf t gs <+> text "?" <+> pppred d gs
pppred' dc (GuardedNeg t d) gs = dc <> text "!" <+> pprtf t gs <+> 
                                       text "?" <+> pppred d gs
pppred' dc (GuardedConstr c t d) gs = dc <> text "(" <> pprtf t gs <+> 
                                            text ("is a "++c++")?") <+> 
                                            pppred d gs

--- Begin Alejandro ---  
pppred' dc (GuardedDinData c t d) gs = dc <> text "(" 
                                          <> text (c ++ " in")
                                          <+> pprtf t gs 
                                          <> text ")?"
                                          <+> pppred d gs

pppred' dc (HasC t1 c t2) gs = dc <> text "HasC" 
                                  <+> pprtf t1 gs 
                                  <+> text c 
                                  <+> pprtf t2 gs 

pppred' dc (ConstrsOf t) gs = dc <> text "ConstrsOf"
                                 <+> pprtf t gs

pppred' dc (IndexOf t) gs = dc <> text "IndexOf"
                               <+> pprtf t gs

pppred' dc (UnifyData t1 cname t2) gs = 
  dc <> text ("(") <> 
   pprtf t1 gs <> text "[" <> text cname <> text "]" 
   <+> text "~" <+> pprtf t2 gs
   <> text ")"
--- End Alejandro ---

pppred' dc (Unify t1 t2)    gs = 
  dc <> text ("(") <> 
   pprtf t1 gs <+> text "~" <+> pprtf t2 gs
  <> text ")"
pppred' dc (Fail msg)       gs = dc <> text ("Fail: " ++ msg)

pptcasestats []     gs = text ""
pptcasestats [c]    gs = pptcasestat c gs <> text ";"
pptcasestats (c:cs) gs = (pptcasestat c gs <> text ";")
                           $$ pptcasestats cs gs

pptcasestat (st1,st2) gs = sep [ pprtf st1 gs
                               , nest 2 ((text "->") <+> (pprt st2 gs))
                               ]

-----------------------------
forAll 0  0  _  = text ""
forAll tm sm gs = text "\\/" <> gentsss tm sm gs <> text "."

gentsss 0  0  _  = text ""
gentsss 0  sm gs = genss sm gs
gentsss tm 0  gs = gents tm gs
gentsss tm sm gs = gents tm gs <+> genss sm gs

gents 0 _  = text ""
gents 1 gs = gent 1 0 gs
gents n gs = gents (n-1) gs <+> gent 1 (n-1) gs

genss 0 _  = text ""
genss 1 gs = gens 1 0 gs
genss n gs = genss (n-1) gs <+> gens 1 (n-1) gs

gent l t gs = let it = lookForRGen l t gs
               in text (showRTGen it)

gens l s gs = let is = lookForSGen l s gs
               in text (showRSGen is)

lookForRGen 0 x sts      = lookFRG x 0 sts
lookForRGen l x (_:sts)  = lookForRGen (l-1) x sts
lookFRG x ac []          = x+ac
lookFRG x ac ((t,_):sts) = lookFRG x (ac+t) sts

lookForSGen 0 x sts      = lookFSG x 0 sts
lookForSGen l x (_:sts)  = lookForSGen (l-1) x sts
lookFSG x ac []          = x+ac
lookFSG x ac ((_,s):sts) = lookFSG x (ac+s) sts

-----------------------------
ppopr s ppe = text s <+> ppe

paren d = sep [ text "(" <+> d, text ")" ]

