module ResidualUnification where

import MoreMonad
import ResidualLang
import ResidualLangPrint
import ResidualReduction
import SpecState
import SpecMonad
import ResidualLangFVs

-------------------------------------------------------
unifErr t1 t2 = do msg <- unifErrMsg t1 t2
                   fail msg
unifErrMsg t1 t2 = do t1' <- unfoldAll t1
                      t2' <- unfoldAll t2
                      return ("\n Cannot unify \n  "
                              ++ ppresidualtype t1' ++ " with \n  "
                              ++ ppresidualtype t2')

shouldNotBeIn fn cn = error (fn ++ " should not be defined for " ++ cn ++ "s!")

-------------------------------------------------------
occurCheckR w t = 
  do ftvs <- freeRTV t
     let ok = not (w `elem` ftvs)
      in do when ok (set "RTVar(tmp in occurCheck)" tmptype w t)
            return ok

occurCheckS w s = 
  do fsvs <- freeRSV s
     let ok = not (w `elem` fsvs)
      in do when ok (set "RSVar(tmp in occurCheck)" tmpscheme w s)
            return ok


-------------------------------------------------------
efective f a a'             = do yes <- f a a'
                                 (if yes then confirmUnification 
                                         else discardUnification)
                                 return yes

efectiveUnify a a'             = efective unify a a'
efectiveMatch tvs svs a a'     = efective (match tvs svs) a a'
efectiveMatchUnif tvs svs a a' = efective (matchUnif tvs svs) a a'

unifiable a a' = do yes <- unify a a'
                    discardUnification
                    return yes

-- --------------------------------------------------------
-- Unification of free variables on residual types 
-- --------------------------------------------------------
{-
unifyAll []  = return True
unifyAll [x] = return True
unifyAll (y:x:xs) = do b  <- unify x y
                       b' <- unifyAll (x:xs)
                       return (b && b')
-}

unify :: (MetaVbles a, UnifnMatch a) => a -> a -> SpecM Bool
unify  a a' = do --newTemporarySubst
                 unify' a a'

unify' :: (MetaVbles a, UnifnMatch a) => a -> a -> SpecM Bool
unify' a a' = do a1  <- fullindir a
                 a1' <- fullindir a'
                 unify'' a1 a1'

unify'' :: UnifnMatch a => a -> a -> SpecM Bool
unify'' a a' =
  umR (\vw@(RTVar w) vx -> 
         case vx of
          RTVar x       -> 
           case compare w x of
            EQ -> return True
            LT -> do set "RTVar(tmp in unify)" tmptype x vw; return True
            GT -> do set "RTVar(tmp in unify)" tmptype w vx; return True
          t             -> occurCheckR w t)
      (\t (RTVar w)     -> occurCheckR w t)
      (\vw@(RSVar w) vx ->
         case vx of
          RSVar x       -> 
           case compare w x of
            EQ -> return True
            LT -> do set "RSVar(tmp in unify)" tmpscheme x vw; return True
            GT -> do set "RSVar(tmp in unify)" tmpscheme w vx; return True
          s             -> occurCheckS w s)
      (\s (RSVar w)     -> occurCheckS w s)
      a a'

-- --------------------------------------------------------
-- Matching of given residual type and scheme variables
-- --------------------------------------------------------
match, match' :: UnifnMatch a => [ResTypeVble] -> [ResSchemeVble] -> 
                           a -> a -> SpecM Bool
match tvs svs a a' = do --newTemporarySubst
                        match' tvs svs a a'
match' tvs svs a a' =
  umR (\vw@(RTVar w)  t -> 
         let okw = w `elem` tvs
          in do when okw (set "RTVar(tmp in match)" tmptype w t)
                return (okw || vw==t))
      (\_ _             -> return False)
      (\vw@(RSVar w)  s -> 
         let okw = w `elem` svs
          in do when okw (set "RSVar(tmp in match)" tmpscheme w s)
                return (okw || vw==s))
      (\_ _             -> return False)
      a a'

-- --------------------------------------------------------
-- Matching of given residual type and scheme variables and 
--  unification of the rest
-- --------------------------------------------------------
matchUnif,matchUnif' :: UnifnMatch a => [ResTypeVble] -> [ResSchemeVble] -> 
                             a -> a -> SpecM Bool
matchUnif tvs svs a a' = do --newTemporarySubst
                            matchUnif' tvs svs a a'

matchUnif' tvs svs a a' = 
   umR (\vw@(RTVar w)  t -> 
          let okw = w `elem` tvs
           in if okw 
               then do set "RTVar(tmp in matchUnif)" tmptype w t
                       return True
               else do unify' vw t)
       (\t vx@(RTVar _)  -> unify' vx t)
       (\vw@(RSVar w)  s -> 
          let okw = w `elem` svs
           in if okw 
               then do set "RSVar(tmp in matchUnif)" tmpscheme w s
                       return True
               else do unify' vw s)
       (\s vx@(RSVar _)  -> unify' vx s)
       a a'

-- --------------------------------------------------------
-- Folding of residual structures for unification and matching
-- --------------------------------------------------------
class (MetaVbles a, Eq a) => UnifnMatch a where
  umR, umR', umR''  
      :: (ResType       -> ResType       -> SpecM Bool) ->
         (ResType       -> ResType       -> SpecM Bool) ->
         (ResTypeScheme -> ResTypeScheme -> SpecM Bool) ->
         (ResTypeScheme -> ResTypeScheme -> SpecM Bool) ->
         a -> a -> SpecM Bool
  umR   ft gt fs gs a a' = do newTemporarySubst
                              umR' ft gt fs gs a a'
  umR'  ft gt fs gs a a' = do a_1  <- fullindir a
                              a'_1 <- fullindir a'
                              umR'' ft gt fs gs a_1 a'_1
  umR'' _  _  _  _  a a' = return (a==a')

instance UnifnMatch Char
instance UnifnMatch Int

-- -------------------------------------------------------------
instance UnifnMatch ResType where
  ---------------------
  umR'' _  _  _  _  t@(RTGen _ _)         
                    t'                   = return (t==t')
  umR'' _  _  _  _  _                     
                    (RTGen _ _)          = return False
  umR'' ft _  _  _  vw@(RTVar w)          
                    t                    = ft vw t
  umR'' _  gt _  _  t                     
                    vx@(RTVar _)         = gt t  vx
  umR'' ft gt fs gs (RTFun tau1  tau2)    
                    (RTFun tau1' tau2')  = liftM2 (&&) 
                                            (umR' ft gt fs gs tau1 tau1')
                                            (umR' ft gt fs gs tau2 tau2')
  umR'' ft gt fs gs (RTTuple  ts)                      
                    (RTTuple  ts')       = umR' ft gt fs gs ts ts'
  umR'' ft gt fs gs (RTTupleS ts)         
                    (RTTupleS ts')       = umR' ft gt fs gs ts ts'
  umR'' ft gt fs gs (RTPoly sigma)        
                    (RTPoly sigma')      = umR' ft gt fs gs sigma sigma'
  umR'' ft gt fs gs (RTData d  ts  ss)        
                    (RTData d' ts' ss')  | d==d' 
                                         = liftM2 (&&) 
                                            (umR' ft gt fs gs ts ts')
                                            (umR' ft gt fs gs ss ss')
  ------------------------
  ---- Begin Alejandro ---
  ------------------------
  umR'' ft gt fs gs (RTDData td ts )        
                    (RTDData td' ts') = do {-traceSM "Unfica" td
                                           traceSM "con" td' 
                                           traceSM "y" ts 
                                           traceSM "con" ts' -}
                                           liftM2 (&&) 
                                            (umR' ft gt fs gs td td')
                                            (umR' ft gt fs gs ts ts')
  ------------------------
  ---- End Alejandro -----
  ------------------------

  umR'' ft gt fs gs (RTConstruc c  ts)        
                    (RTConstruc c' ts')  | c==c' 
                                         = umR' ft gt fs gs ts ts'
  umR'' ft gt fs gs (RTStaticExp e)       
                    (RTStaticExp e')     = do e1  <- unfoldAll e
                                              e'1 <- unfoldAll e'
                                              let e2  = ered e1
                                              let e'2 = ered e'1
                                              return (eqR e2 e'2)
  umR'' ft gt fs gs (RTFixS ts  t  s)   
                    (RTFixS ts' t' s')   = do vts <- umR' ft gt fs gs ts ts'
                                              vt  <- umR' ft gt fs gs t  t'
                                              vs  <- umR' ft gt fs gs s  s'
                                              return (vts&&vt&&vs)
  umR'' ft gt fs gs (RTTypedS ts  t  s)   
                    (RTTypedS ts' t' s') = do vts <- umR' ft gt fs gs ts ts'
                                              vt  <- umR' ft gt fs gs t  t'
                                              vs  <- umR' ft gt fs gs s  s'
                                              return (vts&&vt&&vs)
  umR'' _  _  _  _  tau                   
                    tau'                 = return (tau==tau')

-- -------------------------------------------------------------
instance UnifnMatch ResTypeScheme where
  ---------------------
  umR'' _  _  _  _  s@(RSGen _ _) 
                    s'                     = return (s==s')
  umR'' _  _  _  _  s'            
                    (RSGen _ _)            = return False
  umR'' _  _  fs _  vw@(RSVar _)  
                    s                      = fs vw s
  umR'' _  _  _  gs s             
                    vx@(RSVar _)           = gs s  vx
  umR'' ft gt fs gs (RSForAll tm  sm  qt)
                    (RSForAll tm' sm' qt') = if tm==tm' && sm==sm'
                                              then umR' ft gt fs gs qt qt'
                                              else return False
  umR'' ft gt fs gs (RSTupleS ss)  
                    (RSTupleS ss')         = umR' ft gt fs gs ss ss'
  umR'' _  _  _  _  sigma         
                    sigma'                 = return (sigma==sigma')

-- -------------------------------------------------------------
instance UnifnMatch QualType where
  umR'' ft gt fs gs (ps:=>t) (ps':=>t') = liftM2 (&&) 
                                           (umR' ft gt fs gs ps ps')
                                           (umR' ft gt fs gs t t')

-- -------------------------------------------------------------
instance UnifnMatch a => UnifnMatch [a] where
  umR'' _  _  _  _  []     []     = return True
  umR'' ft gt fs gs (x:xs) (y:ys) = liftM2 (&&) (umR' ft gt fs gs x y) 
                                                (umR' ft gt fs gs xs ys)
  umR'' _  _  _  _  _      _      = return False

-- -------------------------------------------------------------
instance (UnifnMatch a, UnifnMatch b) => UnifnMatch (a,b) where
  umR'' ft gt fs gs (ax,ay) (bx,by) = liftM2 (&&) (umR' ft gt fs gs ax bx) 
                                                  (umR' ft gt fs gs ay by)

-- -------------------------------------------------------------
instance UnifnMatch Pred where
  umR'' ft gt fs gs (IsMG s1  s2) 
                    (IsMG s1' s2')              = liftM2 (&&) 
                                                   (umR' ft gt fs gs s1 s1') 
                                                   (umR' ft gt fs gs s2 s2')
  umR'' ft gt fs gs (IsMGmorph t  s1  s2) 
                    (IsMGmorph t' s1' s2')      = do bt <- umR' ft gt fs gs t t'
                                                     b1 <- umR' ft gt fs gs s1 s1'
                                                     b2 <- umR' ft gt fs gs s2 s2'
                                                     return (bt&&b1&&b2)
  umR'' ft gt fs gs (IsInt  t) 
                    (IsInt  t')                 = umR' ft gt fs gs t t'
  umR'' ft gt fs gs (IsBool t) 
                    (IsBool t')                 = umR' ft gt fs gs t t'
  umR'' ft gt fs gs (IsFunS t1  t2) 
                    (IsFunS t1' t2')            = liftM2 (&&)
                                                   (umR' ft gt fs gs t1 t1')
                                                   (umR' ft gt fs gs t2 t2')
  umR'' ft gt fs gs (IsFixS t1  t2) 
                    (IsFixS t1' t2')            = liftM2 (&&)
                                                   (umR' ft gt fs gs t1 t1')
                                                   (umR' ft gt fs gs t2 t2')
  umR'' ft gt fs gs (IsIf tr  tb  t1  t2) 
                    (IsIf tr' tb' t1' t2')      = do br<-umR' ft gt fs gs tr tr'
                                                     bb<-umR' ft gt fs gs tb tb'
                                                     b1<-umR' ft gt fs gs t1 t1'
                                                     b2<-umR' ft gt fs gs t2 t2'
                                                     return (br&&bb&&b1&&b2)
  umR'' ft gt fs gs (IsCase tr  te  cts) 
                    (IsCase tr' te' cts')       = do br<-umR' ft gt fs gs tr tr'
                                                     be<-umR' ft gt fs gs te te'
                                                     bs<-umR' ft gt fs gs cts cts'
                                                     return (br&&be&&bs)
  umR'' ft gt fs gs (IsArithOp nop  t  t1  t2) 
                    (IsArithOp nop' t' t1' t2') | nop==nop' 
                                                = do b <-umR' ft gt fs gs t t'
                                                     b1<-umR' ft gt fs gs t1 t1'
                                                     b2<-umR' ft gt fs gs t2 t2'
                                                     return (b && b1 && b2)
  umR'' ft gt fs gs (IsBoolOp  bop  t  t1  t2) 
                    (IsBoolOp  bop' t' t1' t2') | bop==bop' 
                                                = do b <-umR' ft gt fs gs t t'
                                                     b1<-umR' ft gt fs gs t1 t1'
                                                     b2<-umR' ft gt fs gs t2 t2'
                                                     return (b && b1 && b2)
  umR'' ft gt fs gs (IsRelOp   rop  t  t1  t2) 
                    (IsRelOp   rop' t' t1' t2') | rop==rop' 
                                                = do b <-umR' ft gt fs gs t  t'
                                                     b1<-umR' ft gt fs gs t1 t1'
                                                     b2<-umR' ft gt fs gs t2 t2'
                                                     return (b && b1 && b2)
  umR'' ft gt fs gs (IsConstrOf sdt  t) 
                    (IsConstrOf sdt' t')        | sdt == sdt'
                                                = do b <-umR' ft gt fs gs t  t'
                                                     return b
  umR'' ft gt fs gs (Guarded    t  d) 
                    (Guarded    t' d')          = liftM2 (&&)
                                                   (umR' ft gt fs gs t t')
                                                   (umR' ft gt fs gs d d')
  umR'' ft gt fs gs (GuardedNeg t  d)   
                    (GuardedNeg t' d')          = liftM2 (&&)
                                                   (umR' ft gt fs gs t t')
                                                   (umR' ft gt fs gs d d')
  umR'' ft gt fs gs (GuardedConstr c  t  d)   
                    (GuardedConstr c' t' d')    | c==c'
                                                = liftM2 (&&)
                                                   (umR' ft gt fs gs t t')
                                                   (umR' ft gt fs gs d d')
  umR'' ft gt fs gs (Unify t1  t2) 
                    (Unify t1' t2')             = liftM2 (&&)
                                                   (umR' ft gt fs gs t1 t1')
                                                   (umR' ft gt fs gs t2 t2')
  ------------------------------
  -------- Begin Alejandro -----
  ------------------------------
  umR'' ft gt fs gs (IndexOf t) 
                    (IndexOf t')    = umR' ft gt fs gs t t' 
  

  umR'' ft gt fs gs (UnifyData td _ t) 
                    (UnifyData td' _ t') = liftM2 (&&)
                                                (umR' ft gt fs gs td td')
                                                (umR' ft gt fs gs t t')

  umR'' ft gt fs gs (ConstrsOf td) 
                    (ConstrsOf td') = umR' ft gt fs gs td td'
                                               

  umR'' ft gt fs gs (HasC td _ t) 
                    (HasC td' _ t') = liftM2 (&&)
                                            (umR' ft gt fs gs td td')
                                            (umR' ft gt fs gs t t')

  umR'' ft gt fs gs (GuardedDinData _ td p) 
                    (GuardedDinData _ td' p') 
                                     = liftM2 (&&)
                                          (umR' ft gt fs gs td td')
                                          (umR' ft gt fs gs p p')

  ------------------------------
  -------- End Alejandro -------
  ------------------------------


  umR'' _  _  _  _  (Fail _) (Fail _)           = return True
  umR'' _  _  _  _  _        _                  = return False

-- -------------------------------------------------------------
instance UnifnMatch ResTerm where
  umR'' ft gt fs gs (RApp    e1  e2) 
                    (RApp    e1' e2')   = do b1 <- umR' ft gt fs gs e1 e1'
                                             b2 <- umR' ft gt fs gs e2 e2'
                                             return (b1&&b2)
  umR'' ft gt fs gs (RLam x e)
                    (RLam x' e')        | x==x'
                                        = umR' ft gt fs gs e e'
  umR'' ft gt fs gs (RFix e)
                    (RFix e')           = umR' ft gt fs gs e e'
  umR'' ft gt fs gs (RLet xes  e)
                    (RLet xes' e')      = do b  <- umR' ft gt fs gs xes xes'
                                             be <- umR' ft gt fs gs e   e'
                                             return (b&&be)
  umR'' ft gt fs gs (RArithOp nop e1 e2)
                    (RArithOp nop' e1' e2') 
                                        | nop==nop'
                                        = do b1 <- umR' ft gt fs gs e1 e1'
                                             b2 <- umR' ft gt fs gs e2 e2'
                                             return (b1&&b2)
  umR'' ft gt fs gs (RIf e  e1  e2)  
                    (RIf e' e1' e2')    = do b  <- umR' ft gt fs gs e  e'
                                             b1 <- umR' ft gt fs gs e1 e1'
                                             b2 <- umR' ft gt fs gs e2 e2'
                                             return (b&&b1&&b2)
  umR'' ft gt fs gs (RBoolOp bop  e1  e2)
                    (RBoolOp bop' e1' e2') 
                                        | bop==bop'
                                        = do b1 <- umR' ft gt fs gs e1 e1'
                                             b2 <- umR' ft gt fs gs e2 e2'
                                             return (b1&&b2)
  umR'' ft gt fs gs (RRelOp rop  e1  e2)
                    (RRelOp rop' e1' e2') 
                                        | rop==rop'
                                        = do b1 <- umR' ft gt fs gs e1 e1'
                                             b2 <- umR' ft gt fs gs e2 e2'
                                             return (b1&&b2)
  umR'' ft gt fs gs (RTuple es)  
                    (RTuple es')        = umR' ft gt fs gs es es'
  umR'' ft gt fs gs (RProj i  n  e   )
                    (RProj i' n' e'  )  | i==i' && n==n'
                                        = do be <- umR' ft gt fs gs e e'
                                             return be
  umR'' ft gt fs gs (RTupleS es)
                    (RTupleS es')       = umR' ft gt fs gs es es'
  umR'' ft gt fs gs (RProjS i  n  e  )
                    (RProjS i' n' e' )  | i==i' && n==n'
                                        = do be <- umR' ft gt fs gs e e'
                                             return be
  umR'' ft gt fs gs (REFill    v  e) 
                    (REFill    v' e')   = do bv <- umR' ft gt fs gs v v'
                                             be <- umR' ft gt fs gs e e'
                                             return (bv&&be)
  umR'' ft gt fs gs (RELam h  e)
                    (RELam h' e')       | h==h'
                                        = umR' ft gt fs gs e e'
  umR'' ft gt fs gs (RELet x  e1  e2) 
                    (RELet x' e1' e2')  | x==x'
                                        = do b1 <- umR' ft gt fs gs e1 e1'
                                             b2 <- umR' ft gt fs gs e2 e2'
                                             return (b1&&b2)
  umR'' ft gt fs gs (REApp    e  v) 
                    (REApp    e' v')    = do be <- umR' ft gt fs gs e e'
                                             bv <- umR' ft gt fs gs v v'
                                             return (be&&bv)
  umR'' ft gt fs gs (REIf v  e1  e2) 
                    (REIf v' e1' e2')   = do b  <- umR' ft gt fs gs v  v'
                                             b1 <- umR' ft gt fs gs e1 e1'
                                             b2 <- umR' ft gt fs gs e2 e2'
                                             return (b&&b1&&b2)
  umR'' ft gt fs gs (RECase v  pes) 
                    (RECase v' pes')    = do b  <- umR' ft gt fs gs v  v'
                                             b1 <- umR' ft gt fs gs pes pes'
                                             return (b&&b1)
  umR'' ft gt fs gs (REAppS e  ea) 
                    (REAppS e' ea')     = do b  <- umR' ft gt fs gs e  e'
                                             bs <- umR' ft gt fs gs ea ea'
                                             return (b&&bs)
  umR'' ft gt fs gs (RData _ e)                                           
                    (RData _ e')        = umR' ft gt fs gs e e'           
  umR'' ft gt fs gs (RConstruc c  es)                                     
                    (RConstruc c' es')  = do be <- umR' ft gt fs gs es es'
                                             return (c==c'&&be)           
  umR'' ft gt fs gs (RCase e     cs)                                      
                    (RCase e'    cs')   = do be <- umR' ft gt fs gs e  e' 
                                             bc <- umR' ft gt fs gs cs cs'
                                             return (be&&bc)              
  umR'' _  _  _  _   e      
                     e'                 =    return (e==e')

 
-- -------------------------------------------------------------
instance UnifnMatch EvTerm where
  umR'' ft gt fs gs (ECxt  c)
                    (ECxt  c')           = umR' ft gt fs gs c c'
  umR'' ft gt fs gs (EComp v     e)     
                    (EComp v'    e')     = do bv <- umR' ft gt fs gs v v'
                                              be <- umR' ft gt fs gs e e'
                                              return (bv&&be)
  umR'' ft gt fs gs (EStaticExp e)      
                    (EStaticExp e')      =    umR' ft gt fs gs e e'
  umR'' ft gt fs gs (EAppS e  ea)
                    (EAppS e' ea')       = do b  <- umR' ft gt fs gs e  e'
                                              bs <- umR' ft gt fs gs ea ea'
                                              return (b&&bs)
  umR'' ft gt fs gs (ETupleS vs) 
                    (ETupleS vs')        =    umR' ft gt fs gs vs vs'
  umR'' ft gt fs gs (EProjS i  n  v   )
                    (EProjS i' n' v'   ) | i==i' && n==n'
                                         = do bv <- umR' ft gt fs gs v v'
                                              return bv
  umR'' _  _  _  _   v
                     v'                  =    return (v==v')


-- -------------------------------------------------------------
instance UnifnMatch CxtTerm where
  umR'' ft gt fs gs (CELet x  c  e)
                    (CELet x' c' e')    | x==x'
                                        = do bc <- umR' ft gt fs gs c c'
                                             be <- umR' ft gt fs gs e e'
                                             return (bc&&be)
  umR'' ft gt fs gs (CELam xs  ct) 
                    (CELam xs' ct')     | xs==xs'
                                        = umR' ft gt fs gs ct ct'
  umR'' ft gt fs gs (CEApp    c  v)
                    (CEApp    c' v')    = do bc <- umR' ft gt fs gs c c'
                                             bv <- umR' ft gt fs gs v v'
                                             return (bc&&bv)
  umR'' _  _  _  _  c
                    c'                  =    return (c==c')

 
