module SolvingFunS(solveFunS, solveFixS) where

import MyPrelude
import MoreMonad

import ResidualLang
import SpecState
import SpecMonad
import ResidualReduction
import ResidualUnification
import PredsSimplify
import SolvingFailure

------------------------------------
-- It solves variable t to a static function type
--  (a closure). No recursion
--  (ftvs and fsvs are variables still free in
--     the environment)
------------------------------------
solveFunS forcing ftvs fsvs t delta = 
  do iubs <- collectRTFunUBs t delta
     let deltaSol = pick iubs delta
     let ubs      = map gimmeRTFunUB deltaSol
     let sigmas   = map gimmeFunType  ubs
     let exps     = map gimmeFunExp   ubs
     let tauss    = map gimmeFunTuple ubs
 --
 -- Calculates the glb of all the upper bounds
     msigma <- glbs ftvs fsvs sigmas
     case msigma of
      Nothing    -> noSolution forcing (msgNoSolution 1) deltaSol delta
      Just sigma -> 
 --
 -- One of the upper bounds should provide the function body...
       do mie <- pickStaticExp (zip [0..] exps)
          case mie of
           Nothing    -> noSolution forcing msgAmbiguous deltaSol delta
           Just (i,e) -> do 
             if i>=length sigmas then fail (outboundIn "solveFunS(1)") else do
             let sigma' = sigmas!!i
 --
 -- How to go from the provider to the glb? 
             mcxtp <- isMoreGeneral delta sigma' sigma
             case mcxtp of
              Nothing -> error ("Cannot happen! It should exist a context."++
                                "(Something may be wrong with glb...look Fun)")
              Just (cxt, _) -> do
 --
 -- All static expressions should coincide with the one provided
 --  (with suitable conversions!)
                let e'    = redfill cxt e
                let taue' = RTStaticExp e'
                yeses <- unifyStaticExps taue' exps sigmas sigma delta
                -- (old version: no conversions considered -- BAAD!
                --yeses <- mapM (efectiveUnify taue') exps
                if not (and yeses)
                 then noSolution forcing (msgNoSolution 2) deltaSol delta
                 else 
                  do if i>=length tauss 
                      then fail (outboundIn "solveFunS(2)") else do
 --
 -- The tuple types should all be the same
                     let taus' = tauss!!i
                     yeses <- mapM (efectiveUnify taus') tauss
                     if not (and yeses)
                      then noSolution forcing (msgNoSolution 3)
                                      deltaSol delta
                      else 
 --
 -- Finally, the variable t is set to the solution
 --  and the resulting context, simplified.
                       do set "RTVar" residualtype t 
                              (RTTypedS taus' taue' sigma)
                          res@(_,_,delta') <- simplifySolving ftvs fsvs delta 
                          when forcing (checkFailure delta')
                          return res

------------------------------------
-- It solves variable t to a recursive
--  static function type (a closure)
--  (ftvs and fsvs are variables still free in
--     the environment)
------------------------------------
solveFixS forcing ftvs fsvs t delta = 
  do iubs <- collectRTFixUBs t delta
     let deltaSol = pick iubs delta
     let ubs      = map gimmeRTFixUB deltaSol
     let sigmas   = map gimmeFunType  ubs
     let exps     = map gimmeFunExp   ubs
     let tauss    = map gimmeFunTuple ubs
 --
 -- Calculates the glb of all the upper bounds
     msigma <- glbs ftvs fsvs sigmas
     case msigma of
      Nothing    -> noSolution forcing (msgNoSolution 4) deltaSol delta
      Just sigma -> 
 --
 -- One of the upper bounds should provide the function body...
       do mie <- pickStaticExp (zip [0..] exps)
          case mie of
           Nothing    -> noSolution forcing msgAmbiguous deltaSol delta
           Just (i,e) -> do 
             if i>=length sigmas then fail (outboundIn "solveFixS(1)") else do
             let sigma' = sigmas!!i
 --
 -- How to go from the provider to the glb? 
             mcxtp <- isMoreGeneral delta sigma' sigma
             case mcxtp of
              Nothing -> error ("Cannot happen! It should exist a context."++
                                "(Something may be wrong with glb...look Fix)")
              Just (cxt, _) -> do
 --
 -- All static expressions should coincide with the one provided
 --  (with suitable conversions!)
                let e'    = redfill cxt e
                let taue' = RTStaticExp e'
                -- (old version: no conversions considered -- BAAD!
                --yeses <- mapM (efectiveUnify taue') exps
                yeses <- unifyStaticExps taue' exps sigmas sigma delta
                if not (and yeses)
                 then noSolution forcing (msgNoSolution 5)
                                 deltaSol delta
                 else 
                  do if i>=length tauss 
                      then fail (outboundIn "solveFixS(2)") else do
 --
 -- The tuple types should all be the same
                     let taus' = tauss!!i
                     yeses <- mapM (efectiveUnify taus') tauss
                     if not (and yeses)
                      then noSolution forcing (msgNoSolution 6)
                                      deltaSol delta
                      else 
 --
 -- Finally, the variable t is set to the solution
 --  and the resulting context, simplified.
                       do set "RTVar" residualtype t 
                              (RTFixS taus' taue' sigma)
                          res@(_,_,delta') <- simplifySolving ftvs fsvs delta 
                          when forcing (checkFailure delta')
                          return res

------
-- It picks the first expression that is not a variable
--  together with its index
------
pickStaticExp []           = return Nothing
pickStaticExp ((i,tau):xs) = do tau' <- indir tau
                                case tau' of
                                 RTStaticExp e -> return (Just (i,e))
                                 _             -> pickStaticExp xs

------
-- Collects IsFunS bounds for a given type variable
------
collectRTFunUBs t xs = collectRTFunBs' gimmeRTFunLB t (zip [0..] xs)
collectRTFunLBs t xs = collectRTFunBs' gimmeRTFunUB t (zip [0..] xs)

collectRTFunBs' _      t [] = return []
collectRTFunBs' gimmeB t ((i,d@(_,IsFunS _ _)):idelta) =
  do ibs <- collectRTFunBs' gimmeB t idelta
     let tv = gimmeB d
     tv' <- indir tv
     case tv' of
       RTVar t' -> if t==t'
                    then return (i:ibs) -- It should check for the other bound
                                        --  be a RTTypedS...
                    else return ibs
       _        -> return ibs
collectRTFunBs' gimmeB t (_:idelta) = collectRTFunBs' gimmeB t idelta
  
gimmeRTFunUB (_,IsFunS tau _)   = tau
gimmeRTFunLB (_,IsFunS _ tau)   = tau
gimmeFunType  (RTTypedS _ _ sigma) = sigma
gimmeFunExp   (RTTypedS _ e _)     = e
gimmeFunExp   t = error ("Bad type in gimmeFunExp:" ++ show t)
gimmeFunTuple (RTTypedS ts _ _)    = ts

------
-- Collects IsFixS bounds for a given type variable
------
collectRTFixUBs t xs = collectRTFixBs' gimmeRTFixLB t (zip [0..] xs)
collectRTFixLBs t xs = collectRTFixBs' gimmeRTFixUB t (zip [0..] xs)

collectRTFixBs' _      t [] = return []
collectRTFixBs' gimmeB t ((i,d@(_,IsFixS _ _)):idelta) =
  do ibs <- collectRTFixBs' gimmeB t idelta
     let tv = gimmeB d
     tv' <- indir tv
     case tv' of
       RTVar t' -> if t==t'
                    then return (i:ibs) -- It should check for the other bound
                                        --  be a RTTypedS...
                    else return ibs
       _        -> return ibs
collectRTFixBs' gimmeB t (_:idelta) = collectRTFixBs' gimmeB t idelta
  
gimmeRTFixUB (_,IsFixS tau _)   = tau
gimmeRTFixLB (_,IsFixS _ tau)   = tau
gimmeFixType  (RTFixS _ _ sigma) = sigma
gimmeFixExp   (RTFixS _ e _)     = e
gimmeFixTuple (RTFixS ts _ _)    = ts

------
-- Calculates the right conversions and then attempts to unify all
--  the static expressions in tes with taue
------
unifyStaticExps _ [] [] _ _ = return []
unifyStaticExps taue@(RTStaticExp e_tau) (te:tes) (s:ss) sigma delta =
  do yese <- 
       case te of
        RTVar _ -> --efectiveUnify taue te
          do mcxtp <- isMoreGeneral delta s sigma
             case mcxtp of
              Nothing -> error ("Cannot happen! It should exist a context."++
                                "(Something may be wrong with glb...look uSE1)")
              Just (cxt, _) -> efectiveUnify te (RTInvertCxt cxt e_tau)
        RTStaticExp e ->  
          do mcxtp <- isMoreGeneral delta s sigma
             case mcxtp of
              Nothing -> error ("Cannot happen! It should exist a context."++
                                "(Something may be wrong with glb...look uSE2)")
              Just (cxt, _) -> do
                let e'    = redfill cxt e
                let taue' = RTStaticExp e'
                efectiveUnify taue taue'
        _ -> fail "Cannot happen! Static expression expected."
     yeses <- unifyStaticExps taue tes ss sigma delta
     return (yese:yeses)
unifyStaticExps _ _ _ _ _ = 
  fail "Cannot happen! Have you messed with solveFunS or solveFixS?"

