module SolvingMG where

import MyPrelude
import MoreMonad

import ResidualLang
import ResidualLangPrint
import SpecState
import SpecMonad

import PredsSimplify
import SolvingFailure

------------------------------------------------------------
------------------------------------------------------------
solveMG forcing ftvs fsvs s delta =     
  do (hsm, vsm, deltaNm) <- replaceMorphs s delta
     traceSM ("Context after morph-removal:" ++ show (hsm,vsm)) deltaNm
     ilbs                <- collectRSLBs s deltaNm
     iubs                <- collectRSUBs s deltaNm  
          -- Will this be enough? ws should answer yes...
          -- Recursive constraints are considered in the next stage
     let ubs = map gimmeRSUB (pick iubs deltaNm)
     traceSM "Upper bounds: " ubs
     msigma <- glbs ftvs fsvs ubs
     case msigma of
      Nothing    -> do (hs,vs,delta') <- noSolution forcing (msgNoSolution 7) 
                                                    (pick iubs delta) delta
                       return (hs,vs,delta',CHole)
      Just sigma -> 
       do traceSM "GLB(ubs): " sigma
          set "RSVar" residualscheme s sigma 
          (hs, vs,  delta') <- simplifySolving ftvs fsvs deltaNm
          -- Aca esta la verga fail (show delta')
          when forcing (checkFailure delta')
          (hs',vs')         <- eliminatingEvidence iubs ilbs hs vs delta s
          return (hs',vs',delta',replaceEv hsm vsm)

replaceMorphs s [] = return ([],[],[])
replaceMorphs s (hd@(EVar h,d):ds) =
  do (hsm, vsm, dsNm) <- replaceMorphs s ds
     case d of
       IsMGmorph te sigma s' ->
         do s'' <- indir s'
            case s'' of
              RSVar s''' -> 
                if s == s'''
                 then 

                  case (te,sigma) of
                   (RTStaticExp (RELam hd e), RSForAll tm sm (delta :=> tau)) ->
                    do e'           <- unfoldAll e
                       let fh        = freeHVs e
                       (vd, delta') <- forgetSomeDs fh delta
                       let hd'       = takeEvidence delta'
                       let cxt       = CELam hd' (CEApp CHole vd)
                       h' <- freshevidence
                       return ( h                      : hsm
                              , (EComp h' (ECxt cxt))  : vsm
                              , ((h',IsMG (RSForAll tm sm (delta':=>tau)) 
                                          (RSVar s))   : dsNm)
                              )
                   _ -> fail (canthpnSo ("Where have you find this IsPolym"
                                         ++ ppresidualpredicate d ++"?"))

                 else return (h:hsm, EVar h:vsm, hd:dsNm)
              _ -> return    (h:hsm, EVar h:vsm, hd:dsNm)
       _ -> return           (h:hsm, EVar h:vsm, hd:dsNm)

forgetSomeDs _ [] = return ([], [])
forgetSomeDs fh (hd@(h,d):delta) = 
   do (vd,delta') <- forgetSomeDs fh delta
      if unEVar h `elem` fh
       then return (h:vd, hd:delta')
       else return (evoid:vd, delta')

------
-- It constructs the context to apply to the corresponding term
--  to change the evidence in it (or eliminate it...)
------
eliminatingEvidence :: [Int] -> [Int] -> [EvVble] -> [EvidenceTerm] 
                    -> Preds -> ResSchemeVble
                    -> SpecM ([EvVble],[EvidenceTerm])
eliminatingEvidence iubs ilbs hs vs delta s =
  do yes <- eliminateEvidence
     traceSM "About to eliminate evidence..." yes
     if not yes
      then do traceSM "Nahh!..." ()
              return (hs, vs)
      else do let vubs = pick iubs vs
              traceSM "Doing it!" ()
              let vlbs = pick ilbs vs
              let (vjlbs', vlbs', ilbs') = remDups (1::Int) [] (zip vlbs ilbs)
              let proys = map snd vjlbs'
              traceSM "ilbs=" ilbs
              traceSM "vlbs=" vlbs
              traceSM "ilbs'=" ilbs'
              traceSM "vlbs'=" vlbs'
              traceSM "proys=" proys
              let n = length vlbs'
              let buildTuple h = ETupleS (map (\hi -> EComp hi h) vlbs')
              theLBs <- mapM (\(_,IsMG _ sigma) -> 
                                  case sigma of
                                   RSForAll 0 0 ([]:=>tau) -> unfoldAll tau
                                   _ -> error (canthpnSo ("Schema as a "
                                               ++ "lower bound."))
                               ) 
                               (pick ilbs' delta)
              let nubs = map buildTuple vubs
              let nlbs = map (ithProj n) proys
              let nvs1 = change (zip iubs nubs) vs
              let nvs2 = change (zip ilbs nlbs) nvs1
              set "RSVar" residualscheme s (RSTupleS theLBs)
              return (hs, nvs2)
              
remDups _ vjs [] = (vjs,[],[])
remDups j vjs ((v,i):vis) = 
    case v `lookup` vjs of
     Nothing -> let (vjs', vs', is') = remDups (j+1) ((v,j):vjs) vis
                 in (vjs', v:vs', i:is')
     Just i' -> let (vjs', vs', is') = remDups (j+1) ((v,i'):vjs) vis
                 in (vjs',vs',is')
                           
 
------
-- Collects bounds for a given scheme variable
------
collectRSUBs s xs = collectRSBs' gimmeRSLB s (zip [0..] xs)
collectRSLBs s xs = collectRSBs' gimmeRSUB s (zip [0..] xs)

collectRSBs' _      s [] = return []
collectRSBs' gimmeB s ((i,d):idelta) | isIsMG d =
  do ibs <- collectRSBs' gimmeB s idelta
     let sv = gimmeB d
     sv' <- indir sv
     case sv' of
       RSVar s' -> if s==s'
                    then return (i:ibs)
                    else return ibs
       _        -> return ibs
collectRSBs' gimmeB s (_:idelta) = collectRSBs' gimmeB s idelta
  
gimmeRSUB (_,IsMG        sigma _) = sigma
gimmeRSUB (_,IsMGmorph _ sigma _) = sigma

gimmeRSLB (_,IsMG        _ sigma) = sigma
gimmeRSLB (_,IsMGmorph _ _ sigma) = sigma

isIsMG (_,IsMG _ _)        = True
isIsMG (_,IsMGmorph _ _ _) = True
isIsMG _                   = False

