module SourceDataManagement 
  ( DataManager, initialManager
  , addData, datatypeOf, buildConstructorType
  , allEqual
  , dataconstrs -- Added by Ale

  ) 
 where

import Error
import BasicData
import SourceLang

allEqual [] = False
allEqual [x] = True
allEqual (x:xs@(y:_)) = x==y && allEqual xs

---------------------
-- Data management --
---------------------
type DataManager = ([DataDef], [ConstructorName], [DataName])

initialManager :: DataManager
initialManager = ([],[],[])

addData :: DataManager -> DataDef -> Error DataManager
addData (ddefs, constrs, datas) d@(DataDef _ name _ cs) = 
  let cnames = map constrName cs
      alreadyIn xs ys = filter (flip elem ys) xs
   in if name `elem` datas
       then Error (name ++ " is an already defined datatype.")
       else case (cnames `alreadyIn` constrs) of
             []  -> Success (d:ddefs, cnames++constrs, name:datas)
             c:_ -> Error (c ++ " is a constructor used in some other datatype.")

buildConstructorType ann reps dataName args =
  let datatype = STData ann dataName (map snd reps)
      args' = args `replacedWith` reps
   in foldr (\t app -> STFun ann t app) datatype args'

datatypeOf :: DataManager -> ConstructorName 
              -> Error (DataDef, [SourceType])
datatypeOf (ddefs, constrs, _) constrName = 
  if constrName `elem` constrs
   then Success (dataOf constrName ddefs)
   else Error (constrName ++ " is not a valid constructor.")

-----------------------
--- Begin Alejandro ---
-----------------------
dataconstrs :: DataManager -> DataName -> Error [ConstructorName]
dataconstrs (ddefs, _, datas) d =
   if d `elem` datas 
      then Success (constrsOf d ddefs) 
      else Error (d ++ "datatype is not defined.")
-----------------------
--- End   Alejandro ---
-----------------------


replacedWith args reps = map (replaceIn reps) args

replaceIn reps st@(STVar _) = 
   case lookup st reps of
     Nothing  -> error ("Cannot happen! Free variable "
                       ++"in constructor argument after check...")
     Just nst -> nst
     
replaceIn reps (STTuple a sts)   = STTuple a (sts `replacedWith` reps)
replaceIn reps (STFun a st1 st2) = STFun a (reps `replaceIn` st1)
                                           (reps `replaceIn` st2)
replaceIn reps (STPoly st)       = STPoly (reps `replaceIn` st)
replaceIn reps (STData a n sts)  = STData a n (sts `replacedWith` reps)
replaceIn _    st                = st

