module SourceLangParse(parsesourceexpr) where

import Char
import Parser
import SourceLang
import BasicData

{- Grammar parsed: see grammar.txt -}

-- Parser
parsesourceexpr :: String -> Error SourceTerm
parsesourceexpr str = do lt <- funlexer str
                         parse (do e <- expr'; eof; return e) lt       

--------------------
-- Typed expressions
--------------------
                                                                       
expr' :: Parser (Tokenized String) SourceTerm                          
expr' = do ddef       <- datas                                         
           expression <- expr                                         
           return (Data ddef expression)                             
                                                                    
-- data annotation TypeName1 [params]                              
--     = Constructor11 [types] | .. | Constructor1N [types]       
-- ..                                                            
-- data annotation TypeNameM [params]                           
--     = ConstructorM1 [types] | .. | ConstructorMN [types]    
                                                              
datas :: Parser (Tokenized String) [DataDef]                 
datas = (do ddef <- singleData `sepBy` key ";"              
            key ";"                                        
            return ddef                                   
        ) `option` []                                    
                                                        
singleData :: Parser (Tokenized String) DataDef        
singleData = do                                       
                key "data"                           
                name <- typeName                    
                ann  <- annot                      
                vars <- dataParams                
                key "="                          
                constDefs <- constructors       
                return (DataDef ann name vars constDefs)               
                                                                      
                                                                     
typeName :: Parser (Tokenized String) String                        
typeName = uppercaseName "Data"                                    
                                                                  
constructorName :: Parser (Tokenized String) String         
constructorName = uppercaseName "Constructor"                
                                                              
                                                               
dataParams :: Parser (Tokenized String) [SourceType]            
dataParams = many typeVar                                        
                                                                  
constructors :: Parser (Tokenized String) [Constructor SourceType]     
constructors = do                                                     
                 cs <- constructor `sepBy` key "|"                   
                 case cs of                                         
                        [] -> fail "Constructor expected!"         
                        otherwise -> return cs                    
                                                                 
constructor :: Parser (Tokenized String) (Constructor SourceType)  
constructor = do                                                 
                name <- constructorName                         
                types <- many typeargfact                      
                return (Constructor name types)              
                                                            
                                                                       
texp :: Parser (Tokenized String) SourceTerm
texp = do e <- expr; t <- typesig
          case t of
            STUnknown -> return e
            _         -> return (Typed e t)

expr :: Parser (Tokenized String) SourceTerm
expr = 
  munchers
  `orelse`
  bexp

bexp  = rassoc (do key "||"; ann <- annot; return (BoolOp ann Or)) bterm
bterm = rassoc (do key "&&"; ann <- annot; return (BoolOp ann And)) bfact
bfact = noassoc rop nexp
nexp  = lassoc nop nterm
nterm = lassoc (do key "*"; ann <- annot; return (ArithOp ann Mul)) nfact
nfact = lassoc (do key "^"; ann <- annot; return (ArithOp ann Exp)) fexp

fexp :: Parser (Tokenized String) SourceTerm
fexp = lassoc (do key "@"; ann <- annot; return (App ann STUnknown)) prim

prim :: Parser (Tokenized String) SourceTerm
prim = 
  (do key "poly"; e <- aexp; return (Poly e))
  `orelse`
  (do key "spec"; e <- aexp; return (Spec e))
  `orelse`
  (do key "polym"; e <- aexp; return (Polymorph e))
  `orelse`
  (do key "lift"; e <- aexp; return (Lift e))
  `orelse`
  (do key "fix"; ann <- annot; e <- aexp; return (Fix ann e))
  `orelse`
  (proj $> aexp)
  `orelse`
  aexp

aexp :: Parser (Tokenized String) SourceTerm
aexp = 
  (return Var $> kind Ident)
  `orelse`
  (do key "()"; ann <- annot; return (Tuple ann []))
  `orelse`
  liter
  `orelse`
  caseStatement
  `orelse`
  (do me <- paren aexpTupleCont
      case me of
        Left e   -> return e
        Right fe -> return fe $> annot
  )

aexpTupleCont :: Parser (Tokenized String) 
                        (Either SourceTerm (Annot -> SourceTerm))
aexpTupleCont = 
  do ts <- texp `sepBy` key ","
     case ts of
      []  -> fail "Expression expected!"
      [e] -> return (Left e)
      _   -> return (Right (\ann -> Tuple ann ts))

munchers =
  (do key "\\"; ann <- annot; i <- kind Ident; t <- typesig
      key "->"; e <- expr
      return (Lam ann i t e)
  )
  `orelse`
  (do key "let"; ann <- annot; ds <- decls
      key "in"; e <- expr
      return (Let ann ds e)
  )
  `orelse`
  (do key "if"; ann <- annot; e <- expr
      key "then"; e1 <- expr
      key "else"; e2 <- expr
      return (If ann e e1 e2)
  )

caseStatement =
  do key "case"; ann <- annot; e <- expr; key "of"                
     xs <- some casestat                                         
     key "esac"                                                 
     return (Case ann e STUnknown xs)                          
                                                              
-----------
--begin @--
-----------
casestat :: Parser (Tokenized String) ((SourceTerm, SourceType), SourceTerm)
casestat = do p <- pattern
              key "->" 
              e <- expr 
              key ";"
              return ((p,STUnknown),e)
                  
pattern :: Parser (Tokenized String) SourceTerm
pattern = do c <- construc
             patternarguments c
          `orelse`
          patternfact

patternfact :: Parser (Tokenized String) SourceTerm
patternfact = liter `orelse` variable `orelse` paren pattern

patternarguments :: SourceTerm -> Parser (Tokenized String) SourceTerm 
patternarguments s = 
  (do key "@"                                  
      ann <- annot                              
      arg <- patternfact
      patternarguments (App ann STUnknown s arg))
  `option` s                             
---------
--end @--
---------

--------------------
-- Type signatures
--------------------
typesig :: Parser (Tokenized String) SourceType
typesig = (do key "::"; t <- typeexp; return t)
          `orelse`
          (return STUnknown)
            
typeexp :: Parser (Tokenized String) SourceType
typeexp = rassoc (do key "->"; ann <- annot; return (STFun ann))
                 typefact

typefact :: Parser (Tokenized String) SourceType
typefact = typeprefact `orelse` dataType

typeargfact = typeprefact `orelse` argdataType

typeprefact :: Parser (Tokenized String) SourceType
typeprefact =                   
  (do key "Int"; ann <- annot; return (sTypeInt ann))
  `orelse`
  (do key "Bool"; ann <- annot; return (sTypeBool ann))
  `orelse`
  (do key "poly"; t <- typefact; return (STPoly t))
  `orelse`
  (do key "()"; ann <- annot; return (STTuple ann []))
  `orelse`
  (do mt <- paren typefactTupleCont
      case mt of
        Left t   -> return t
        Right ft -> return ft $> annot
  )
  `orelse`
  typeVar      

typefactTupleCont :: Parser (Tokenized String) 
                            (Either SourceType (Annot -> SourceType))
typefactTupleCont = 
  (do ts <- typeexp `sepBy` key ","
      case ts of
       []  -> fail "Type expected!"
       [t] -> return (Left t)
       _   -> return (Right (\ann -> STTuple ann ts))
  )

--------------------
-- Auxiliaries
--------------------
annot :: Parser (Tokenized String) Annot
annot = (do key tokenStatic; return Static)
        `orelse` 
        (do key tokenDynamic; return Dynamic)
        `orelse` 
        (return (AVar 0))
        
checkAnnot :: Annot -> Parser (String,Token) ()
checkAnnot a = do a' <- annot
                  if a == a' then return () else failP

rop = anyOf (\(sop,r) -> (do key sop; ann <- annot
                             return (RelOp ann r STUnknown)
            ))
            [ ("==", Eq), ("/=", Neq)
            , ("<",  Lt), ("<=", Leq)
            , (">",  Gt), (">=", Geq)
            ]

nop = anyOf (\(sop,n) -> (do key sop; ann <- annot; return (ArithOp ann n)))
            [ ("+", Add), ("-",Sub)
            ]

proj = (do key "fst"; ann <- annot; return (Proj ann 1 2 STUnknown))
       `orelse`
       (do key "snd"; ann <- annot; return (Proj ann 2 2 STUnknown))
       `orelse`
       (do t <- kind (Other "TagId")
           let pruins = span (/='_') t
           case pruins of
             ("Proj", '_':ins) -> 
               do let iun = span (/='_') ins
                  case iun of
                   (si, '_':sn) -> 
                     if (all isDigit si) && (all isDigit sn)
                      then do let i = read si
                              let n = read sn
                              if i>n || i==0 then fail "Bad projection!" else do
                              ann <- annot
                              return (Proj ann i n STUnknown)
                      else fail "Bad projection!"
                   _          -> fail "Bad projection!"
             _       -> fail "Bad projection!"
       )

liter :: Parser (Tokenized String) SourceTerm
liter = (do n <- num; ann <- annot; return (Int ann n))
        `orelse` 
        (do b <- bool; ann <- annot; return (Bool ann b))
        `orelse`                                        
        construc                                       
                                                      
construc = do                                        
              name <- uppercaseName "Constructor"   
              ann  <- annot                        
              return (Construc ann name [])       

decls = decl `sepBy` key ";"
decl  = do x <- kind Ident; t <- typesig; key "="; e <- expr
           return ((x, t), e)

num :: Parser (Tokenized String) Int
num = return read $> kind Numeral

bool :: Parser (Tokenized String) Bool
bool = (do key "True"; return True)
       `orelse`
       (do key "False"; return False)

uppercaseName :: String -> Parser (Tokenized String) String            
uppercaseName msg = do                                                
                      t <- kind (Other "TagId")                      
                      if isUpper (head t)                           
                         then return t                             
                         else fail (msg ++ " names are uppercase!") 
                                                         
typeVar :: Parser (Tokenized String) SourceType           
typeVar = do                                               
            t <- kind (Other "TagId")                       
            case t of                                        
                 'T':'_':ns -> return (STVar (read ns))       
                 _          -> failP                           
                                                                
variable = do v <- (kind Ident); return (Var v)                  
                                                                  
dataType :: Parser (Tokenized String) SourceType                   
------------
--begin @F--
------------
argdataType = do t   <- typeName
                 ann <- annot
                 return (STData ann t [])
              `orelse`
              paren dataType

dataType = do t    <- typeName
              ann  <- annot
              opws <- many (return addArgument <*> typefact)
              return (appRev (foldl (\u (oper,w)->u `oper` w) 
                     (STData ann t []) opws))
     where addArgument (STData ann t tes) te = STData ann t (te:tes)
           appRev (STData ann t tes) = STData ann t (reverse tes)
----------
--end @F--
----------

paren p = do key "("; v <- p; key ")"; return v

--------------------
-- Lexical analyizer
--------------------
funlexer :: String -> Error (TokenList String)
funlexer = lexer symbols keywords comment procIdents

symbols :: [String]
symbols = [tokenStatic, tokenDynamic, "::", "()", "\\", "->" ]
          ++ [ "==", "/=", "<=", ">=", "&&", "||" ]
          ++ map (:[]) "=()@,;{}^><+-*|"                        

keywords :: [String]
keywords = words ("let in fix if then else poly spec polym lift fst snd data case esac of"
                 ++ " Int Bool True False")

comment :: Parser Char String
comment = (string "--" <* skipto (literal '\n'))
           `orelse`
          (string "{-" <* skipto (string "-}"))
          
procIdents :: String -> Maybe String
procIdents w = if isUpper (head w) 
                then Just "TagId" 
                else Nothing

