module SourceUnification
  (unify,unifyAll,unifyAnnot)      
 where

import SourceLang
import SourceLangPrint
import SourceTypeCheckMonad

-- --------------------------------------------------------
-- Unification of source types
-- --------------------------------------------------------

unifyAll []  = return ()
unifyAll [x] = return ()
unifyAll (y:x:xs) = do unify x y
                       unifyAll (x:xs)
                       return ()


unify t t' = unify' [] [] t t'

unify'  eqs  seqs t 
                  t'                 = if (t,t') `elem` eqs 
                                        then return ()
                                        else do t1  <- indir t
                                                t1' <- indir t'
                                                unify'' ((t,t'):eqs) seqs t1 t1'
unify'' eqs seqs vw@(STVar w) 
                 vx@(STVar x)        = case compare w x of
                                         EQ -> return ()
                                         LT -> set stype x vw
                                         GT -> set stype w vx
unify'' _   _    (STVar w)    
                 t                   = occurCheck w t
unify'' _   _    t            
                 (STVar w)           = occurCheck w t
unify'' _   _    t1@(STBase an1 b1) 
                 t2@(STBase an2 b2)
                                     = do yes <- unifyAnnot an1 an2
                                          if not yes then uniferr t1 t2 else do
                                          yes <- unifySTBase b1 b2
                                          if not yes then uniferr t1 t2 else do
                                          return ()
unify'' eqs seqs t1@(STTuple an1 as) 
                 t2@(STTuple an2 bs) = do yes <- unifyAnnot an1 an2
                                          if not yes then uniferr t1 t2 else do
                                          yes <- unifys eqs seqs as bs
                                          if not yes then uniferr t1 t2 else do
                                          return ()
unify'' eqs seqs t1@(STFun an1 a b)  
                 t2@(STFun an2 c d)  = do yes <- unifyAnnot an1 an2
                                          if not yes then uniferr t1 t2 else do
                                          unify' eqs seqs a c
                                          unify' eqs seqs b d
unify'' eqs seqs t1@(STData an1 n1 xs)                                               
                 t2@(STData an2 n2 ys) = if (n1 /= n2) then uniferr t1 t2 else do    
                                             yes <- unifyAnnot an1 an2               
                                             if not yes then uniferr t1 t2 else do   
                                             yes <- unifys eqs seqs xs ys            
                                             if not yes then uniferr t1 t2 else do   
                                             return ()                               

unify'' eqs seqs (STPoly a)   
                 (STPoly b)          = unify' eqs seqs a b
unify'' _   _    a            
                 b                   = uniferr a b

-- --------------------------------------------------------
-- Auxiliaries
-- --------------------------------------------------------
unifys _   _    []     []     = return True
unifys eqs seqs (a:as) (b:bs) = do unify' eqs seqs a b
                                   unifys eqs seqs as bs
unifys _   _    _      _      = return False

occurCheck w t@STUnknown = uniferr (STVar w) t
occurCheck w t = do t' <- unfoldV False t
                    if w `elem` freeSTV t'
                     then uniferr (STVar w) t
                     else set stype w t

uniferr a b = do a' <- unfoldV False a
                 b' <- unfoldV False b
                 fail ("In 2-level type checking: cannot unify\n  "
                       ++ppsourcetype a'++"\nand\n  "++ppsourcetype b')

-- --------------------------------------------------------
-- Unification of annotations
-- --------------------------------------------------------
unifyAnnot   ann1      ann2      = do ann1' <- indir ann1
                                      ann2' <- indir ann2
                                      unifyAnnot' ann1' ann2'
unifyAnnot'  vi@(AVar i)  
             vj@(AVar j)         = case compare i j of
                                    EQ -> return True
                                    LT -> do set annot j vi; return True
                                    GT -> do set annot i vj; return True
unifyAnnot'  (AVar i)  ann       = do set annot i ann; return True
unifyAnnot'  ann       (AVar i)  = do set annot i ann; return True
unifyAnnot'  ann1      ann2      = return (ann1 == ann2)

-- --------------------------------------------------------
-- Unification of base types
-- --------------------------------------------------------
unifySTBase  b1        b2        = do b1' <- indir b1
                                      b2' <- indir b2
                                      unifySTBase' b1' b2' 
unifySTBase' vi@(SBVar i) 
             vj@(SBVar j)        = case compare i j of
                                    EQ -> return True
                                    LT -> do set basetype j vi; return True
                                    GT -> do set basetype i vj; return True
unifySTBase' (SBVar i) b         = do set basetype i b; return True
unifySTBase' b         (SBVar i) = do set basetype i b; return True
unifySTBase' b1        b2        = return (b1 == b2)
