module SpecAuxRels where

import List((\\), union)
import MyPrelude
import MoreMonad (liftM,liftM2)

import SourceLang
import ResidualLang
import SpecMonad
import ResidualGeneralizable
import ResidualUnification
import ResidualLangFVs
import BasicData(ConstructorName)

infix 4 :~>

-- ------------------------------------
-- Source-Residual relation algorithm
-- ------------------------------------
srw :: SourceType -> SpecM (Preds, ResidualType)
srw st = do st' <- unfoldAll st
            srw' False st'

srw_for_datadef st = do st' <- unfoldAll st
                        srw' True st'

srw' :: Bool -> SourceType -> SpecM (Preds, ResidualType)
srw' p (STBase Static SBInt)   = do t <- freshresidualtype
                                    h <- freshevidence
                                    return ([(h, IsInt t)], t)
srw' p (STBase Dynamic SBInt)  = return ([], RTIntD)
srw' p (STBase Static SBBool)  = do t <- freshresidualtype
                                    h <- freshevidence
                                    return ([(h, IsBool t)], t)
srw' p (STBase Dynamic SBBool) = return ([], RTBoolD)
srw' p (STFun Static  t1 t2)   = do (delta, t') <- srw' p (STFun Dynamic t1 t2)
                                    (sigma,_,_) <- gen emptyAssgn (delta:=>t')
                                    te  <- freshresidualtype
                                    tts <- freshresidualtype
                                    t   <- freshresidualtype
                                    h   <- freshevidence
                                    let tau = RTTypedS tts te sigma
                                    return ([(h, IsFunS tau t)], t)
srw' p (STFun Dynamic t1 t2)   = do (delta1, t1') <- srw' p t1
                                    (delta2, t2') <- srw' p t2
                                    return (delta1 ++ delta2, RTFun t1' t2')
srw' p (STTuple Static ts)     = do (delta, ts') <- srw's p ts
                                    return (delta, RTTupleS ts')
srw' p (STTuple Dynamic ts)    = do (delta, ts') <- srw's p ts
                                    return (delta, RTTuple ts')
srw' p (STPoly t)              = do (delta, t') <- srw' p t
                                    s           <- freshresidualscheme
                                    h           <- freshevidence
                                    (sigma,_,_) <- gen emptyAssgn (delta :=> t')
                                    return ([(h, IsMG sigma s)], RTPoly s)
srw' p sdt@(STData Static _ _) = do t <- freshresidualtype
                                    h <- freshevidence
                                    return ([(h, IsConstrOf sdt t)], t)
---------------------
-- Begin Alejandro --
{-
We add here my WSR rules, which is different from the previosly put 
here. The old code is

srw' p (STData Dynamic d _)    = do (ps,dt) <- dataNamed d
                                    return (ps,dt)
-}

srw' p (STData Dynamic d _) = do cs_st <- constrsOfSourceData d
                                 psts <- srConstrs p cs_st
                                 t   <- freshresidualtype
                                 let cs = map fst cs_st 
                                 let cname_psts = zip cs psts
                                 let tvars = map snd psts
                                 ps  <- guardPreds t cname_psts 
                                 (h, delta) <- entlw ps 
                                                  (ConstrsOf t)
                                 return (ps++delta, t)
{-
                                 return (ps, RTDData t (RTTuple tvars))
-}

srw' p (STVar t)               = if p then return ([],RTVar t)
                                      else error ("The source language is "++
                                                  "a monomorphic one!")
srw' _ t                       = error ("Cannot happen! Have you " ++
                                        "called srw' directly with argument " ++
                                        show t ++ "?")

srw's p [] = return ([],[])
srw's p (t:ts) = do (deltat,  t')  <- srw' p t
                    (deltats, ts') <- srw's p ts
                    return (deltat++deltats, t':ts')

--------------------
-- Begin Alejadro --
--------------------
-- Doing WSR for types that are taken by constructors.
-- I only considered constructor with arity 1
srConstrs :: Bool -> [(ConstructorName, SourceType)] 
             -> SpecM [(Preds,ResidualType)] 
srConstrs p [] = return []
srConstrs p ((_, (STFun Dynamic tp _)):psts)  
       =  do (ps, rt) <- srw' p tp  
             others   <- srConstrs p psts
             return ((ps,rt):others)

------------------------------------------------------------
-- For puting predicates inside a Dinamic Data Type Guard form and 
-- also add HasC predicate 
guardPreds :: ResidualType -> [(ConstructorName, (Preds, ResidualType))] 
              -> SpecM Preds 
guardPreds t []         = return []
guardPreds t ((cname, (ps, rt)):rs) =
       do others  <- guardPreds t rs 
          let guarded_preds = guardDinDataPreds t cname ps   
          h <- freshevidence 
          let hasc = guardDinDataPreds t cname [(h, HasC t cname rt)]
          return (hasc++guarded_preds++others)

guardDinDataPreds t cname ps = 
    map g ps 
    where g (h,p) =  (h, GuardedDinData cname t p)

------------------------------------------------------------
-- Taking constructor ST and its source type
-- C @ x::tau ---> (x, tau)
takeConstrFromPattern (App _ _ (Lam _ _ tau _) (Var x)) = return (x, tau)
takeConstrFromPattern _ = error ("Dynamic Data Types" ++
                             " in a case branch has to have the form C @ x"
                             ++ "(arithy 1)")

takeConstrSTsAndBranchs [] = return []
takeConstrSTsAndBranchs (((tpattern, patterntype), texps):brs) =
       do (x, tau) <- takeConstrFromPattern tpattern
          others <- takeConstrSTsAndBranchs brs
          return ((x,tau,texps):others)

------------------------------------------------------------
-- Taking constructor names from a branch
-- C @ x::tau ---> C 
takeConstrNameFromPattern (App _ _ (Lam _ _ _ e) _) = 
            case e of 
                 (Construc _ cname _) -> return cname 
                 _ -> error "bad application of dynamic constructor!"   


takeConstrsFromBranchs [] = return []
takeConstrsFromBranchs (((tpattern, patterntype), texp):brs) = 
      do cname <- takeConstrNameFromPattern tpattern 
         others <- takeConstrsFromBranchs brs 
         return (cname:others)

------------------------------------------------------------
-- Constructing Dinamic Lambdas for using in Dinamic Case Rule
-- C @ x::tau1 -> texp::tau2 
-- 
-- (x, tau1, texp) tau2 ---> \x->texpr :: tau1 -> tau2
buildLamsForDinCaseSpec [] tau = return []
buildLamsForDinCaseSpec ((x, tau1, texp):others) tau2 =
   do exp <- buildLamForDinCaseSpec (x, tau1, texp) tau2
      exps <- buildLamsForDinCaseSpec others tau2
      return (exp:exps)
    
buildLamForDinCaseSpec (x, tau1, expr) tau2 =
	return (Lam Dynamic x tau1 expr, STFun Dynamic tau1 tau2)   

------------------------------------------------------------
-- Taking evidence and constructor of a HasC
-- (h,HasC t1 C t2) ---> (h, C)
takeEvidenceConstrNameHasC :: (EvTerm, Pred) -> [(EvTerm,ConstructorName)]
takeEvidenceConstrNameHasC (h,GuardedDinData _ _ (HasC _ cname _)) = [(h,cname)]
takeEvidenceConstrNameHasC  _ = []

takeEvidenceConstrNameHasCs :: [(EvTerm, Pred)] 
                               -> [(EvTerm,ConstructorName)]
takeEvidenceConstrNameHasCs [] = []
takeEvidenceConstrNameHasCs (p:ps) = 
   ls ++ others 
   where ls = takeEvidenceConstrNameHasC p
         others = takeEvidenceConstrNameHasCs ps 


------------------------------------------------------------
-- Taking residual variable and body of the residual dynamic 

builtResidualBranch :: ((EvTerm, ConstructorName), ResTerm) 
                       -> (ResTerm, ResTerm)
builtResidualBranch ((h,cname), RLam x' e') =
      (RDConstruc cname (evidence2residual h) [RVar x'], e')


builtResidualBranchs :: [((EvTerm, ConstructorName), ResTerm)] 
                       -> [(ResTerm, ResTerm)]
builtResidualBranchs [] = []
builtResidualBranchs (b:bs) = 
   rbr:others 
   where rbr = builtResidualBranch b 
         others = builtResidualBranchs bs

--------------------
-- End Alejadro --
--------------------

checkSRfor [] [] = return ([],True)
checkSRfor (rt:rts) (st:sts) =
  do (ps,rt')   <- srw st
     yes        <- efectiveUnify rt rt'
     (pss,yes2) <- checkSRfor rts sts
     return (ps:pss, yes&&yes2)
checkSRfor _  _  = fail (canthpnSo "Have you called checkSRfor directly?")

-- ------------------------------------
-- Algorithm for entailment used in the spec algorithm
-- ------------------------------------
entlw :: Preds -> Pred -> SpecM (EvidenceTerm, Preds)
entlw [] d = do h <- freshevidence
                return (h, [(h,d)])
entlw ((h,d):delta) dr =
  do d'  <- unfoldAll d
     dr' <- unfoldAll dr
     if (d'==dr')
      then return (h, [])
      else entlw delta dr

entlsw delta [] = return ([],[])
entlsw delta (d:ds) = 
  do (v,d') <- entlw delta d
     (vs,ds') <- entlsw (delta++d') ds
     return (v:vs,d'++ds')

-- ------------------------------------
-- Assignments
-- ------------------------------------
newtype Assignment = Assgn [ ((SourceTermVble, SourceType), TypedResidualTerm) ]
                   deriving Show

emptyAssgn :: Assignment
emptyAssgn = Assgn []

extendedWith :: Assignment 
                -> ((SourceTermVble, SourceType), TypedResidualTerm) 
                -> Assignment
extendedWith (Assgn as) source_residual = Assgn (source_residual:as)

replaceWith (Assgn as) source_residual@((x,_),_) = 
  Assgn (source_residual:filter (\((x',_),_) -> x'/=x) as)

multiextendedWith g srs = foldl (\g' xt_xt' -> g' `extendedWith` xt_xt') g srs

lookupFor :: Assignment -> (SourceTermVble, SourceType) 
             -> Maybe TypedResidualTerm
lookupFor (Assgn xs) sourcevar = sourcevar `lookup` xs

filterAssgn :: Assignment -> [String] 
             -> [((SourceTermVble,SourceType),TypedResidualTerm)]
filterAssgn (Assgn xs) ys = filter (\((x,_),_) -> x `elem` ys) xs

instance Generalizable SourceType
instance Generalizable Assignment where
  gr tf tg sf sg (Assgn assgn) = do (assgn',tv,sv) <- gr tf tg sf sg assgn
                                    return (Assgn assgn',tv,sv)
  
gen :: Assignment -> QualifiedType 
       -> SpecM (ResidualTypeScheme, [ResTypeVble], [ResSchemeVble])
gen g qt = do qtvs <- freeRTV qt
              qsvs <- freeRSV qt
              genThese g qtvs qsvs qt

genThese :: Assignment 
       -> [ResTypeVble] -> [ResSchemeVble]
       -> QualifiedType 
       -> SpecM (ResidualTypeScheme, [ResTypeVble], [ResSchemeVble])
genThese g qtvs qsvs qt@(preds:=>_) = 
   do gtvs        <- freeRTV g
      gsvs        <- freeRSV g
      fs          <- fundeps preds
      let gtvsclos = closure gtvs fs
      let tvs      = qtvs \\ gtvsclos
      let tvsclos  = closure tvs (proper fs)
      let svs      = qsvs \\ gsvs
      sigma       <- generalize tvsclos svs qt
      return (sigma, tvsclos, svs)

constructScheme g delta tau e = do let h  = takeEvidence delta
                                   let e' = RELam h e
                                   tau' <- unfoldAll tau
                                   (sigma, _, _) <- gen g (delta:=>tau')
                                   return (e', sigma)

---------------------------------
-- Instances for Assignments
instance MetaVbles  SourceType
instance Unfoldable SourceType
instance MetaVbles  Assignment
instance Unfoldable Assignment where
  unfoldVbles (Assgn gs) = liftM Assgn (unfoldVbles gs)
  unfoldAll'  (Assgn gs) = liftM Assgn (unfoldAll' gs)
instance FRV Assignment 

-- ------------------------------------
-- Functional dependencies
-- ------------------------------------
data FunDep = [ResTypeVble] :~> [ResTypeVble]
            -- | FDGuarded [ResTypeVble] FunDep FunDep
            deriving (Eq, Show)

fundeps [] = return []
fundeps ((_,d):ds) = do fsd  <- singlefundeps d
                        fsds <- fundeps ds
                        return (fsd `union` fsds)

proper :: [FunDep] -> [FunDep]
proper = filter (\(ps:~>_) -> not (null ps))

singlefundeps (IsIf taur taub tau1 tau2) =
    do tv1 <- freeRTV tau1
       tv2 <- freeRTV tau2
       tvb <- freeRTV taub
       tvr <- freeRTV taur
       let fd = tvb++tv1++tv2 :~> tvr
       -- let gfd = FDGuarded tvb (tv1 :~> tvr) (tv2 :~> tvr)
       return [fd]
singlefundeps (IsCase taur taue cts) =
    do tv1 <- freeRTV cts
       tve <- freeRTV taue
       tvr <- freeRTV taur
       let fd = tve++tv1 :~> tvr
       return [fd]
singlefundeps (IsArithOp _ tau tau1 tau2) =
    do tv1 <- freeRTV tau1
       tv2 <- freeRTV tau2
       tv  <- freeRTV tau
       let fd = tv1++tv2 :~> tv
       return [fd]
singlefundeps (IsBoolOp  _ tau tau1 tau2) =
    do tv1 <- freeRTV tau1
       tv2 <- freeRTV tau2
       tv  <- freeRTV tau
       let fd = tv1++tv2 :~> tv
       return [fd]
singlefundeps (IsRelOp   _ tau tau1 tau2) =
    do tv1 <- freeRTV tau1
       tv2 <- freeRTV tau2
       tv  <- freeRTV tau
       let fd = tv1++tv2 :~> tv
       return [fd]

{-
singlefundeps (HasC tau cname tau') =
    do tv  <- freeRTV tau
       tv' <- freeRTV tau'
       let fd = tv' :~> tv 
       return [fd]

singlefundeps (GuardedDinData cname tau pred) =
    do tvp <- freeRTV pred 
       tv  <- freeRTV tau
       let fd = tvp :~> tv 
       return [fd]
-}


{-
singlefundeps (IsFunS tau tau') = 
    do tv  <- freeRTV tau
       tv' <- freeRTV tau'
       --let vs = tv `union` tv'
       --return [[] :~> vs]
       return [tv :~> tv']
singlefundeps (IsFixS tau tau') = 
    do tv  <- freeRTV tau
       tv' <- freeRTV tau'
       --let vs = tv `union` tv'
       --return [[] :~> vs]
       return [tv :~> tv']
-}
singlefundeps _ = return []

closure,onepass :: [ResTypeVble] -> [FunDep] -> [ResTypeVble]
closure v fs = let vclos = onepass v fs
                in if v==vclos
                    then v
                    else closure vclos fs

onepass v [] = v
onepass v fs@((x:~>y):fs') = if x `subseteq` v
                              then onepass (v `union` y) fs'
                              else onepass v fs'

--------------------------------------------
-- Doesn't work: do not bind these variables
nonBoundableRTV tau = do tau' <- unfoldAll tau
                         nbRTV tau'
nbRTV (RTVar i) = return [i]
nbRTV (RTTypedS taut taue sigma) = 
  do tv <- nbRTV taut
     ev <- nbRTV taue
     return (tv `union` ev)
nbRTV (RTFixS taut taue sigma) = 
  do tv <- nbRTV taut
     ev <- nbRTV taue
     return (tv `union` ev)
nbRTV _ = return []

