module Spec where

import List(union,(\\))

import MyPrelude
import Error
import MoreMonad

import BasicData
import SourceLang
import SourceLangPrint
import qualified SourceTypeCheckMonad
import ResidualLang
import ResidualLangPrint
import SpecState
import SpecMonad
import ResidualReduction
import SpecAuxRels
import PredsSimplify
import PredsSolving
import ResidualUnification
import ResidualLangFVs

instance Unfoldable SourceTerm

specialisesource :: SourceTypeCheckMonad.SourceState 
                    -> Bool -> Bool -> Bool
                    -> TypedSourceTerm 
                    -> Error (ResidualTerm, ResidualTypeScheme)
specialisesource source_state solveCxts eliminateEv raiseArity 
                 (Data ddefs tst, st) = 
  let f = do if solveCxts   then onSolveContextsFlag else offSolveContextsFlag
             if eliminateEv then onEvidenceElimFlag  else offEvidenceElimFlag
             if solveCxts   then onTracing else offTracing
             setSourceTCstate source_state
             exceptVblesUsedIn ddefs
             ddefs'             <- specDatas ddefs
             (delta, (e',tau')) <- specw emptyAssgn (tst,st)
             (delta',  e_2')    <- simplifyAndSolve True
                                                    emptyAssgn delta tau' e'
             (e_3',  tau'')     <- return (e_2', tau') --arityRaise raiseArity e_2' tau'
             (e_w',  sigma_w)   <- constructScheme emptyAssgn delta' tau'' e_3'
             e_w''              <- unfoldAll e_w'
             sigma_w'           <- unfoldAll sigma_w 
             --stopTrace 
             return (RData ddefs' (ered e_w''), sigma_w')
   in case (runSM f) of
        Error e            -> Error e
        Success (re_rt, _) -> Success re_rt

simplifyAndSolve forcing g delta tau' e' = 
   do (ftvs,  fsvs)      <- freeRTVs g tau'
      (cxt,   delta')    <- simplify ftvs fsvs delta
      (ftvs', fsvs')     <- freeRTVs g tau'
      e''                <- unfoldVbles e'
      let e_r1            = redfill cxt e''
      --DEBUG
     {- 
      yes <- solveContexts
      when (yes && forcing)
       (do traceSMraw "(e'',tau'):" (e'',tau') 
           traceSMraw "(ftvs,fsvs):" (ftvs,fsvs)
       )
     -}  
      --DEBUG
      (cxt',  delta'')   <- solve forcing ftvs' fsvs' delta'
      let e_r             = redfill cxt' e_r1
      --let e_r             = noredfill cxt' (noredfill cxt e'')
      return (delta'', e_r)
             
exceptVblesUsedIn ddefs = 
  let stvbles = concat (map datadefParams ddefs)
      rtvbles = map (\(STVar i) -> i) stvbles
   in setexcepts residualtype rtvbles

-------------------------------------------------------------------
-- Specialisation algorithm
-------------------------------------------------------------------
specw :: Assignment -> TypedSourceTerm -> SpecM (Preds, TypedResidualTerm)

specw g (xv@(Var x), tau) =
   case g `lookupFor` (x,tau) of
    Nothing  -> fail ("Unbound variable:" ++ ppsourceexpr xv 
                                  ++ "::" ++ ppsourcetype tau)
    Just et' -> return ([], et')

------------
-- Functions
specw g (App Static tau2 e1 e2, tau1) =
  specwAppS g [e2] [tau2] tau1 (e1, STFun Static tau2 tau1)
 --  do (delta1, (e1', tau21')) <- specw g (e1, STFun Static tau2 tau1)
 --     (delta2, (e2', tau2'))  <- specw g (e2, tau2)
 --     (delta3, tau1')         <- srw tau1
 --     taue'                   <- freshresidualtype
 --     taus'                   <- freshresidualtype
 --     let delta                = delta1 ++ delta2 ++ delta3
 --     (v, delta')             <- entlw delta
 --                                      (IsFunS tau21'                                          
 --                                              (RTTypedS taus' taue' 
 --                                                (scheme (RTFun tau2' tau1')))
 --                                      )     
 --     let e' = REAppS (REAppS (evidence2residual v) e1') e2'
 --     return (delta++delta', (e', tau1'))

specw g (App Dynamic tau2 e1 e2, tau1) = 
   do (delta1, (e1', tau2tau1')) <- specw g (e1, STFun Dynamic tau2 tau1)
      (delta2, (e2', tau2''))    <- specw g (e2, tau2)
      tau'                       <- freshresidualtype
      (_, delta')                <- entlw (delta1++delta2) 
                                          (Unify tau2tau1' (RTFun tau2'' tau'))
      return (delta1++delta2++delta', (RApp e1' e2', tau'))
                                
specw g (f@(Lam Static x t e1), STFun Static tau2 tau1) =  
   do let ys              = freeSV f
      let ys'             = filterAssgn g ys
      let taus'           = map (snd . snd) ys'
      sf                 <- freshresidualtermBasedOn "sf"
      let ((eys, newys),tauys') 
                          = ys' `replacedWithProjectionsFor` (RVar sf)
      let newg            = foldl replaceWith g newys
      (delta, (e',tau')) <- specwLamS newg [(x,t)] [tau2] (e1,tau1)
      --(delta, (e',tau')) <- specw newg (Lam Dynamic x t e1
      --                                 ,STFun Dynamic tau2 tau1)
      (delta', e'')      <- simplifyAndSolve False newg delta 
                                             (RTFun tauys' tau') 
                                             (RLam sf e')
      (e''',sigma)       <- constructScheme newg delta' tau' e''
      tau                <- freshresidualtype
      h                  <- freshevidence
      let d              = IsFunS (RTTypedS (RTTupleS taus') 
                                            (RTStaticExp e''') sigma
                                  ) tau
      return ([(h,d)],(RTupleS eys, tau))

  where replacedWithProjectionsFor ys f = 
            let taus = map (snd . snd) ys
                tauys = RTTupleS taus
             in (repProj 1 (length ys) ys f tauys, tauys)
        repProj _ _ [] _ _ = ([],[])
        repProj i n ((tst, (ei',taui')):ys') f tau = 
             let (eys', newys') = repProj (i+1) n ys' f tau
              in (ei' : eys', (tst, (RProjS i n f, taui')) : newys')

specw g (Lam Dynamic x _ e, STFun Dynamic tau2 tau1) = 
   do (delta, tau2')      <- srw tau2
      x'                  <- freshresidualtermBasedOn x
      (delta',(e',tau1')) <- specw (g `extendedWith` ((x,tau2),(RVar x',tau2')))
                                   (e, tau1)
      return (delta++delta', (RLam x' e', RTFun tau2' tau1'))
      
specw g (Fix Static e, tau) = 
   do (delta, (e', tautau')) <- specw g (e,STFun Static tau tau)
      tau'                   <- freshresidualtype
      (_, delta')            <- entlw delta (IsFixS tautau' tau')
      return (delta++delta', (e', tau'))

specw g (Fix Dynamic e, tau) = 
   do (delta, (e', tautau')) <- specw g (e,STFun Dynamic tau tau)
      tau'                   <- freshresidualtype
      (_, delta')            <- entlw delta (Unify tautau' (RTFun tau' tau'))
      return (delta++delta', (RFix e', tau'))

specw g (Let ann decls e, tau) = 
   do (delta1, decls', xdecls') <- specwdecls ann g decls
      (delta2, (e',tau'))       <- specw (g `multiextendedWith` xdecls') (e,tau)
      let e'' = case ann of
                 Dynamic -> RLet decls' e'
                 Static  -> e'
      return (delta1++delta2, (e'', tau'))
      
-------------------
-- Basic types: Int
specw _ (Int Static  n, STBase Static  SBInt) = return ([], (rvoid, RTInt n))

specw _ (Int Dynamic n, STBase Dynamic SBInt) = return ([], (RInt n, RTIntD))

specw g (Lift e, tau@(STBase Dynamic base)) = 
   do (delta, (_,tau')) <- specw g (e, STBase Static base)
      (v, delta')       <- entlw delta (assertIs base tau')
      (_, rtau')        <- srw tau
      return (delta++delta', (evidence2residual v, rtau'))

specw g (ArithOp Static  nop e1 e2, tau@(STBase Static  SBInt)) =
   do (delta1, (_, tau1')) <- specw g (e1, tau)
      (delta2, (_, tau2')) <- specw g (e2, tau)
      tau'                 <- freshresidualtype
      (_, delta')          <- entlw (delta1++delta2)
                                    (IsArithOp nop tau' tau1' tau2')
      return (delta1++delta2++delta', (rvoid, tau'))

specw g (ArithOp Dynamic nop e1 e2, tau@(STBase Dynamic SBInt)) =
   do (delta1, (e1', tau1')) <- specw g (e1, tau)
      (delta2, (e2', tau2')) <- specw g (e2, tau)
      (_,delta')             <- entlw (delta1++delta2) 
                                      (Unify tau1' RTIntD)
      (_,delta'')            <- entlw (delta1++delta2++delta') 
                                      (Unify tau2' RTIntD)
      return (delta1++delta2++delta'++delta'', (RArithOp nop e1' e2', RTIntD))
                      
--------------------
-- Basic types: Bool
specw _ (Bool Static  b, STBase Static  SBBool) = return ([],(rvoid, RTBool b))

specw _ (Bool Dynamic b, STBase Dynamic SBBool) = return ([],(RBool b, RTBoolD))

specw g (If Static  e e1 e2, tau) =
   do (delta,  (_,   taub')) <- specw g (e, sTypeBool Static)
      -- Too eager!! It actually PERFORMS unification in BOTH branches!!!!
      --  (Is this still true?? We're using Unify...)
      (delta1, (e1', tau1')) <- specw g (e1, tau)  
      (delta2, (e2', tau2')) <- specw g (e2, tau)
      (v, deltaA)            <- entlw delta (IsBool taub')
      taur'                  <- freshresidualtype
      let deltaB = delta++deltaA 
      (_, deltaC)            <- entlw deltaB
                                      (IsIf taur' taub' tau1' tau2')

      (deltar, taur'')       <- srw tau
      let deltaD = deltaB++deltaC++deltar
      (_, deltares)          <- entlw deltaD 
                                      (Unify taur' taur'')

      let delta1' = map (\(v,d) -> (v, Guarded    taub' d)) delta1
      let delta2' = map (\(v,d) -> (v, GuardedNeg taub' d)) delta2
      return (deltares++deltaD++delta1'++delta2', (REIf v e1' e2', taur'))

specw g (If Dynamic e e1 e2, tau) =
   do (delta,  (e',  tau'))  <- specw g (e, sTypeBool Dynamic)
      (delta1, (e1', tau1')) <- specw g (e1, tau)
      (delta2, (e2', tau2')) <- specw g (e2, tau)
      (_,delta')             <- entlw (delta++delta1++delta2) 
                                      (Unify tau1' tau2')
      return (delta1++delta2++delta', (RIf e' e1' e2', tau1'))

specw g (BoolOp  Static  bop e1 e2, tau@(STBase Static  SBBool)) =
   do (delta1, (_, tau1')) <- specw g (e1, tau)
      (delta2, (_, tau2')) <- specw g (e2, tau)
      tau'                 <- freshresidualtype
      (_, delta')          <- entlw (delta1++delta2)
                                    (IsBoolOp bop tau' tau1' tau2')
      return (delta1++delta2++delta', (rvoid, tau'))

specw g (BoolOp  Dynamic bop e1 e2, tau@(STBase Dynamic SBBool)) =
   do (delta1, (e1', tau1')) <- specw g (e1, tau)
      (delta2, (e2', tau2')) <- specw g (e2, tau)
      (_,delta')             <- entlw (delta1++delta2) 
                                      (Unify tau1' RTBoolD)
      (_,delta'')            <- entlw (delta1++delta2++delta') 
                                      (Unify tau2' RTBoolD)
      return (delta1++delta2++delta'++delta'', (RBoolOp bop e1' e2', RTBoolD))
                      
specw g (RelOp   Static  rop tau e1 e2, STBase Static  SBBool) =
   do (delta1, (_, tau1')) <- specw g (e1, tau)
      (delta2, (_, tau2')) <- specw g (e2, tau)
      tau'                 <- freshresidualtype
      (_, delta')          <- entlw (delta1++delta2)
                                    (IsRelOp rop tau' tau1' tau2')
      return (delta1++delta2++delta', (rvoid, tau'))

specw g (RelOp   Dynamic rop tau e1 e2, STBase Dynamic SBBool) =
   do (delta1, (e1', tau1')) <- specw g (e1, tau)
      (delta2, (e2', tau2')) <- specw g (e2, tau)
      (_,delta')             <- entlw (delta1++delta2) (Unify tau1' tau2')
      return (delta1++delta2++delta', (RRelOp rop e1' e2', RTBoolD))
                      
---------
-- Tuples
specw g (Tuple Dynamic es, STTuple Dynamic taus) = 
   do (delta, (es', taus')) <- specws g (es, taus)
      return (delta, (RTuple es', RTTuple taus'))

specw g (Tuple Static es, STTuple Static taus) = 
   do (delta, (es', taus')) <- specws g (es, taus)
      return (delta, (RTupleS es', RTTupleS taus'))

specw g (Proj Dynamic i n taue e, tau) =
   do (delta, (e', taue')) <- specw g (e, taue)
      taus'                <- mapM (\_ -> freshresidualtype) [1..n]
      (_,delta')           <- entlw delta (Unify taue' (RTTuple taus'))
      if i>length taus' then fail (outboundIn "Proj Dynamic") else do
      let tau'             = taus'!!(i-1)
      return (delta++delta', (RProj i n e', tau'))
   
specw g (Proj Static i n taue e, tau) =
   do (delta, (e', taue')) <- specw g (e, taue)
      taus'                <- mapM (\_ -> freshresidualtype) [1..n]
      (_,delta')           <- entlw delta (Unify taue' (RTTupleS taus'))
      if i>length taus' then fail (outboundIn "Proj Static") else do
      let tau'             = taus'!!(i-1)
      return (delta++delta', (RProjS i n e', tau'))

---------------
-- Polyvariance
specw g (Poly e, STPoly tau) = 
   do (delta, (e',tau')) <- specw g (e, tau)
      (delta'', e'')     <- simplifyAndSolve False g delta tau' e'
      (e''',sigma)       <- constructScheme g delta'' tau' e''
      h                  <- freshevidence
      s                  <- freshresidualscheme
      return ([(h,IsMG sigma s)],(REFill h e''',RTPoly s))

specw g (Spec e, tau) =
   do (delta, (e',polysigma)) <- specw g (e, STPoly tau)
      sigma                   <- freshresidualscheme
      (_,delta'')             <- entlw delta (Unify polysigma (RTPoly sigma))
      (delta', tau')          <- srw tau
      h                       <- freshevidence
      let deltar = (   delta' 
                    ++ [ (h, IsMG sigma (scheme tau')) ] 
                    ++ delta ++ delta''
                   )
      return (deltar, (REFill h e', tau'))

specw g (Polymorph e, STPoly tau) = 
   do (delta, (e',tau')) <- specw g (e, tau)
      (delta'', e'')     <- simplifyAndSolve False g delta tau' e'
      (e''',sigma)       <- constructScheme g delta'' tau' e''
      h                  <- freshevidence
      s                  <- freshresidualscheme
      return ([(h,IsMGmorph (RTStaticExp e''') sigma s)],(REFill h e''',RTPoly s))

--------------
-- Datatypes
specw g (Construc Static  c tes, STData Static _ _) = 
   do let (es, taus) = unzip tes
      (delta, (es', taus')) <- specws g (es, taus)
      return (delta, (RTupleS es', RTConstruc c taus'))

specw g (Construc Dynamic c tes, d@(STData Dynamic name _)) = 
   do let (es, args)        =  unzip tes
      (delta, (es', args')) <- specws g (es, args)
      (deltac, tauc)        <- constructorRType c
      taud                  <- freshresidualtype
      let tauc2             =  buildConstrType args' taud
      (_,delta')            <- entlw delta (Unify tauc2 tauc)
      return (delta++deltac++delta', (RConstruc c es', taud))

specw g (Case Static  e t cs, tau) =
   do (deltae,  (e', te')) <- specw g (e,t)
      (deltat, t')         <- srw t
      case deltat of
        [(v,IsConstrOf _ _)] ->
         do let deltaA = deltae++deltat
            (_,deltate)    <- entlw deltaA (Unify t' te')
            let deltaB = deltaA++deltate

            (cnames, deltase, (tpes, pets)) <- specwCases g te' cs tau

            let pts = map fst pets
            cs' <- newTypesFor freshresidualtype pts
            deltast <- mapM (\(ti,eti)->
                              entlw deltaB (Unify ti eti)
                            ) (zip (map snd cs') (map snd pets))
            let deltaC = concat (map snd deltast)
            let deltaD = deltaB++deltaC
            ntau'            <- freshresidualtype
            (_,deltacs)      <- entlw deltaD (IsCase ntau' te' cs')
            (deltatau, tau') <- srw tau
            let deltaE = deltaD++deltatau
            (_, deltant)     <- entlw deltaE (Unify ntau' tau')
            
            let gdelta = concat (
                 map (\(c,ds)-> 
                       map (\(v,d)->(v,GuardedConstr c te' d)) ds
                     )
                     (zip cnames deltase))
            let deltaF = deltaE++deltacs++gdelta++deltant
            tpes' <- mapM (bindPattern e' te') (zip cnames tpes)
            return (deltaF, (RECase v tpes', tau'))
        _ -> fail "Case with non-data argument!"
     
specw g (Case Dynamic e t cs, tau) = 
   do (delta1, (e',  te'))      <- specw g (e, t)
      (deltat, t')              <- srw t
      let delta' = delta1++deltat
      (_,deltat')               <- entlw delta' (Unify t' te')
      (_, deltas, (tpes, pets)) <- specwCases g te' cs tau
      let delta = delta' ++ deltat' ++ concat deltas
      (_,delta3) <- entlsw delta [ Unify te' patt | (patt,_) <- pets ]
      et <- freshresidualtype
      (_,delta4) <- entlsw delta [ Unify et expt | (_,expt) <- pets ]
      return (delta++delta3++delta4, (RCase e' tpes, et))

--------------
-- Other cases
specw _ _     = error (canthpnSo "Something's wrong in source type checking.")
   
-----------------------------------------------
-- Specialising a list
-----------------------------------------------
specws _ ([],[])      = return ([],([],[]))
specws g (e:es, t:ts) = do (delta, (e',t'))    <- specw g (e,t)
                           (deltas, (es',ts')) <- specws g (es,ts)
                           return (delta++deltas, (e':es',t':ts'))
specws _ _            = error "Cannot happen! Tuple with different arity."

-----------------------------------------------
-- Specialising declarations
-----------------------------------------------
specwdecls _ _ [] = return ([], [], [])
specwdecls ann g (d:ds) =
   do (deltad, d', xd')    <- specwdecl ann g d
      (deltads, ds', xds') <- specwdecls ann g ds
      return (deltad++deltads, d':ds', xd':xds')

specwdecl ann g (xt@(x,t),e) =
   do (delta, (e',t')) <- specw g (e, t)
      x'               <- freshresidualtermBasedOn x
      let xe' = case ann of
                  Dynamic -> RVar x'
                  Static  -> e'
      return (delta, (x',e'), (xt,(xe', t')))

-----------------------------------------------
-- Specialising datatypes
-----------------------------------------------
specDatas :: [DataDef] -> SpecM [RDataDef]
specDatas ds = do fws <- newTypesFor freshresidualtype (map datadefName ds)
                  mapM_ forwardDataFor fws
                  let fwvbles = map snd fws
                  ds' <- specDatas' fwvbles ds
                  resolveForwards fws
                  unfoldOnly fwvbles ds'

resolveForwards [] = return ()
resolveForwards ((dn,td):fws) =
   do (_,d) <- resolveDataNamed dn
      yes <- efectiveUnify td d
      if yes 
       then resolveForwards fws
       else error "Cannot happen! Forward datatype is not a variable..."

specDatas' _      []     = return []
specDatas' noargs (d:ds) = do d'  <- specData noargs d
                              ds' <- specDatas' noargs ds
                              return (d':ds')

specData noargs (DataDef _ name _ cs) = 
  do (delta, cs') <- specDataConstrs cs
     let csargs = map constrArgs cs'
     tsdelta <- typevars delta
     tscs    <- typevars csargs
     ssdelta <- schemevars delta
     sscs    <- schemevars csargs
     let targs = map RTVar (tsdelta`union`tscs) \\ noargs
     let sargs = map RSVar (ssdelta`union`sscs) 
     let d' = RDataDef name targs sargs delta cs'
     addData d'
     return d'

specDataConstrs [] = return ([], [])
specDataConstrs (Constructor c cargs:cs) =
  do (delta, cargs') <- specConstrArgs cargs
     (delta', cs')   <- specDataConstrs cs
     return (delta++delta', Constructor c cargs':cs')

specConstrArgs [] = return ([],[])
specConstrArgs (t:ts) = 
  do (delta, t')   <- srw_for_datadef t
     (delta', ts') <- specConstrArgs ts
     return (delta++delta', t':ts')

-----------------------------------------------
-- Specialising case alternatives
-----------------------------------------------
specwCases _ _ [] _ = return ([],[],([],[]))
specwCases g tres (c:cs) tau = 
  do (c , deltace , (tpe , pet))  <- specwCase g tres c tau
     (cs, deltacse, (tpes, pets)) <- specwCases g tres cs tau
     return (c:cs, deltace:deltacse, (tpe:tpes, pet:pets))

specwCase g tres ((pat,tp),e) te = 
  do let c = patternMainConstr (stripEtas pat)
     (g', deltap, (pat', tp')) <- specwpattern g tres pat
        --   tres is the residual of tp
        --   g' is the assignment including the bindings of pat
     (deltae, (e'  , te')) <- specw g' (e,te)
     return (c, deltap++deltae, ((pat',e'), (tp', te')))

----------------------------
-- Patterns 
specwpattern g tres pat = specwpat1 g [] tres [] [] pat

specwpat1 g delta tres es ts (App ann ti e (Var xi)) =
  do (ds, ti')      <- srw ti
     xi'            <- freshresidualtermBasedOn xi
     let g' = g `extendedWith` ((xi,ti),(RVar xi',ti'))
     specwpat1 g' (delta++ds) tres (RVar xi':es) (ti':ts) e
specwpat1 g delta tres es ts pat = specwpat2 g delta tres [] es ts pat

specwpat2 g delta tres rxs es ts (Lam _ x t e) = 
  specwpat2 g delta tres ((Var x,t):rxs) es ts e
specwpat2 g delta tres rxs es ts (Construc ann c args) =
  if (reverse rxs == args)
   then case ann of
         Static  -> return (g, delta, (RTupleS es, RTConstruc c ts))
         Dynamic -> 
           do (deltac, tauc)        <- constructorRType c
              let tauc2             =  buildConstrType ts tres
              (_,delta')            <- entlw [] (Unify tauc2 tauc)
              return (g, delta++deltac++delta', (RConstruc c es, tres))
   else fail (canthpnSo "Pattern does not satisfy check.")

-----------------
bindPattern :: ResTerm -> ResType -> (ConstructorName,(ResTerm,ResTerm)) 
               -> SpecM (ConstructorName, ResTerm)
bindPattern e t (c,(RTupleS pivs,ei)) =
  do let n = length pivs
     vs <- mapM (\v -> case v of
                        RVar i -> return i
                        _ -> fail (canthpnSo ("Residual of a pattern "
                                  ++"must be a tuple of variables."))
                )
                pivs
     let rs = map (\i -> RProjS i n e) [1..n]
     let ei' =  accuml ressubst ei vs rs
     return (c,ei')
bindPattern _ _ _ = 
  fail (canthpnSo "Residual of a pattern must be a tuple.")

-- Static functions all at once!
specwLamS g xts ts (Lam Static x t e1, STFun Static tau2 tau1) =  
  specwLamS g ((x,t):xts) (tau2:ts) (e1, tau1)
specwLamS g xts ts (e1,tau1) =
  specw g (foldr (\(x,t) e -> Lam Dynamic x t e) e1 (reverse xts)
          ,foldr (\tau2 tau -> STFun Dynamic tau2 tau) tau1 (reverse ts)
          )

specwAppS g es ts tres (App Static tau2 e1 e2, tau1) =
  specwAppS g (e2:es) (tau2:ts) tres (e1, STFun Static tau2 tau1)
specwAppS g es ts tres (e1, tau1) =
  do (delta1, (e1', tau21')) <- specw g (e1, tau1)
     (delta2, (e's, tau's))  <- specws g (es, ts)
     (delta3, tres')         <- srw tres
     taue'                   <- freshresidualtype
     taus'                   <- freshresidualtype
     let delta                = delta1 ++ delta2 ++ delta3
     let tfun = foldr RTFun tres' tau's
     (v, delta')             <- entlw delta
                                      (IsFunS tau21'                                          
                                              (RTTypedS taus' taue' 
                                                        (scheme tfun))
                                      )     
     let e' = foldr (\e v -> REAppS v e)
                    (REAppS (evidence2residual v) e1') 
                    e's
     return (delta++delta', (e', tres'))

