module SpecState where

import MyPrelude(canthpnSo)

import MetaVbles
import qualified SourceTypeCheckMonad
import ResidualLang
import ResidualDataManagement

----------------------------------
-- Type for SpecM states
----------------------------------
data State = SS SpecManager GenManager Others 
           deriving (Read, Show)

type SpecManager =
      ( MetaVble ResidualType             -- type assignments
      , MetaVble ResidualTypeScheme       -- scheme assignments
      , ( [MetaVble ResidualType]         -- temporary type assignments
        , [MetaVble ResidualTypeScheme]   -- temporary scheme assignments
        )
      , MetaVble EvidenceTerm             -- evidence vbles
      , (RDataManager                     -- Datatypes
        ,SourceTypeCheckMonad.SourceState
        )
      )
type GenManager =
      ( [(String, Int)]                 -- vbles already used
      , VbleGenerator                   -- generic vbles generator
      )

type VbleGenerator = (Int,   Int,    Int,     [(Int,Int)], [(Int,Int)])
                  -- (level, tyvars, schvars, tyseen,      schseen    )

type Others = (Bool, Bool, Bool, Tracer) 
              -- Flags: 'Solve contexts' 
              --        'Eliminate evidence'
              --        'Perform tracing'
              --   and trace information

newtype Tracer = TR () deriving (Read, Show)
--newtype Tracer = TR ShowS
--instance Show Tracer where
--  showsPrec _ (TR s) = s

----------------------------------
-- Initial state
initialTracer = TR ()
--initialTracer = TR (showString "\n")

initialState :: State
initialState = SS ( initialmetavbles, initialmetavbles
                  , ([], [])
                  , initialmetavbles
                  , (initialRDataMngr
                    , SourceTypeCheckMonad.initialSourceState
                    )
                  )
                  ([],initialVbleGenerator)
                  (False, False, False, initialTracer)

initialVbleGenerator :: VbleGenerator
initialVbleGenerator = (1,0,0,[],[])

----------------------------------
-- Toogling the flags
peekSolveContextsFlag   (SS _  _  (bf,_,_  ,_ )) = bf
pokeSolveContextsFlag b (SS mv vg (_,ee,trf,tr)) = SS mv vg (b ,ee,trf,tr)

peekEvidenceElimFlag    (SS _  _  (_,ee,_  ,_ )) = ee
pokeEvidenceElimFlag  b (SS mv vg (bf,_,trf,tr)) = SS mv vg (bf,b ,trf,tr)

----------------------------------
-- Tracing
peekTrace     (SS _  _  (_ ,_ ,_  ,TR s)) = s --s ""
updateTrace f (SS mv vg (bf,ee,trf,TR s)) = SS mv vg (bf,ee,trf,TR (f s))

peekTraceFlag   (SS _  _  (_ ,_ ,trf,_ )) = trf
pokeTraceFlag b (SS mv vg (bf,ee,_  ,tr)) = SS mv vg (bf,ee,b,tr)

---------------------------------
-- Substitutions
getSubst    (SS (rt,_,_,_,_) _ _)      = rt
getTmpSubst (SS (_,_,(rts,_),_,_) _ _) = rts

----------------------------------
-- Operations for metavariables of different kinds
residualtype   :: ST (MetaVble ResidualType) a -> ST State a
residualscheme :: ST (MetaVble ResidualTypeScheme) a -> ST State a
tmptype        :: ST (MetaVble ResidualType) a -> ST State a
tmpscheme      :: ST (MetaVble ResidualTypeScheme) a -> ST State a
evidence       :: ST (MetaVble EvidenceTerm) a -> ST State a

residualtype   f (SS (t ,s ,gtgs        ,v ,dd) vg fs) = let (i,t' ) = f t  in
              (i, SS (t',s ,gtgs        ,v ,dd) vg fs)
residualscheme f (SS (t ,s ,gtgs        ,v ,dd) vg fs) = let (i,s' ) = f s in
              (i, SS (t ,s',gtgs        ,v ,dd) vg fs)
tmptype        f (SS (_ ,_ ,([]     ,_ ),_ ,_ ) _  _ ) = error (canthpnSo "Initialize tmp.")
tmptype        f (SS (t ,s ,(gt:gts ,gs),v ,dd) vg fs) = let (i,gt') = f gt in
              (i, SS (t ,s ,(gt':gts,gs),v ,dd) vg fs)
tmpscheme      f (SS (_ ,_ ,(_ ,[]     ),_ ,_ ) _  _ ) = error (canthpnSo "Initialize tmp.")
tmpscheme      f (SS (t ,s ,(gt,gs:gss ),v ,dd) vg fs) = let (i,gs') = f gs in
              (i, SS (t ,s ,(gt,gs':gss),v ,dd) vg fs)
evidence       f (SS (t ,s ,gtgs        ,v ,dd) vg fs) = let (i,v' ) = f v  in
              (i, SS (t ,s ,gtgs        ,v',dd) vg fs)

----------------------------------
-- Operations for source typechecking
peekSourceTCstate (SS (_,_,_,_,(_,stcs)) _ _) = stcs
pokeSourceTCstate stcs (SS (t,s,tt,v,(dd,_   )) cs fs) = 
                        SS (t,s,tt,v,(dd,stcs)) cs fs

----------------------------------
-- Operations for datatypes
peekData (SS (_,_,_,_,(dd,_)) _ _) = dd
updateData f (SS (t,s,tts_tss,v,(dd  ,stcs)) cs fs) =
              SS (t,s,tts_tss,v,(f dd,stcs)) cs fs

----------------------------------
-- Operations for temporary substition
peektmptype (SS (_,_,(tts,_),_,_) _ _) = tts
updatetmptype f (SS (t,s,(tts  ,tss),v,dd) cs fs) =
                 SS (t,s,(f tts,tss),v,dd) cs fs

peektmpscheme (SS (_,_,(_,tss),_,_) _ _) = tss
updatetmpscheme f (SS (t,s,(tts,tss  ),v,dd) cs fs) =
                   SS (t,s,(tts,f tss),v,dd) cs fs

stateConfirmUnification (SS (t,s,([tt],[ts]),v,dd) cs fs) =
 let tt_table = fst (getflattable tt)
     ts_table = fst (getflattable ts)
     nt = snd (setmetavbles tt_table t)
     ns = snd (setmetavbles ts_table s)
  in SS (nt,ns,([],[]),v,dd) cs fs
stateConfirmUnification (SS (t,s,(tt:tt2:tts,ts:ts2:tss),v,dd) cs fs) =
 let tt_table = fst (getflattable tt)
     ts_table = fst (getflattable ts)
     ntt = snd (setmetavbles tt_table tt2)
     nts = snd (setmetavbles ts_table ts2)
  in SS (t,s,(ntt:tts,nts:tss),v,dd) cs fs

----------------------------------
-- Operations for generic variables
pokeGeneric gg (SS mv (cs,_) fs) = SS mv (cs,gg) fs
updateQuantifierScope f (SS mv (cs,(gl ,gti,gsi,ta,sa)) fs) =
           let gl1 = f gl
            in gl1 `seq` SS mv (cs,(gl1,gti,gsi,ta,sa)) fs
      
peekScopeLevel (SS _ (_,(gl,_  ,_  ,_  ,_  )) _) = gl
peekMaxRTGen   (SS _ (_,(_ ,gti,_  ,_  ,_  )) _) = gti
peekMaxRSGen   (SS _ (_,(_ ,_  ,gsi,_  ,_  )) _) = gsi
peekCurRTGen   (SS _ (_,(_ ,_  ,_  ,gta,_  )) _) = gta
peekCurRSGen   (SS _ (_,(_ ,_  ,_  ,_  ,gsa)) _) = gsa
updateRTGen f  (SS mv (cs,(gl,gti ,gsi ,ta ,sa )) fs) =
      let (res, gti',ta') = f (gti, ta)
       in (res, SS mv (cs,(gl,gti',gsi ,ta',sa )) fs)
updateRSGen f  (SS mv (cs,(gl,gti ,gsi ,ta ,sa )) fs) =
      let (res, gsi',sa') = f (gsi, sa)
       in (res, SS mv (cs,(gl,gti ,gsi',ta ,sa')) fs)

projectRTGen f (SS _ (_,(_,_,_,ta,_)) _) = f ta
projectRSGen f (SS _ (_,(_,_,_,_,sa)) _) = f sa
      
----------------------------------
-- Operations for residual term variables
updateTermVbles f (SS mv (cs,g) fs) = 
  let (res,cs') = f cs
   in (res, SS mv (cs',g) fs)

---------------------------------------------------------------
-- Auxiliaries
---------------------------------------------------------------
getNext x [] = (0, [(x,1)])
getNext x (xix@(x',ix):cs) =
         if x==x' 
          then (ix, (x,ix+1):cs)
          else let (ix, cs') = getNext x cs
                in (ix, xix:cs')

lookUpdate t i [] = let i1 = i+1 in (i, i1, [(t,i)])
lookUpdate t i s@(tit@(t',it'):tits) =
  if t==t'
   then (it', i, s)
   else let (it, i', tits') = lookUpdate t i tits
         in (it, i', tit:tits')

